#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

from pprint import pprint

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from matplotlib.colors import colorConverter

import euclid3 as eu
import numpy as np
import pandas as pd
import cx_Oracle as cx
import copy as cp

from skspatial.objects import Points, Plane
from skspatial.plotting import plot_3d

import OCC.Core.AIS as ais
import OCC.Core.BRepBuilderAPI as brep
from OCC.Core.BRep import BRep_Builder
from OCC.Core.BRepAlgoAPI import BRepAlgoAPI_Cut
import OCC.Core.GC as gc
import OCC.Core.gp as gp
import OCC.Core.TopoDS as tp
import OCC.Core.Quantity as qu

from OCCUtils.Topology import Topo

def inject(id_, geom, table, data=None):
    with cx.connect('scott','tiger','localhost:1521/orcl') as con:
        with con.cursor() as cur:
            
            typeObj = con.gettype("MDSYS.SDO_GEOMETRY")
            elementInfoTypeObj = con.gettype("MDSYS.SDO_ELEM_INFO_ARRAY")
            ordinateTypeObj = con.gettype("MDSYS.SDO_ORDINATE_ARRAY")
            obj = typeObj.newobject()
            obj.SDO_GTYPE = geom.gtype
            obj.SDO_SRID = geom.srid
            obj.SDO_ELEM_INFO = elementInfoTypeObj.newobject()
            obj.SDO_ELEM_INFO.extend(geom.geom.elem_info(True))
            obj.SDO_ORDINATES = ordinateTypeObj.newobject()
            obj.SDO_ORDINATES.extend(geom.geom.ord_info(True))
            
            if  id_ is not None:
                
                data = {"id": id_,
                        "geom": obj}        
                statement = "INSERT INTO %s (ID, GEOM) VALUES (:id, :geom)"%table
                
            elif data is not None:
                data.update({"geom": obj})
                fieldnames = ",".join([each.upper() for each in data] )
                values = ",".join([":"+each.lower() for each in data] )
                statement = "INSERT INTO %s (%s) VALUES (%s)" %(table, fieldnames, values) 
                
            cur.execute(statement, data)
            
        con.commit()
        
def coord_from_row(elem_array, row):
            
    '''Returns row in ordinate array referenced by element array row
    ''' 
    return int((elem_array[row,0] - 1  ) / 3)

def parse_elem_info(elem_array,
                    element,
                    collection,
                    solid,
                    composite,
                    surf,
                    void_,
                    face,
                    int_face,
                    hole,
                    int_hole,
                    last_coord,
                    ):

    '''elem_array'''

    # flags for use during parsing
    intsurf = False
    last_element_belonged_to_composite = False
    solids_to_read = 0
    last_element_was_a_face = False
    last_element_was_a_hole = False

    # initial element ids

    collection_id = 0
    composite_id = 0
    solid_id = 0
    surf_id = 0
    void_id = 0
    hole_id = 0
    face_id = 0
    int_face_id = 0
    int_hole_id = 0

    number_of_elements = elem_array.shape[0]

    for each_row in range(0, number_of_elements):
        etype = elem_array[each_row][1]

        if etype == 1008:

            # Handle Composite solid                

            raise Exception ("Functionality does not yet exist to parse "+
                             "Composite Solids")

            composite_id +=1

            # Flags for showing that next n solids to belong composite                
            belong_to_composite = True 
            solids_to_read = elem_array[each_row, 2]


            collection[collection_id]['composites'].append(composite_id)

            composite[composite_id] = {'elem_row': each_row,
                                       'solids': [],
                                                'start' : coord_from_row(elem_array,each_row),
                                                'parent': (10, collection_id)}

            if last_element_was_a_hole:
                if intsurf:
                    int_hole[int_hole_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    void_[void_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    hole[hole_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    surf[surf_id]['end'] = coord_from_row(elem_array,each_row) - 1

            elif last_element_was_a_face:
                if intsurf:
                    int_face[int_face_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    void_[void_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    face[face_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    surf[surf_id]['end'] = coord_from_row(elem_array,each_row) - 1                

            if last_element_was_a_hole or last_element_was_a_face:
                solid[solid_id]['end'] = coord_from_row(elem_array,each_row) - 1

            element.append((9, composite_id))

            last_element_was_a_face = False
            last_element_was_a_hole = False 

            last_element_belonged_to_composite = True
            intsurf = False # Reset Internal Surface to External surface

        elif etype == 1007:
            # Handle Solid

            solid_id += 1

            if solids_to_read == 0:
                last_element_belonged_to_composite = False

            if elem_array[each_row, 2] == 3:
                raise Exception("Functionality does not yet exist to parse Solid Cube")


            solid[solid_id]= {'elem_row' : each_row,
                              'surfaces' : [],
                                             'voids'    : [],
                                             'start'    : coord_from_row(elem_array,each_row)}                

            if solids_to_read:
                composite[composite_id]['solids'].append(solid_id)
                solid[solid_id]['parent'] = (9, composite_id)
                solids_to_read -= 1
            else:
                collection[collection_id]['solids'].append(solid_id)
                solid[solid_id]['parent'] = (10, collection_id)              

            solid[solid_id - 1]['end'] = coord_from_row(elem_array,each_row) - 1

            element.append((8, solid_id))

            last_element_was_a_face = False
            last_element_was_a_hole = False 

            intsurf = False # Reset Internal Surface to External surface

        elif etype == 1006:
            # Handle exterior surfaces

            surf_id += 1

            surf[surf_id] = {'elem_row': each_row,
                                      'parent'  : (8, solid_id),
                                      'start'   : coord_from_row(elem_array,each_row),
                                      'faces'   : []}

            solid[solid_id]['surfaces'].append(surf_id)

            element.append((7,surf_id))

            if last_element_was_a_hole:
                if intsurf:
                    int_hole[int_hole_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    void_[void_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    hole[hole_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    surf[surf_id]['end'] = coord_from_row(elem_array,each_row) - 1

            elif last_element_was_a_face:
                if intsurf:
                    int_face[int_face_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    void_[void_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    face[face_id]['end'] = coord_from_row(elem_array,each_row) - 1
                    surf[surf_id]['end'] = coord_from_row(elem_array,each_row) - 1

            intsurf = False
            last_element_was_a_face = False
            last_element_was_a_hole = False  

        elif etype == 2006:
        # Handle interior surfaces (voids)

            void_id += 1

            void_[void_id] = {'elem_row': each_row,
                                      'parent'  : (8, solid_id),
                                      'start'   : coord_from_row(elem_array,each_row),
                                      'faces'   : []}

            solid[solid_id]['voids'].append(void_id)

            element.append((6,void_id))


            if last_element_was_a_face or last_element_was_a_hole:
                if not intsurf:
                    surf[surf_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    void_[void_id - 1]['end'] = coord_from_row(elem_array,each_row) - 1

            if last_element_was_a_hole:
                if intsurf:
                    int_hole[int_hole_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    hole[hole_id]['end'] = coord_from_row(elem_array,each_row) - 1

            elif last_element_was_a_face:
                if intsurf:
                    int_face[int_face_id]['end'] = coord_from_row(elem_array,each_row) - 1
                else:
                    face[face_id]['end'] = coord_from_row(elem_array,each_row) - 1

            intsurf = True
            last_element_was_a_face = False
            last_element_was_a_hole = False                

        elif etype == 1003 or etype ==3:
        # Handle interior holes and exterior faces

            if intsurf:                    

                int_hole_id += 1 
                int_hole[int_hole_id] = {'elem_row': each_row,
                                                  'parent'  : (4, int_face_id),
                                                  'start'   : coord_from_row(elem_array,each_row)}

                int_face[int_face_id]['holes'].append(int_hole_id)

                if last_element_was_a_hole:
                    int_hole[int_hole_id-1]['end'] = coord_from_row(elem_array,each_row) - 1

                if last_element_was_a_face:
                    int_face[int_face_id-1]['end'] = coord_from_row(elem_array,each_row) - 1                    

                element.append((2,int_hole_id))

                last_element_was_a_face = False
                last_element_was_a_hole = True

            else:
                face_id += 1   

                face[face_id] = {'elem_row': each_row,
                                          'holes'   : [],
                                          'parent'  : (7, surf_id),
                                          'start'   : coord_from_row(elem_array,each_row)}

                surf[surf_id]['faces'].append(face_id)

                element.append((5,face_id))                    

                if last_element_was_a_hole:
                    hole[hole_id]['end'] = coord_from_row(elem_array,each_row) - 1

                if last_element_was_a_face:
                    face[face_id-1]['end'] = coord_from_row(elem_array,each_row) - 1

                last_element_was_a_face = True
                last_element_was_a_hole = False


        elif etype == 2003:
            # Handle holes and interior faces 

            if intsurf:
                int_face_id +=1 

                int_face[int_face_id] = {'elem_row': each_row,
                                                  'start'   : coord_from_row(elem_array,each_row),
                                                  'holes'   : [],
                                                  'parent'  : (6, void_id)}

                void_[void_id]["faces"].append(int_face_id)

                element.append((4,int_face_id))

                if last_element_was_a_hole:
                    int_hole[int_hole_id]['end'] = coord_from_row(elem_array,each_row) - 1

                if last_element_was_a_face:
                    int_face[int_face_id-1]['end'] = coord_from_row(elem_array,each_row) - 1                    

                last_element_was_a_face = True 
                last_element_was_a_hole = False

            else:
                hole_id += 1 

                hole[hole_id] = {'elem_row': each_row,
                                          'parent'  : (5, face_id),
                                          'start'   : coord_from_row(elem_array,each_row)}

                face[face_id]['holes'].append(hole_id)

                element.append((3,hole_id))

                if last_element_was_a_hole:
                    hole[hole_id-1]['end'] = coord_from_row(elem_array,each_row) - 1

                if last_element_was_a_face:
                    face[face_id]['end'] = coord_from_row(elem_array,each_row) - 1

                last_element_was_a_face = False
                last_element_was_a_hole = True                    

    # Having parsed each element provide 'end' to the last elements

    #last_coord = np.shape(ordinates)[0]

    collection[collection_id]['end'] = last_coord

    if last_element_belonged_to_composite:
        composite[composite_id]['end'] = last_coord

    solid[solid_id]['end'] = last_coord


    if last_element_was_a_hole:
        if intsurf:
            int_hole[int_hole_id]['end'] = last_coord
            void_[void_id]['end'] = last_coord
        else:
            hole[hole_id]['end'] = last_coord
            surf[surf_id]['end'] = last_coord

    elif last_element_was_a_face:
        if intsurf:
            int_face[int_face_id]['end'] = last_coord
            void_[void_id]['end'] = last_coord
        else:
            face[face_id]['end'] = last_coord
            surf[surf_id]['end'] = last_coord       
            


class SdoElemInfo (object):
    
    '''A class wrapping elem_info and ordinate arrays that have been read from
    Oracle SDO_GEOMETRY object and method for parsing
    
    The class has not been developed yet to handle:
    points, linestrings, heterogenous collections
    
    '''
    
    def __init__(self, gtype, element_info, ordinates):
        '''Arguments:
        
        gtype (integer)         : SDO_GTYPE
        element_info (np.array) : Array of shape (n,3)
        ordinates (np.array)    : Array of shape (n,3)
        '''
        
        self.gtype = gtype                  # GTYPE - not currently used 
        self.ordinates = ordinates          # 
        self.elem_array = element_info      #
        
        
        self.length = self.elem_array.shape[0]
        
        self.element = []  # list referencing each row in element array
        
        element_codes = { 0 : 'point',
                          1 : 'linestring',
                          2 : 'hole on interior face (int_hole)',
                          3 : 'hole on exterior hole (hole)',
                          4 : 'interior face (int_face)',
                          5 : 'exterior face (face)',
                          6 : 'interior surface (void)',
                          7 : 'exterior surface (surf)',
                          8 : 'solid',
                          9 : 'composite',
                          10: 'collection'}

        self.collection = {}
        self.collection[0] = {'elem_row'  : None,
                                    'composites': [0], 
                                    'solids'    : []}

        self.solid ={}
        self.solid[0] = {'elem_row': None,
                               'surfaces': [0],
                               'voids'   : [],
                               'parent'  : (9,0)}

        self.composite = {}
        self.composite[0] = {'elem_row':None, 'solids' :[0]}

        self.surf = {}
        self.surf[0] = {'elem_row': None,
                              'parent'  : (8,0),
                              'faces'   : []}

        self.void = {}    # dictionary of all voids
        self.void[0] = {'elem_row': None,
                              'parent'  : (8,0),
                              'faces'   : []}
  
        # dictionary of all faces and holes including void faces and holes
        self.face = {}       # dict of faces keyed by face_id
        self.int_face = {} 
        self.hole = {}
        self.int_hole = {}

        last_coord = np.shape(self.ordinates)[0]

        parse_elem_info(self.elem_array,
                        self.element,
                        self.collection,
                        self.solid,
                        self.composite,
                        self.surf,
                        self.void,
                        self.face,
                        self.int_face,
                        self.hole,
                        self.int_hole,
                        last_coord,
                        )
        
        
        def collect(geomdict):
            '''Function for returning list of elements by element row
            Used for legacy functions
            '''
            keys = geomdict.keys()
            ret = []
            for each in keys:
                if each:
                    ret.append(geomdict[each]['elem_row'])
            return ret
                    
        self.collections = collect(self.collection)
        self.composites  = collect(self.composite)
        self.solids      = collect(self.solid) 
        self.ext_surf    = collect(self.surf)     
        self.int_surf    = collect(self.void)
        self.ext_ring    = collect(self.face)
        self.int_ring    = collect(self.int_face)
        self.ext_holes   = collect(self.hole)
        self.int_holes   = collect(self.int_hole)
        
    def parse_elem_info_(self):
    
        '''elem_array'''
    
        
        def coord_from_row(row):
            
            '''Returns row in ordinate array referenced by element array row
            ''' 
            return int (( self.elem_array[row,0] - 1  ) / 3)
        
        # flags for use during parsing
        intsurf = False
        last_element_belonged_to_composite = False
        solids_to_read = 0
        last_element_was_a_face = False
        last_element_was_a_hole = False
        
        # initial element ids
                
        collection_id = 0
        composite_id = 0
        solid_id = 0
        surf_id = 0
        void_id = 0
        hole_id = 0
        face_id = 0
        int_face_id = 0
        int_hole_id = 0
        
        number_of_elements = self.elem_array.shape[0]
        
        for each_row in range(0, number_of_elements):
            etype = self.elem_array[each_row][1]
            
            if etype == 1008:
                
                # Handle Composite solid                
                
                raise Exception ("Functionality does not yet exist to parse "+
                                 "Composite Solids")
                
                composite_id +=1
                
                # Flags for showing that next n solids to belong composite                
                belong_to_composite = True 
                solids_to_read = self.elem_array[each_row, 2]
                
                
                self.collection[collection_id]['composites'].append(composite_id)
                
                self.composite[composite_id] = {'elem_row': each_row,
                                                'solids': [],
                                                'start' : coord_from_row(each_row),
                                                'parent': (10, collection_id)}
                
                if last_element_was_a_hole:
                    if intsurf:
                        self.int_hole[int_hole_id]['end'] = coord_from_row(each_row) - 1
                        self.void[void_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.hole[hole_id]['end'] = coord_from_row(each_row) - 1
                        self.surf[surf_id]['end'] = coord_from_row(each_row) - 1
                        
                elif last_element_was_a_face:
                    if intsurf:
                        self.int_face[int_face_id]['end'] = coord_from_row(each_row) - 1
                        self.void[void_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.face[face_id]['end'] = coord_from_row(each_row) - 1
                        self.surf[surf_id]['end'] = coord_from_row(each_row) - 1                
                
                if last_element_was_a_hole or last_element_was_a_face:
                    self.solid[solid_id]['end'] = coord_from_row(each_row) - 1
                
                self.element.append((9, composite_id))
                
                last_element_was_a_face = False
                last_element_was_a_hole = False 
                
                last_element_belonged_to_composite = True
                intsurf = False # Reset Internal Surface to External surface
                                
            elif etype == 1007:
                # Handle Solid
                
                solid_id += 1
                
                if solids_to_read == 0:
                    last_element_belonged_to_composite = False
                
                if self.elem_array[each_row, 2] == 3:
                    raise Exception("Functionality does not yet exist to parse Solid Cube")
                
                
                self.solid[solid_id]= {'elem_row' : each_row,
                                             'surfaces' : [],
                                             'voids'    : [],
                                             'start'    : coord_from_row(each_row)}                
                
                if solids_to_read:
                    self.composite[composite_id]['solids'].append(solid_id)
                    self.solid[solid_id]['parent'] = (9, composite_id)
                    solids_to_read -= 1
                else:
                    self.collection[collection_id]['solids'].append(solid_id)
                    self.solid[solid_id]['parent'] = (10, collection_id)              

                self.solid[solid_id - 1]['end'] = coord_from_row(each_row) - 1
                
                self.element.append((8, solid_id))
                
                last_element_was_a_face = False
                last_element_was_a_hole = False 
                
                intsurf = False # Reset Internal Surface to External surface
                
            elif etype == 1006:
                # Handle exterior surfaces
                
                surf_id += 1
                
                self.surf[surf_id] = {'elem_row': each_row,
                                      'parent'  : (8, solid_id),
                                      'start'   : coord_from_row(each_row),
                                      'faces'   : []}
                
                self.solid[solid_id]['surfaces'].append(surf_id)
                
                self.element.append((7,surf_id))
                
                if last_element_was_a_hole:
                    if intsurf:
                        self.int_hole[int_hole_id]['end'] = coord_from_row(each_row) - 1
                        self.void[void_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.hole[hole_id]['end'] = coord_from_row(each_row) - 1
                        self.surf[surf_id]['end'] = coord_from_row(each_row) - 1
                        
                elif last_element_was_a_face:
                    if intsurf:
                        self.int_face[int_face_id]['end'] = coord_from_row(each_row) - 1
                        self.void[void_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.face[face_id]['end'] = coord_from_row(each_row) - 1
                        self.surf[surf_id]['end'] = coord_from_row(each_row) - 1

                intsurf = False
                last_element_was_a_face = False
                last_element_was_a_hole = False  
                
            elif etype == 2006:
            # Handle interior surfaces (voids)
                
                void_id += 1
                
                self.void[void_id] = {'elem_row': each_row,
                                      'parent'  : (8, solid_id),
                                      'start'   : coord_from_row(each_row),
                                      'faces'   : []}
                
                self.solid[solid_id]['voids'].append(void_id)
                
                self.element.append((6,void_id))
                
                
                if last_element_was_a_face or last_element_was_a_hole:
                    if not intsurf:
                        self.surf[surf_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.void[void_id - 1]['end'] = coord_from_row(each_row) - 1
                    
                if last_element_was_a_hole:
                    if intsurf:
                        self.int_hole[int_hole_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.hole[hole_id]['end'] = coord_from_row(each_row) - 1
                        
                elif last_element_was_a_face:
                    if intsurf:
                        self.int_face[int_face_id]['end'] = coord_from_row(each_row) - 1
                    else:
                        self.face[face_id]['end'] = coord_from_row(each_row) - 1
                                           
                intsurf = True
                last_element_was_a_face = False
                last_element_was_a_hole = False                
                
            elif etype == 1003 or etype ==3:
            # Handle interior holes and exterior faces

                if intsurf:                    
                    
                    int_hole_id += 1 
                    self.int_hole[int_hole_id] = {'elem_row': each_row,
                                                  'parent'  : (4, int_face_id),
                                                  'start'   : coord_from_row(each_row)}
                    
                    self.int_face[int_face_id]['holes'].append(int_hole_id)
                    
                    if last_element_was_a_hole:
                        self.int_hole[int_hole_id-1]['end'] = coord_from_row(each_row) - 1
                        
                    if last_element_was_a_face:
                        self.int_face[int_face_id-1]['end'] = coord_from_row(each_row) - 1                    
                    
                    self.element.append((2,int_hole_id))
                    
                    last_element_was_a_face = False
                    last_element_was_a_hole = True
                    
                else:
                    face_id += 1   
                    
                    self.face[face_id] = {'elem_row': each_row,
                                          'holes'   : [],
                                          'parent'  : (7, surf_id),
                                          'start'   : coord_from_row(each_row)}
                    
                    self.surf[surf_id]['faces'].append(face_id)
                    
                    self.element.append((5,face_id))                    
                    
                    if last_element_was_a_hole:
                        self.hole[hole_id]['end'] = coord_from_row(each_row) - 1
                        
                    if last_element_was_a_face:
                        self.face[face_id-1]['end'] = coord_from_row(each_row) - 1
                    
                    last_element_was_a_face = True
                    last_element_was_a_hole = False
                    
                    
            elif etype == 2003:
                # Handle holes and interior faces 
                
                if intsurf:
                    int_face_id +=1 
                    
                    self.int_face[int_face_id] = {'elem_row': each_row,
                                                  'start'   : coord_from_row(each_row),
                                                  'holes'   : [],
                                                  'parent'  : (6, void_id)}
                                       
                    self.void[void_id]["faces"].append(int_face_id)
                    
                    self.element.append((4,int_face_id))
                    
                    if last_element_was_a_hole:
                        self.int_hole[int_hole_id]['end'] = coord_from_row(each_row) - 1
                        
                    if last_element_was_a_face:
                        self.int_face[int_face_id-1]['end'] = coord_from_row(each_row) - 1                    
                    
                    last_element_was_a_face = True 
                    last_element_was_a_hole = False
                    
                else:
                    hole_id += 1 
                    
                    self.hole[hole_id] = {'elem_row': each_row,
                                          'parent'  : (5, face_id),
                                          'start'   : coord_from_row(each_row)}
                    
                    self.face[face_id]['holes'].append(hole_id)
                    
                    self.element.append((3,hole_id))
                    
                    if last_element_was_a_hole:
                        self.hole[hole_id-1]['end'] = coord_from_row(each_row) - 1
                        
                    if last_element_was_a_face:
                        self.face[face_id]['end'] = coord_from_row(each_row) - 1

                    last_element_was_a_face = False
                    last_element_was_a_hole = True                    
                    
        # Having parsed each element provide 'end' to the last elements
        
        last_coord = np.shape(self.ordinates)[0]
            
        self.collection[collection_id]['end'] = last_coord
        
        if last_element_belonged_to_composite:
            self.composite[composite_id]['end'] = last_coord
        
        self.solid[solid_id]['end'] = last_coord

            
        if last_element_was_a_hole:
            if intsurf:
                self.int_hole[int_hole_id]['end'] = last_coord
                self.void[void_id]['end'] = last_coord
            else:
                self.hole[hole_id]['end'] = last_coord
                self.surf[surf_id]['end'] = last_coord
                
        elif last_element_was_a_face:
            if intsurf:
                self.int_face[int_face_id]['end'] = last_coord
                self.void[void_id]['end'] = last_coord
            else:
                self.face[face_id]['end'] = last_coord
                self.surf[surf_id]['end'] = last_coord       
            
        
        
        def collect(geomdict):
            '''Function for returning list of elements by element row
            Used for legacy functions
            '''
            keys = geomdict.keys()
            ret = []
            for each in keys:
                if each:
                    ret.append(geomdict[each]['elem_row'])
            return ret
                    
        self.collections = collect(self.collection)
        self.composites  = collect(self.composite)
        self.solids      = collect(self.solid) 
        self.ext_surf    = collect(self.surf)     
        self.int_surf    = collect(self.void)
        self.ext_ring    = collect(self.face)
        self.int_ring    = collect(self.int_face)
        self.ext_holes   = collect(self.hole)
        self.int_holes   = collect(self.int_hole)

    def pprint(self):
        
        '''Method for printing out elements for checking and debugging'''
        
        print ("Element Array :\n%s"%pd.DataFrame(self.elem_array))
        print ("Ordinate Array :\n%s"%pd.DataFrame(self.ordinates))
        
        for each in self.collection.keys():
            print ("Collection: %s"%each)
            pprint (self.collection[each])
            print("")
            for each_solid in self.collection[each]['solids']:
                print ("Solid: %s"%each_solid)
                pprint (self.solid[each_solid])
                print("")
                
                for each_surface in self.solid[each_solid]['surfaces']:
                    print ("Surface: %s"%each_surface)
                    pprint (self.surf[each_surface])
                    
                    for each_face in self.surf[each_surface]['faces']:
                        print ("Faces : %s"%each_face)
                        pprint (self.face[each_face])
                    
                        print (self.get_ords(self.face[each_face]['start'],
                                             self.face[each_face]['end']+1))
                        
                        for each_hole in self.face[each_face]['holes']:
                            print ("Holes: %s"%each_hole)
                            pprint (self.hole[each_hole])
                            
                            print (self.get_ords(self.hole[each_hole]['start'],
                                                 self.hole[each_hole]['end']+1))                            
                    
                for each_void in self.solid[each_solid]['voids']:
                    print ("Void: %s"%each_void)
                    pprint (self.void[each_void])
                    
                    for each_face in self.void[each_void]['faces']:
                        print ("Faces : %s"%each_face)
                        pprint (self.face[each_face])
                    
                        print (self.get_ords(self.face[each_face]['start'],
                                             self.face[each_face]['end']+1))
                        
                        for each_hole in self.face[each_face]['holes']:
                            print ("Holes: %s"%each_hole)
                            pprint (self.hole[each_hole])
                            
                            print (self.get_ords(self.hole[each_hole]['start'],
                                                 self.hole[each_hole]['end']+1)) 
            
            for each_composite in self.collection[each]['composites']:
                print ("Composite: %s"%each_composite)
                pprint (self.composite[each_composite]) 
                
                for each_solid in self.composite[each]['solids']:
                    print ("Solid: %s"%each_solid)
                    pprint (self.solid[each_solid])
                    print("")
                    
                    for each_surface in self.solid[each_solid]['surfaces']:
                        print ("Surface: %s"%each_surface)
                        pprint (self.surf[each_surface])
                        
                        for each_face in self.surf[each_surface]['faces']:
                            print ("Faces : %s"%each_face)
                            pprint (self.face[each_face])
                        
                            print (self.get_ords(self.face[each_face]['start'],
                                                 self.face[each_face]['end']+1))
                            
                            for each_hole in self.face[each_face]['holes']:
                                print ("Holes: %s"%each_hole)
                                pprint (self.hole[each_hole])
                                
                                print (self.get_ords(self.hole[each_hole]['start'],
                                                     self.hole[each_hole]['end']+1))                            
                        
                    for each_void in self.solid[each_solid]['voids']:
                        print ("Void: %s"%each_void)   
                        
                        pprint (self.void[each_void])
                        
                        for each_face in self.void[each_void]['faces']:
                            print ("Faces : %s"%each_face)
                            pprint (self.face[each_face])
                        
                            print (self.get_ords(self.int_face[each_face]['start'],
                                                 self.int_by_id[each_face]['end']+1))
                            
                            for each_hole in self.int_face[each_face]['holes']:
                                print ("Holes: %s"%each_hole)
                                pprint (self.int_hole[each_hole])
                                
                                print (self.get_ords(self.int_hole[each_hole]['start'],
                                                     self.int_hole[each_hole]['end']+1)) 
    
    def get_ords(self, s, e):
        '''Method for returning coords from ordinate array using start and end'''
        return self.ordinates[s:e]
    
class SDO_Geometry (object):
    '''Class for holding SDO_GEOMETRY object either by parsing Element Information
    Array or by adding Geometry objects
    '''
    def __init__(self,
                 geom = None,
                 gtype =3001,
                 srid = 7405,
                 pt = None,
                 elem_info=None,
                 ords = None,
                 debug=False):
        
        self.srid = srid
        self.alt_pt = False
        self.debug = debug
        
        if type(geom) == type(None):    
            
            self.pt = pt
            self.elem_info = elem_info
            self.ords = ords
            self.ordinates=np.array([], float)
                    
            if elem_info and gtype != 3001:
                self.elem_array = np.array(elem_info, int).reshape(int(len(elem_info)/3.0),3)
            if ords and gtype != 3001:
                self.ordinates = np.array(ords, float).reshape(int(len(ords)/3.0),3)  
                
            if gtype not in [3001,3002]:
                self.element = SdoElemInfo(gtype, self.elem_array, self.ordinates)
                
            if gtype == 3001 and pt: # Point
                self.geom = Point(*pt)
            elif gtype == 3001 and elem_info:  # Point
                self.geom = Point(ords[0],ords[1],ords[2])
            elif gtype == 3002: # LineString
                self.geom = LineString(self.ordinates)
            elif gtype == 3003 and elem_info[1] == 3 and elem_info[2] == 1 : # Simple Polygon 
                self.geom = Polygon(PolygonRing(self.ordinates), simple = True)
            elif gtype == 3003 and elem_info[1] == 3 and elem_info[2] == 3 : # Simple Polygon - rectangle
                self.geom = Polygon(PolygonRing(self.ordinates, corners = True), simple = True)               
            elif gtype == 3003 and elem_info[1] == 1003 : # Polygon 
                face_id = self.element.surf[0]['faces'][0]                                   
                self.geom = self.polygon_build(face_id) 
                
            elif gtype == 3003 and elem_info[1] == 1006: # Composite Polygon
                
                surf_id = self.element.solid[0]['surfaces'][1]  
                # Note: The default solid contains two surfaces - the default
                #       surface [0] and the described surface [1]
                self.geom = self.surface_build(surf_id)  
                
            elif gtype == 3008 and elem_info[1] == 1007 and elem_info[2] == 1: # Simple Solid
                solid_id = self.element.collection[0]['solids'][0]
                self.geom = self.solid_build(solid_id)
            elif gtype == 3008 and elem_info[1] == 1007 and elem_info[2] == 3: # Simple Solid - cube
                raise Exception ("Functionality does not yet exist to parse Cuboid Solid")
            elif gtype == 3008 and elem_info[1] == 1008: # Composite Solid
                raise Exception ("Functionality does not yet exist to parse Composite Solids")
            elif gtype == 3009: # MultiSolid
                raise Exception ("Functionality does not yet exist to parse Multi Solids")
            elif gtype == 3004: # Heterogenous Collection
                raise Exception ("Functionality does not yet exist to parse Heterogenous Collections")
            elif gtype == 3005: # MultiPoint
                raise Exception ("Functionality does not yet exist to parse Multi Points")
            elif gtype == 3006: # MultiLine
                raise Exception ("Functionality does not yet exist to parse Multi Lines")
            elif gtype == 3007: # MultiSurface 
                surf_id = 0
                self.geom = self.surface_build(surf_id)
    
            self.gtype = gtype
            
        else:
            self.gtype = geom.gtype   
            self.geom = geom
            
    def polygon_build(self, face_id, interior = False):
        
        if interior:
            face = self.element.int_face[face_id]
        else:
            face = self.element.face[face_id]
        coords = self.element.get_ords(face['start'],face['end']+1)
        
        #If only two coordinates assume that [1, 1003, 3] element is being read
        corners = (np.shape(coords)[0] ==2)
        p = Polygon(PolygonRing(coords, corners=corners),
                    debug=self.debug)
        
        for each in face['holes']:
            if interior:
                hole = self.element.int_hole[each]
            else:    
                hole = self.element.hole[each]
            coords = self.element.get_ords(hole['start'],hole['end']+1)
            #If only two coordinates assume that [1, 2003, 3] element is being handle
            corners = (np.shape(coords)[0] ==2)
            p.addHole([PolygonRing(coords, corners = corners, interior = True)])
        
        return p
    
    def surface_build(self, surf_id, interior= False):
        
        if interior:
            surface = self.element.void[surf_id]
        else:
            surface = self.element.surf[surf_id]
        
        polygons = []
        poly_list = surface['faces']
        
        for each_face in poly_list:
            
            
            polygons.append(self.polygon_build(each_face, interior))

        return Surface(polygons,
                       interior=interior,
                       debug=self.debug)
    
    def solid_build(self, solid_id):
        
        solid = self.element.solid[solid_id]
        void_list = solid['voids']
        s = SimpleSolid(self.surface_build(solid['surfaces'][0]),
                        debug=self.debug)
        
        for each_void in void_list:
            to_add = [self.surface_build(each_void, interior = True)]
            s.addIntSurf(to_add)
            
        return s

    def setAltPt(self, boolean):
        
        self.alt_pt = boolean
    
    @property
    def SDO_GEOMETRY(self):
        
        if self.gtype == 3001 and not self.alt_pt:
            return "SDO_GEOMETRY (%s, %s, SDO_POINT_TYPE %s, NULL, NULL)"%(self.gtype,
                                                                           self.srid,
                                                                           str(self.geom.xyz))          
        else:            
            return "SDO_GEOMETRY (%s, %s, NULL, SDO_ELEM_INFO_ARRAY %s, SDO_ORDINATE_ARRAY %s)"%\
                    (self.gtype, self.srid, tuple(self.geom.elem_info(True)),
                                            tuple(self.geom.ord_info(True)))
            
    def __repr__(self):
        
        return "SDO_GEOMETRY %s"%(self.gtype)
    
        
class Point(eu.Point3):
    
    '''Point class inheritting from Euclid Point3 class'''

    def __init__(self, *args, **kwargs):
        
        '''Arguments:
        
        x :
        y :
        z :
        
        '''
        
        super(eu.Point3,self).__init__(*args,**kwargs) 
        
    def set_point(self, pt):
        
        self.x = pt.x
        self.y = pt.y
        self.z = pt.z
        
    @property
    def xyz(self):
        return (self.x,self.y,self.z)
   
    
    def elem_info(self, full = False):
        return [1,1,1]
    
    def ord_info(self, full = False):
        return [self.x,self.y,self.z]
    
    @property
    def gtype(self):
        return 3001 
    
    
class Vertices(object):
    
    def __init__(self, initPoints = None, **kwargs):
        
        '''initPoints (np.array) : 
        
        '''
        if not isinstance(initPoints, type(None)):
            self._points = initPoints.copy()
        else:
            self._points = np.array([], float)
        
    
    def set_points(self, pts):
        '''pts (np.array) : 
                
        '''        
        self._points = cp.deepcopy(pts) 
     
    @property
    def num_points(self):
        return np.shape(self._points)[0]

    # Returns list of Point3 objects as per list
    @property   
    def stored_points(self):
        ''' returns np array
        '''
        return self._points
    

    @property   
    def xcoords(self):                    
        '''Returns x coords as list
        '''
        return self._points[:,0]
    
    @property   
    def ycoords(self): 
        '''Returns y coords as list
        '''
        return self._points[:,1]  

    @property   
    def zcoords(self):
        '''Returns list of z coordinates'''
        return self._points[:,2]     
        
    # Clears all the points
    def clear_points(self):                   
        self._points = np.array([], float)

    def __repr__(self):
        return str(self.__class__).split(".")[1][:-2]+" of "+\
            str(self.num_points)+" Points"

class LineSegment(eu.LineSegment3):
    
    """
    Identical to LineSegment3 with provision for returning midpoint of a line
    """
    
    @property
    def mid_point(self):
        return eu.Point3(self.p.x + (self.v.x/2),
                         self.p.y + (self.v.y/2),
                         self.p.z + (self.v.z/2))


class LineString(Vertices):

    """
    Class of 2 coordinate line objects
    """
    
    def __init__(self, *args, **kwargs):
        super(LineString, self).__init__(*args, **kwargs)
        self.gtype = 3002
        
    def ord_info(self, full):
        return list(self._points.ravel())


    def elem_info(self, full):
        return [1,2,1]

def make_face(points):

    coords = [gp.gp_Pnt(each[0],each[1],each[2]) for each in points]               
    if len(coords) < 4:
        return None
    
    poly = brep.BRepBuilderAPI_MakePolygon()
    for each_pt in coords[:-1]:
        poly.Add(each_pt)
    poly.Close()
    face = brep.BRepBuilderAPI_MakeFace(poly.Wire())
    if face.IsDone():
        return face.Face()
    else:
        return None
        



class PolygonRing(Vertices):

    """
    Class of 2 coordinate line objects
    """
    
    def __init__(self, *args, **kwargs):
        super(PolygonRing, self).__init__(*args, **kwargs)
        self.interior = kwargs.get('interior', False )
        
        self.gtype = 0
        self._ord_info= cp.deepcopy(list(self._points.ravel()))
        self._elem_info= [1,1003,1]
        self.corners = kwargs.get('corners', False)
        if self.corners:
            self.calculate()
    
    def area(self):
        pass
    
    @property        
    def plot_coords(self):                       
        coords = []               
        if self.num_points != 0:
            for each in self._points[:,]:     
                coords.append((each[0],each[1],each[2]))
        return [coords]

    def set_colour(self, colour):
        
        self.colour = colour
    
    @property        
    def occ_coords(self):                       
        coords = []               
        if self.num_points not in [0,1,2]:
            for each in self._points[:,]:     
                coords.append(gp.gp_Pnt(each[0],each[1],each[2]))
        return coords
            
    # Plots polygon   
    def plot(self, mplot): 
        
        def min_none(a,b):
            if b == None:
                return a
            else:
                return min(a,b)

        def max_none(a,b):
            if b == None:
                return a
            else:
                return max(a,b)
            
        min_x,min_y,min_z = self._points.min(axis = 0)
            
        mplot.min_b[0] = min_none(min_x,mplot.min_b[0])    
        mplot.min_b[1] = min_none(min_y,mplot.min_b[1])
        mplot.min_b[2] = min_none(min_z,mplot.min_b[2]) 
        
        max_x,max_y,max_z = self._points.max(axis = 0)
        
        mplot.max_b[0] = max_none(max_x,mplot.max_b[0]) 
        mplot.max_b[1] = max_none(max_y,mplot.max_b[1])
        mplot.max_b[2] = max_none(max_z,mplot.max_b[2])         
        
        poly_ext  = Poly3DCollection(self.plot_coords, facecolors = [mplot.cc('r',0)])
        poly_fill = Poly3DCollection(self.plot_coords, facecolors = [mplot.cc('r',0.2)], linewidths = (0,))
        
        mplot.ax.add_collection3d(poly_ext)
        mplot.ax.add_collection3d(poly_fill)
        
    @property
    def occ_wire(self):
        
        mkWire = brep.BRepBuilderAPI_MakeWire()
    
        for n in range(len(self.occ_coords[:-1])):
            segment = gc.GC_MakeSegment(self.occ_coords[n],self.occ_coords[n+1])
            edge = brep.BRepBuilderAPI_MakeEdge(segment.Value())
            mkWire.Add(edge.Edge())
    
        return mkWire.Wire()      
    
    @property
    def occ_face(self):
        
        mkWire = brep.BRepBuilderAPI_MakeWire()
        
        if len(self.occ_coords) == 0:
            return None

        for n in range(len(self.occ_coords[:-1])):
            segment = gc.GC_MakeSegment(self.occ_coords[n],self.occ_coords[n+1])
            edge = brep.BRepBuilderAPI_MakeEdge(segment.Value())
            mkWire.Add(edge.Edge())
            
        face = brep.BRepBuilderAPI_MakeFace(mkWire.Wire()) 
    
        return face 
    
    def plotocc(self, display):
        
        topods_shape = self.occ_face.Face()
        
        display.DisplayShape(topods_shape, color = self.colour)
        
        return topods_shape
    
    def calculate(self):
        
        '''Method for adding missing corners in a 2 coordinate defined polygon'''
        
        points = self._points.copy()
        
        yz = points[0,0] == points[1,0]
        zx =  points[0,1] == points[1,1]
        xy = points[0,2] == points[1,2]
            
        if yz+zx+xy != 1:
            raise Exception ("Opposite corners must lie on orthogonal plane" )

        first_point = points[0]        
        third_point = points[1]
        
        if xy:
            z = points[0,2]
            second_point = np.array([points[0,0], points[1,1], z])
            fourth_point = np.array([points[1,0], points[0,1], z])
        if yz:
            x = points[0,0]
            second_point = np.array([x, points[1,1], points[0,2]])
            fourth_point = np.array([x, points[0,1], points[1,2]])            
        if zx:
            y = points[0,1]
            second_point = np.array([points[0,0], y, points[1,2]])
            fourth_point = np.array([points[1,0], y, points[0,2]])            
        
        self._points = np.array([first_point,second_point,third_point,fourth_point, first_point])
                         
        
    
    def elem_info(self, full=False):
        if self.corners or not full:
            return [1,1003,3]
        else:
            return [1,1003,1] 
 
    
    def ord_info(self, full =False):
        
        if not self.corners or not full:
            return list(self._points.ravel()) # Copy not required because list( .ravel()) makes copy 
        else:
            return list(self._points[0].copy())+list(self._points[2].copy())
            
class Polygon(object):
    
    def __init__(self, *args, **kwargs):
            
        self.gtype = 3003    
        self.exterior = cp.deepcopy(args[0])
        self.holes = kwargs.get("holes", [])
        self.simple = kwargs.get('simple', False )
        self.debug = kwargs.get("debug", False)
        self.calculate() 
        
    def plot(self, mplot):
        
        self.exterior.plot(mplot)
        
    def plotocc(self, display):
        
        topods_shape = self.exterior.plotocc(display)  
    
        return topods_shape
    
    def set_colour(self, colour):
        
        self.colour = colour    
    
    def addHole(self, hole_list):
        if not self.simple:
            self.holes.extend(cp.deepcopy(hole_list))
            self.calculate()    
            
    def calculate(self):

        SOA= self.exterior.ord_info()
        
        if self.simple:
            SEIA = [1, 3, self.exterior.elem_info(True)[2]]
            
        else:
            SEIA= [1,1003, self.exterior.elem_info(True)[2]]    
            counter = 1
            counter += len(SOA)
            for each in self.holes:
                SEIA.extend([counter,2003,each.elem_info(True)[2]])
                SOA.extend(each.ord_info(True))
                counter += len(each.ord_info(True))
                
        self._ord_info= cp.deepcopy(SOA)
        self._elem_info= cp.deepcopy(SEIA)
        

    def ord_info(self, full = False):
        return self._ord_info
 

    def elem_info(self, full = False):
        return self._elem_info
    
    @property
    def simple_elem_info(self):
        if self.simple:
            return [1,3,1]
        else:
            return [1,1003,1]
        
    def wires(self):
        
        w = [self.exterior.occ_wire]
        if not self.simple:
            for each in self.holes:
                w.extend(each.occ_wire)
        return w
    
    @property
    def multi_surface(self):
        
        compound = tp.TopoDS_Compound()
        builder = BRep_Builder()
        builder.MakeCompound(compound)         
        
        # print("Py3DGeom - %s" % 1, flush=True)
        
        if len(self.holes) > 0:
            print("Py3DGeom - Holes here", flush=True) 
            
            
        face = self.exterior.occ_face

        builder.Add(compound, face.Face())          
        
        return compound


def remove_consecs(array):
    
    first = array[:-1,:] 
    last = array[1:,:]
    diff = first - last
    rowsum = diff.sum(1)
    rowsum = np.abs(rowsum)
    rowsum[rowsum<0.00000000001] = 0
    rowsum = rowsum.astype(np.bool)
    toberemoved = ~rowsum
    
    to_be_removed = toberemoved.nonzero()[0]
    # print(to_be_removed)
    # print(len(to_be_removed))
    # print(len(rowsum) - len(to_be_removed))
    if len(to_be_removed) == 0:
        return "good", None 
    if (len(rowsum) - len(to_be_removed)) < 3:
        return "bad", None
    
    removed = np.delete(array,
                        to_be_removed,
                        0)
    
    return "fixed", removed
    
    

class Surface(object):
    
    def __init__(self, *args, **kwargs):
        '''Geometry class for OGC Surface
        Arguments:
        polygons(list) : list of Polygon objects
        interior (bool): make interior surface (True)
        
        
        '''
            
        self.polygons = cp.deepcopy(args[0])
        if  kwargs.get('interior', False):
            self.interior = []
        else:
            
            self.interior = kwargs.get('interior', False)
        self.debug = kwargs.get("debug", False)
        self.gtype = 3003
        self.tol = 0.000001
        self.calculate()
        
    def set_colour(self, rgb = (0,0,1)):
        
        r,g,b = rgb
        self.colour = qu.Quantity_Color(r,g,b,0)
        
    def plot(self, mplot):
    
        for each in self.polygons:
            each.plot(mplot)
    
    @property        
    def occ_shell(self):
        
        mkShell = brep.BRepBuilderAPI_Sewing(self.tol)
        
        for each in self.polygons:
            
            mkShell.Add(each.exterior.occ_face.Face())
            
        mkShell.Perform()
        
        sew = mkShell.SewedShape()
        #sew.Reverse()
    
        shell = tp.topods.Shell(sew)  
        
        return shell
    
    @property        
    def multi_surface(self):
            
        compound = tp.TopoDS_Compound()
        builder = BRep_Builder()
        builder.MakeCompound(compound) 
        
        for n, each_face in enumerate(self.polygons):
            if False and n in [346, 376, 383, 209, 2133, 2912, 2942, 2977]:
                continue
            if self.debug:
                print("Py3DGeom - %s" % n, flush=True)
                
                # t = each_face.exterior._points[:-1]
                        
                stat, new_arr = remove_consecs(each_face.exterior._points)
                
                if stat=="good":
                    coords = each_face.exterior._points
                elif stat == "bad":
                    print("Py3DGeom - %s" % n, flush=True)
                    print("bad")
                    continue
                else:
                    print("Py3DGeom - %s" % n, flush=True)
                    print("fixing")
                    coords = new_arr
                
                face = make_face(coords)
                
                if face is None:
                    continue                    
            else:        
                face = each_face.exterior.occ_face.Face()
          
            if len(each_face.holes) > 0:
                if self.debug:
                    print("Py3DGeom - %s" % n, flush=True)
                    print("Py3DGeom - Hole" , flush=True)                  
                for each_hole in each_face.holes:
                    cutter = BRepAlgoAPI_Cut(face, each_hole.occ_face.Face())
                    cutter.Build()
                    shape = cutter.Shape()
                    face = list(Topo(shape).faces())[0]
                    
            if face is None:
                continue
            
            builder.Add(compound, face)        
              
        return compound
            
    def plotocc(self, display):
        
        topods_shape = self.shape()
        ais_shape = ais.AIS_Shape(topods_shape)
        ais_shape.SetColor(self.colour)
        display.Context.Display(ais_shape.GetHandle())
        
        return topods_shape
    
    def shape(self, tol = 0.000001):
        self.tol = tol
        return self.occ_shell
    
        
    def addPolygon(self, polygon_list):
        self.polygons.extend(cp.deepcopy(polygon_list))
        self.calculate()
        
    def calculate(self):

        SOA= []
        SEIA= [1,1006,len(self.polygons)]
        
        count = 0
        for each in self.polygons:
            
            #print "lenght SOA", len(each.SDO[1])
            SEIA_info = cp.deepcopy(each.elem_info(True))
            for n in range(len(SEIA_info)):
                #print "n", n
                if not n%3:
                    SEIA_info[n] = each.elem_info(True)[n] + count
                        
            count += len(each.ord_info(True))
            
            SEIA.extend(SEIA_info)
            SOA.extend(each.ord_info(True))
            
        self._ord_info= cp.deepcopy(SOA)
        self._elem_info= cp.deepcopy(SEIA)    


    def elem_info(self, full = False):
        return self._elem_info
    
    def ord_info(self, full = False):
        return self._ord_info
    
    def wires(self):
        
        w = []
        for each in self.polygons:
            w.extend(each.wires())
        
        return w
        
        
    
    

class SimpleSolid(object):
    
    def __init__(self, *args, **kwargs):
        '''Geometry class for OGC Simple Solid
        
        Arguments:
        ext_surf       : Surface object
        int_surf(list) : list of interior surfaces
        
        
        '''
            
        self.ext_surf = cp.deepcopy(args[0])
        self.int_surf = cp.deepcopy(kwargs.get('int_surf', []))
        self.debug = kwargs.get("debug", False)
        self.gtype = 3008
        self.calculate()
        
    def plot(self, mplot):

        self.ext_surf.plot(mplot)
        
    def set_colour(self, rgb = (0,0,1)):
        
        r,g,b = rgb
        self.colour = qu.Quantity_Color(r,g,b,0)
        
    @property
    def occ_solid(self):
        
        return brep.BRepBuilderAPI_MakeSolid(self.ext_surf.occ_shell)
    
    @property
    def multi_surface(self):
        
        return self.ext_surf.multi_surface
        
    def plotocc(self, display):

        #self.ext_surf.plotocc(display)
        topods_shape = self.occ_solid.Solid()
        display.DisplayShape(topods_shape, color = self.colour )

        return topods_shape
    
    def addIntSurf(self, int_surf):
        
        '''Arguments:
        
        int_surf (list) : list of Surface objects
        
        '''    

        self.int_surf.extend(cp.deepcopy(int_surf))

        self.calculate()
        
    def calculate(self):

        SOA= []
        SEIA= [1,1007,1]
        
        count = 0
        
        SEIA_info = cp.deepcopy(self.ext_surf.elem_info(True))
        
        for n in range(len(SEIA_info)):

            if not n%3:
                SEIA_info[n] = self.ext_surf.elem_info(True)[n] + count
                    
        count += len(self.ext_surf.ord_info(True))
        
        SEIA.extend(SEIA_info)
        SOA.extend(self.ext_surf.ord_info(True)) 
        
        
        for each in self.int_surf:
            
            #print "lenght SOA", len(each.SDO[1])
            SEIA_info = cp.deepcopy(each.elem_info(True))
            
            for n in range(len(SEIA_info)):

                if not n%3:
                    SEIA_info[n] = each.elem_info(True)[n] + count
                elif n%3 ==1:
                    
                    if SEIA_info[n] == 1003:
                        SEIA_info[n] = 2003
                    elif SEIA_info[n] == 2003:
                        SEIA_info[n] = 1003                        
                    elif SEIA_info[n] == 1006:
                        SEIA_info[n] = 2006
                        
            count += len(each.ord_info(True))
            
            SEIA.extend(SEIA_info)
            SOA.extend(each.ord_info(True))
            
            
            
        self._ord_info= cp.deepcopy(SOA)
        self._elem_info= cp.deepcopy(SEIA)    


    def elem_info(self, full= False):
        return self._elem_info
    
    def ord_info(self, full = False):
        return self._ord_info
    
    def shape(self, tol = 0.000001):
        
        return self.ext_surf.shape(tol)
    
    def wires(self):
        
        w = self.ext_surf.wires()
        for each in self.int_surf:
            w.extend(each.wires())

        return w

class Plot_mlab(object):
    
    def __init__(self):
        
        fig = plt.figure()
        self.ax = Axes3D(fig)
        self.cc = lambda arg,alpha = 0.7: colorConverter.to_rgba(arg, alpha)
        
        self.ax.set_xlabel('X Label')
        self.ax.set_ylabel('Y Label')
        self.ax.set_zlabel('Z Label') 
        
        self.min_b = [None,None,None]
        self.max_b = [None,None,None]
        
    def setLimits(self):
    
    
        xRange = self.max_b[0] - self.min_b[0]
        yRange = self.max_b[1] - self.min_b[1]
        zRange = self.max_b[2] - self.min_b[2]
        maxRange = max([xRange,yRange,zRange])
        self.ax.set_xlim3d(self.min_b[0],self.min_b[0]+maxRange)
        self.ax.set_ylim3d(self.min_b[1],self.min_b[1]+maxRange)
        self.ax.set_zlim3d(self.min_b[2],self.min_b[2]+maxRange)

      
