# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

# Python library imports

from pathlib import Path
import time

# Python third party library imports

import numpy as np
from skimage.morphology import binary_dilation, watershed

# Local imports

from modules.class_progress import Progress
from modules.class_scene import Scene
from modules.class_group import VoxelArrayGroup
from modules.class_grid import VoxelGrid

from modules.form_solid_from_voxels import solidify_rooms, minimize, bigger
from modules.adjacency import determine_adjaceny, identify_disconnected_
from modules.thirtytwo_bit import t32_bit
from modules.filing import create_folder
from modules.display import displayer, apply_colour
from modules.display_room import color_pallete
from modules.disp_illust_modules import view_illust
import modules.display_q_voxels as make_solid

# Local cython imports

import neighbour_ski_ig as neighbour
from topology import convertTopology
import Py3DGeom as Py3DGeom
import check_manifold

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)
log_debug = (log.getEffectiveLevel() <= 10)

t = time.time


class Timings(object):

    def __init__(self):

        self.mask = 0
        self.voxelise = 0
        self.dft = 0
        self.source_markers = 0
        self.watershed = 0
        self.disconnected = 0
        self.total = 0

    def asdict(self):

        dict_ = {"mask": self.mask,
                 "voxelise": self.voxelise,
                 "dft": self.dft,
                 "source_markers": self.source_markers,
                 "watershed": self.watershed,
                 "disconnected": self.disconnected,
                 "total": self.total,
                 }

        return dict_


'''------------------------------------------------------------------------'''


def main(setting, environment, test=False):

    timings = Timings()

    '''Loading settings'''

    home_folder = Path.home()

    if environment == "macbook":
        pickles = home_folder / "Dropbox" / "PhD copy" / "Filing"

    setting.progress_update.update({})

    progress = Progress(setting.progress,
                        setting.display,
                        update=setting.progress_update,
                        display_update=setting.display_update,
                        )

    progress.pickles_path = pickles

    log.info("Scene setting imported: %s " % setting.scene)
    log.info("Grid setting imported: %s " % setting.grid)
    log.info("Display setting imported: %s " % setting.display)
    log.info("Progress setting imported: %s " % setting.progress)

    '''--------------------------------------------------------------------'''
    if progress.s.load is True:

        year, month, day = progress.s.load_folder.split("-")
        folder = (progress.pickles_path
                  / (year + "-" + month) / progress.s.load_folder)
        path = folder / progress.s.load_file
        log.info("Loading from %s - %s"
                 % (progress.s.load_folder, progress.s.load_file))

        progress.load(path)
        progress.loaded = True
        progress.scene_loaded = True
        progress.vox_grid_set_up = True
        progress.vox_group_set_up = True

    if progress.scene_loaded and not progress.s.make_scene:
        log.info("Loading Scene")
        scene = Scene(load=path)
        progress.voxelised = True
    else:
        scene = Scene(setting.scene, update=setting.scene_update)
        scene.load_scene_elements()
        progress.vox_grid_set_up = False
        progress.vox_group_set_up = False
        progress.loaded = False

    if progress.vox_grid_set_up and not progress.s.make_vox_grid:
        log.info("Loading Vox Grid")
        vox_grid = VoxelGrid(load=path)
    else:
        log.info("Making Vox Grid")
        vox_grid = VoxelGrid(setting.grid,
                             offset=scene.s.offset,
                             update=setting.grid_update,
                             scene=scene)

        progress.vox_group_set_up = False
        progress.loaded = False
        progress.voxelised = False

    if progress.vox_group_set_up and not progress.s.make_vox_group:
        log.info("Loading Vox Group")
        vox_group = VoxelArrayGroup(vox_grid, progress, path)

    else:
        log.info("Making Vox Group")
        vox_group = VoxelArrayGroup(vox_grid, progress)
        progress.loaded = False
        progress.voxelised = False

    if progress.loaded is False:
        if progress.s.voxelise:
            log.info('''Generating voxel scene elements''')
            scene.generate_voxel_scene(vox_grid)
        else:
            scene.vox_grid = vox_grid

        '''----------------------------------------------------------------'''
        log.info('''Indexing scene elements''')
        ts = t()
        scene.index_elements()
        if progress.s.voxelise:
            scene.vox_index_elements()
        scene.bounding_box()
        log.info("time to index: %s" % (t() - ts))

        '''----------------------------------------------------------------'''
        if progress.s.make_hull:
            log.info('''Making hull''')
            ts = t()
            scene.gen_hull(progress.s.voxelise)
            time_to_make_hull = (t() - ts)
            log.info("time to make hull: %s" % time_to_make_hull)

        '''----------------------------------------------------------------'''
        if progress.s.voxelise:
            log.info('''Making mask''')
            ts = t()
            vox_group.generate_mask(scene, progress.s.hull_buffer)
            time_to_make_mask = (t() - ts)
            log.info("time to make mask: %s" % time_to_make_mask)
            timings.mask = round(time_to_make_hull + time_to_make_mask, 2)

        '''----------------------------------------------------------------'''
        if progress.s.voxelise:
            log.info('''Voxelising''')

            ts = t()
            vox_group.perform_voxelise_trimesh(scene)

            timings.voxelise = round(t() - ts, 2)
            log.info("time to voxelise: %s" % timings.voxelise)
            progress.voxelised = True

        '''----------------------------------------------------------------'''
        if progress.voxelised:

            ts1 = t()

            vox_group.array2 = vox_group.array.copy()
            vox_group.array2[vox_group.mask] = 1

            '''------------------------------------------------------------'''

            # perform_watershed_barrier(vox_group,
            # scene,
            # progress,
            # )

            '''-------------------------------------------------------------'''
            log.info("Performing DFT")

            ts = t()
            vox_group.perform_dft(dft_factor=progress.s.dft_factor)
            timings.dft = round(t() - ts, 2)

            log.info('time taken dft%s' % timings.dft)

            '''-------------------------------------------------------------'''
            log.debug("Connecting Up and Identifying source markers")

            ts = t()
            vox_group.sources = {}

            mask = vox_group.mask.astype(t32_bit())

            no_barriers = bool(progress.s.no_barriers)

            neighbour.connect_up(vox_group.n,
                                 vox_group.sources,
                                 vox_group.dft,
                                 vox_group.raw_dft,
                                 mask,
                                 log_debug,
                                 vox_group.start,
                                 no_barriers,
                                 use_only_6=not progress.s.adjacency_26)

            timings.source_markers = round(t() - ts, 2)
            log.info("time to identify sources: %s" % timings.source_markers)

            '''------------------------------------------------------------'''
            log.info("Performing watershed")
            ts = t()

            markers = np.full_like(vox_group.dft, 0)

            for each in vox_group.sources:
                log.debug("Marker %s" % each)
                x, y, z = vox_group.sources[each]['pos']
                log.debug("%s, %s, %s" % (x, y, z))
                markers[x, y, z] = each

            vox_group.connect = watershed(-vox_group.raw_dft,
                                          markers,
                                          mask=~(vox_group.array
                                                 + vox_group.mask),
                                          watershed_line=False)
            timings.watershed = round((t() - ts), 2)
            log.info("time to perform watershed: %s" % timings.watershed)

            '''------------------------------------------------------------'''
            if progress.s.remove_buffer:
                log.info("Removing dilation/buffer")
                ts = t()
                vox_group.connect[vox_group.buffer] = 0
                log.debug("time to remove buffer: %s" % (t() - ts))

            '''------------------------------------------------------------'''
            log.debug("Identifying disconected")
            ts = t()
            identify_disconnected_(vox_group)
            timings.disconnected = round(t() - ts, 2)
            log.info("time to find disconnected: %s" % timings.disconnected)

            if progress.s.determine_internals:
                ts = t()
                scene.internals = determine_adjaceny(vox_group, scene)
                log.info("time to identify internals: %s" % (t() - ts))
                log.info(scene.internals)
            else:
                scene.internals = list(vox_group.sources)

            '''------------------------------------------------------------'''
            log.info("Total time to identify spaces : %s" % (t() - ts1))
            progress.barrier_built = True
            progress.dft = True

    else:
        '''----------------------------------------------------------------'''
        log.info('''Indexing scene elements''')
        ts = t()
        scene.index_elements()
        scene.vox_grid = vox_grid
        scene.bounding_box()
        log.info("time to index: %s" % (t() - ts))

    '''--------------------------------------------------------------------'''
    if progress.voxelised and progress.s.merge:
        log.info("Merging rooms")

        vox_group.sources = {}
        scene.room_hash = {}
        scene.rooms = {}

        next_room = vox_group.connect.max()
        lowest = vox_group.connect.max() + 1
        rooms = {}
        for each_room in progress.s.merge:
            next_room += 1
            for every in progress.s.merge[each_room]:
                rooms[every] = next_room
                vox_group.sources[next_room] = {'internal': True}

        for each_rm in rooms:
            vox_group.connect[vox_group.connect == each_rm] = rooms[each_rm]
        vox_group.connect[vox_group.connect < lowest] = 0

    if progress.voxelised and progress.s.make_solid:
        log.info('Making Solid Rooms')
        room, room_hash = solidify_rooms(vox_group, scene, faces=False)
        scene.rooms.update(room)
        scene.room_hash.update(room_hash)

    log.info('Number of rooms: %s' % (len(scene.rooms)))

    try:
        scene.internals
    except AttributeError:
        log.info("Reinstating empty internal rooms list")
        scene.internals = []
    try:
        scene.externals
    except AttributeError:
        log.info("Reinstating empty external rooms list")
        scene.externals = []

    '''--------------------------------------------------------------------'''
    if progress.s.make_solid:
        log.info('Colourising Rooms')
        # GXB _elements?
        apply_colour(scene._elements, scene.rooms, progress.d.special_cases)

        if False:
            sc_elems = scene._elements  # GXB _elements?
            sc_elems[scene.rooms[98]].update({'rgb': tuple(color_pallete[7])})
            sc_elems[scene.rooms[99]].update({'rgb': tuple(color_pallete[8])})
            sc_elems[scene.rooms[100]].update({'rgb': tuple(color_pallete[9])})
            sc_elems[scene.rooms[101]].update({'rgb': tuple(color_pallete[0])})
            sc_elems[scene.rooms[102]].update({'rgb': tuple(color_pallete[1])})

    '''--------------------------------------------------------------------'''
    '''Print out timings'''
    timings.total = (timings.dft
                     + timings.mask
                     + timings.watershed
                     + timings.disconnected
                     + timings.voxelise
                     + timings.source_markers
                     )

    for k, v in timings.asdict().items():
        log.info("time to %s: %s" % (k, v))

    '''--------------------------------------------------------------------'''
    if progress.s.preview:
        log.info('Previewing Rooms')
        displayer(scene, vox_group, progress, test)

    '''--------------------------------------------------------------------'''
    if progress.voxelised and progress.s.sort_rooms:
        log.info('Sorting Internal Rooms')
        scene.internals = []
        scene.externals = []
        for each in scene.rooms:
            if scene._elements[scene.rooms[each]].get('visible', False):
                scene.internals.append(each)
            else:
                scene.externals.append(each)

        scene.internals.sort(key=lambda x: scene._elements
                             [scene.rooms[x]].get('voxel_volume', 0.0),
                             reverse=True)

    '''--------------------------------------------------------------------'''
    log.debug('making folder for saving')
    if (progress.s.save
        or progress.s.save_to_svg
        or progress.s.save_to_qb
            or progress.s.save_supersolid):
        file_number = str(int(vox_grid.s.voxel_scale_factor * 100))

        try:
            override = progress.s.save_number_override
        except:
            override = None
        folder_path = create_folder(pickles,
                                    scene.s.scene,
                                    file_number,
                                    override)

    '''--------------------------------------------------------------------'''
    if progress.s.display_illustration:
        log.info('2D illustration')
        if progress.s.save_to_svg:
            folder = Path(folder_path)
            filepath1 = (folder / (folder.stem + "_plan"))
            filepath2 = (folder / (folder.stem + "_side"))
        else:
            filepath1 = None
            filepath2 = None
        view_illust(scene, vox_grid, vox_group, filepath1, filepath2)

    '''--------------------------------------------------------------------'''

    if progress.voxelised and progress.s.check_manifold:
        log.info("Checking Manifold")

        new_array = vox_group.connect.copy()
        values = set(vox_group.connect.flatten().tolist())
        non_internals = values.difference(scene.internals)
        non_internals.difference_update([0])

        for each in non_internals:
            new_array[new_array == each] = 0

        for each in scene.internals:
            new, new1, origin = minimize(new_array, each, True)
            new, origin = bigger(new, origin)  # 1st array is just the value
            new1, origin1 = bigger(new1, origin, 1)  # 2nd array test vs others
            new[new == each] = 1
            new1[new1 == each] = 1

            (int_vert_conflict,
             int_edge_conflict,
             ext_vert_conflict,
             ext_edge_conflict) = check_manifold.iterate2(new, new1)
            log.info("Checking %s is ext manifold - Corner: %s, Edge: %s:"
                     % (each, not ext_vert_conflict, not ext_edge_conflict))
            log.info("Checking %s is int manifold - Corner: %s, Edge: %s:"
                     % (each, not int_vert_conflict, not int_edge_conflict))

    '''--------------------------------------------------------------------'''

    if progress.voxelised and progress.s.make_supersolid:

        if progress.s.save_supersolid:
            folder = Path(folder_path)
            filepath = (folder / folder.stem)
            supersolid_path = filepath.with_suffix(".pkl")
        elif progress.s.save_supersolid:
            supersolid_path = progress.s.supersolid_file
        else:
            supersolid_path = None

        new_array = vox_group.connect.copy()
        values = set(vox_group.connect.flatten().tolist())
        non_internals = values.difference(scene.internals)
        non_internals.difference_update([0])

        for each in non_internals:
            new_array[new_array == each] = 0

        if progress.s.indi_supersolids:
            for n, each in enumerate(scene.internals):
                log.info("%s - %s" % (n, each))
                # GXB _elements?
                info = scene._elements[scene.rooms[each]].copy()
                minmax = [None] * 6
                new, origin = minimize(new_array, each, minmax=minmax)
                new, origin = bigger(new, origin)

                minmax = list(map(lambda x, y: x + y,
                                  minmax, [-1, 1, -1, 1, -1, 1]))

                if progress.s.dilate_rooms:

                    # TODO - handle case that minmax > vox_group.connect

                    new, origin = bigger(new, origin)
                    minmax = list(map(lambda x, y: x + y,
                                      minmax, [-1, 1, -1, 1, -1, 1]))
                    d = np.ones((3, 3, 3), bool)
                    dilated_room = binary_dilation(new, d)

                    new[dilated_room] = each
                    new[~dilated_room] = 0
                    cutter = vox_group.connect[minmax[0]: minmax[1],
                                               minmax[2]: minmax[3],
                                               minmax[4]: minmax[5]].copy()

                    for each_adjacent in scene.rooms:
                        if each_adjacent == each:
                            continue
                        try:

                            new[cutter == each_adjacent] = 0
                        except IndexError:
                            pass

                supersolid_expansion = progress.s.supersolid_expansion
                returned = make_solid.solidify(new,
                                               origin,
                                               vox_grid.trsf_Voxel2Global,
                                               supersolid_expansion)
                rooms, room_hash, elements = returned

                r, g, b = scene.elements[scene.rooms[each]]['rgb']
                # GXB _elements?
                scene._elements[rooms[each]] = info
                scene.rooms[each] = rooms[each]
                scene.room_hash[rooms[each]] = each
                scene._elements[rooms[each]].update(elements[rooms[each]])
                scene._elements[rooms[each]]['rgb'] = (r, g, b)
                # GXB _elements?

                if progress.s.export is True and False:
                    shape = scene._elements[scene.rooms[each]]['shape']
                    simple_solid = convertTopology(shape)
                    SDO_GEOM = Py3DGeom.SDO_Geometry(simple_solid)
                    r, g, b = scene._elements[scene.rooms[each]]['rgb']
                    data = {"guid": each,
                            "red": r,
                            "green": g,
                            "blue": b,
                            "usid": n,
                            "subid": 0,
                            "filename": "spaces",
                            }

                    Py3DGeom.inject(None, SDO_GEOM, progress.s.export_table,
                                    data)

        else:
            new, origin = minimize(new_array)
            new, origin = bigger(new, origin)
            returned = make_solid.solidify(new,
                                           origin,
                                           vox_grid.trsf_Voxel2Global,
                                           progress.s.supersolid_expansion,
                                           supersolid_path,
                                           progress.s.load_supersolid)
            rooms, room_hash, elements = returned

            for each in scene.internals:
                info = scene.elements[scene.rooms[each]].copy()
                r, g, b = scene.elements[scene.rooms[each]]['rgb']
                scene.elements[rooms[each]] = info
                scene.rooms[each] = rooms[each]
                scene.room_hash[rooms[each]] = each
                scene.elements[rooms[each]].update(elements[rooms[each]])
                scene.elements[rooms[each]]['rgb'] = (r, g, b)

    '''--------------------------------------------------------------------'''
    if progress.s.save:

        log.info("Saving")

        progress.room_info = {"internals": len(scene.internals),
                              "externals": len(scene.externals)}

        progress.scene_loaded = True
        progress.vox_grid_set_up = True
        progress.vox_group_set_up = True

        vox_grid.save(folder_path)
        progress.save(folder_path, timings)
        vox_group.save(folder_path)
        scene.save(folder_path)
        log.info("Saving to %s" % folder_path.as_posix())

    '''--------------------------------------------------------------------'''
    log.info('''Display elements and voxels''')

    if progress.s.display:
        displayer(scene, vox_group, progress, test)

    '''--------------------------------------------------------------------'''
    if progress.s.export:
        log.info("Exporting to Oracle")
        for n, each in enumerate(scene.internals):
            shape = scene._elements[scene.rooms[each]]['shape']
            log.info(shape)
            simple_solid = convertTopology(shape)
            SDO_GEOM = Py3DGeom.SDO_Geometry(simple_solid)
            r, g, b = scene._elements[scene.rooms[each]]['rgb']
            print(type(each))
            print(each)
            data = {"guid": int(each),
                    "red": r,
                    "green": g,
                    "blue": b,
                    "usid": n,
                    "subid": 0,
                    "filename": "spaces",
                    }

            Py3DGeom.inject(None, SDO_GEOM, progress.s.export_table, data)
        log.info("Export to Oracle complete")


if __name__ == '__main__':

    import getpass
    import sys

    from settings.setting_combinations import Setting

    if getpass.getuser() == 'gareth' and sys.platform == 'win32':
        environment = "PC"
    elif getpass.getuser() == 'garethboyes' and sys.platform == "darwin":
        environment = "macbook"
        from settings.mac_settings import setting_id
    elif getpass.getuser() == 'garethboyes' and sys.platform == "win32":
        environment = "parallels"
        from settings.par_settings import setting_id

    setting = Setting(setting_id)

    main(setting, environment)

    log.info("The End")
