# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import scipy.ndimage as si
import numpy as np
from itertools import product
import networkx as nx

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)

adjacencies = ((1, 0, 0), (0, 1, 0), (0, 0, 1))


def edge_count_inc(G, val_1, val_2):

    if not G.has_edge(val_1, val_2):
        G.add_edge(val_1, val_2, count=0)
    G.edges[val_1, val_2]['count'] += 1


def determine_adjaceny(vox_group, scene):

    sx = vox_group.sx
    sy = vox_group.sy
    sz = vox_group.sz

    G = nx.Graph()
    G.add_nodes_from((0, 1, 2, "E"))
    G.add_nodes_from((vox_group.sources))

    for x, y, z in product(range(sx), range(sy), range(sz)):

        value = vox_group.connect[x, y, z]

        for dx, dy, dz in adjacencies:

            if x + dx == sx or y + dy == sy or z + dz == sz:
                if vox_group.connect[x, y, z]:
                    edge_count_inc(G, value, "E")
                continue
            else:
                adj_val = vox_group.connect[x + dx, y + dy, z + dz]

            if x - dx == -1 or y - dy == -1 or z - dz == -1:
                if vox_group.connect[x, y, z]:
                    edge_count_inc(G, value, "E")

            # If the adjacent voxel has the same value then do nothing
            if vox_group.connect[x, y, z] == vox_group.connect[x + dx,
                                                               y + dy,
                                                               z + dz]:
                continue

            # If the voxel is not zero and the adjacent voxel is in
            # the buffer zone then increase by one.
            if (vox_group.connect[x, y, z]
                and vox_group.buffer[x + dx, y + dy, z + dz]):
                edge_count_inc(G, value, "E")

            # And vice versa
            elif (vox_group.connect[x + dx, y + dy, z + dz]
                  and vox_group.buffer[x, y, z]):
                pass

            else:
                edge_count_inc(G, value, adj_val)

    internal = []

    # for each in count:

    # if each < vox_group.start:
    # continue

    # if count[each][1] == 0:
    # continue
    # if count[each][0] > count[each][1]:
    # continue
    # if ((count[each][1] - count[each][0]) /
    # (count[each][0] + count[each][1])) > 0.25:
    # internal.append(each)

    # pprint(count)

    for each_node in G.nodes:

        if each_node in [0, 1, 2, "E"]:
            continue

        if identify_external(G, each_node):

            vox_group.sources[each_node]['internal'] = False

        else:
            internal.append(each_node)
            vox_group.sources[each_node]['internal'] = True

    scene.adjacency_graph = G

    return internal


def identify_external(G, node):

    adjacents = G._adj[node]
    externals = 0
    internals = 0

    for each in adjacents:

        if each == "E":
            externals += adjacents[each]['count']
        else:
            internals += adjacents[each]['count']

    print(node, externals, internals)

    if externals + internals == 0:
        return True

    return externals / (externals + internals) > 0.3


def identify_disconnected(vox_group):

    sx = vox_group.sx
    sy = vox_group.sy
    sz = vox_group.sz

    graphs = {}
    for each in [0, 1, 2]:
        graphs[each] = nx.Graph()
    for each in vox_group.sources:
        graphs[each] = nx.Graph()

    # ---------------
    # TODO
    # Incorporate correct room number
    # ---------------

    if len(vox_group.sources) == 0:
        next_source = 11
    else:
        next_source = max(list(vox_group.sources)) + 1

    for x, y, z in product(range(sx), range(sy), range(sz)):

        value = vox_group.connect[x, y, z]

        isolated = True
        for dx, dy, dz in adjacencies:

            if x + dx == sx or y + dy == sy or z + dz == sz:
                continue
            else:
                adj_val = vox_group.connect[x + dx, y + dy, z + dz]

            if value == adj_val:
                graphs[value].add_edge((x, y, z), (x + dx, y + dy, z + dz))
                isolated = False

        if isolated:
            graphs[value].add_node((x, y, z))

    new_graphs = {}
    for each_graph in graphs:

        if each_graph in [0, 1, 2]:
            continue

        if len(graphs[each_graph].nodes) == 0:
            continue

        subs = [graphs[each_graph].subgraph(c).copy()
                for c in nx.connected_components(graphs[each_graph])]

        if len(subs) <= 1:
            continue

        subs.sort(key=lambda x: -len(x.nodes))

        for ea_sub in subs[1:]:

            new_graphs[next_source] = ea_sub.copy()
            vox_group.sources[next_source] = {'belongs_to': each_graph,
                                              'value': next_source,
                                              'pos': next(iter(ea_sub.nodes))}

            for each_node in ea_sub.nodes:

                x, y, z = each_node
                vox_group.connect[each_node] = next_source

            next_source += 1


def identify_disconnected_(vox_group):

    if len(vox_group.sources) == 0:
        max_source = 10
    else:
        max_source = max(list(vox_group.sources))

    new_connect = vox_group.connect.copy()

    for each_source in list(vox_group.sources):
        array = vox_group.connect.copy()
        array[array != each_source] = 0

        new_array, count = si.label(array)

        counts = list(range(1, count + 1))

        if count != 1:
            counts = list(range(1, count + 1))
            counts.sort(key=lambda x: (new_array[new_array == x]).sum() / x,
                        reverse=True)
            n = 0

            for n, each_count in enumerate(counts[1:]):

                first_vx = np.transpose(np.nonzero(new_array == each_count))[0]
                first_voxel = tuple(map(float, first_vx))

                new_connect[new_array == each_count] = max_source + n + 1
                next_source = max_source + n + 1
                vox_group.sources[next_source] = {'belongs_to': each_source,
                                                  'value': next_source,
                                                  'pos': first_voxel,
                                                  "internal": False}

            max_source = max_source + n + 1

        else:
            pass

    vox_group.connect = new_connect
