#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import operator
import json
from collections import namedtuple

import OCC.Core.gp as gp
import OCC.Core.BRepBndLib as blib
import OCC.Core.Bnd as bnd

from OCC.Core.BRepBuilderAPI import (BRepBuilderAPI_MakeVertex)
from OCC.Core.BRepBuilderAPI import BRepBuilderAPI_Transform as brep_Transform

import numpy as np

from settings.load_settings import get_grid_settings

from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)


class VoxelGrid(object):

    def __init__(self,
                 grid_no=None,
                 load=False,
                 offset=(0, 0, 0),
                 update={},
                 scene=None):

        if grid_no is None:
            grid_no = "_default"

        if isinstance(grid_no, (list, tuple)) and len(grid_no) == 3:
            s, _ = get_grid_settings("_from_size")
            self.grid_no = "_from_size"
            self.s = s._replace(voxel_shape=grid_no, grid_method=3)

        elif not load:
            self.s, _ = get_grid_settings(grid_no)
            self.grid_no = grid_no
        else:
            self.load(load)
            self.grid_no = "_loaded"

        self.s = self.s._replace(**update)
        self.scene = scene

        self.offset = offset
        self.setup()

    def load(self, folder_path):

        grid_path = folder_path / (folder_path.stem + "_grid.json")

        with grid_path.open('r') as f:
            grid_dict = json.load(f)

        _, s = get_grid_settings("_default")
        s.update(grid_dict)

        self.s = namedtuple("grid_settings", s)(**s)

    def save(self, folder_path):

        grid_path = folder_path / (folder_path.stem + "_grid.json")

        grid_dict = {
            'grid_method': 3,
            'corner1': self.origin,
            'voxel_shape': self.voxel_grid,
            'voxel_scale_factor': self.voxel_scale_factor,
            'x_dir': self.x_dir,
            'z_dir': self.z_dir,

        }
        with grid_path.open('w') as f:
            # json.dump(self.s._asdict(), f)
            json.dump(grid_dict, f)

    def setup(self):

        if self.s.grid_method == 2:

            self.setup_from_corners(self.s.corner1,
                                    self.s.corner2,
                                    self.s.x_dir,
                                    self.s.z_dir,
                                    self.s.voxel_scale_factor,
                                    )

        elif self.s.grid_method == 1:
            self.setup_from_centre(self.s.centre,
                                   self.s.voxel_grid_dist,
                                   self.s.x_dir,
                                   self.s.z_dir,
                                   self.s.voxel_scale_factor,
                                   )

        elif self.s.grid_method == 3:

            self.setup_from_array_size(self.s.x_dir,
                                       self.s.z_dir,
                                       self.s.corner1,
                                       self.s.voxel_shape,
                                       self.s.voxel_scale_factor)

        elif self.s.grid_method == 4:

            self.setup_from_features()

    def setup_from_centre(self,
                          centre,
                          voxel_grid_dist,
                          x_dir,
                          z_dir,
                          voxel_scale_factor):

        self.centre = tuple(map(operator.sub, centre, self.offset))

        self.voxel_grid_dist = voxel_grid_dist
        self.x_dir = x_dir
        self.z_dir = z_dir
        self.voxel_scale_factor = float(voxel_scale_factor)

        self.voxel_centre = gp.gp_Pnt(*self.centre)

        self.voxel_x_dir = gp.gp_Dir(*self.x_dir)
        self.voxel_z_dir = gp.gp_Dir(*self.z_dir)

        self.voxel_axes = gp.gp_Ax3(self.voxel_centre,
                                    self.voxel_z_dir,
                                    self.voxel_x_dir)
        self.trsf_Voxel2Global = gp.gp_Trsf()
        self.global_axes = gp.gp_Ax3()
        self.trsf_Voxel2Global.SetTransformation(self.voxel_axes,
                                                 self.global_axes)

        self.voxel_origin = gp.gp_Vec(*self.voxel_grid_dist).Reversed()
        self.voxel_origin.Transform(self.trsf_Voxel2Global)
        self.voxel_origin.Add(gp.gp_Vec(*self.centre))

        voxel_origin_vec = gp.gp_Vec(self.voxel_origin.XYZ())
        self.trsf_Voxel2Global.SetTranslationPart(voxel_origin_vec)
        self.trsf_Voxel2Global.SetScaleFactor(self.voxel_scale_factor)

        self.sx = int(self.voxel_grid_dist[0] * 2 / self.voxel_scale_factor)
        self.sy = int(self.voxel_grid_dist[1] * 2 / self.voxel_scale_factor)
        self.sz = int(self.voxel_grid_dist[2] * 2 / self.voxel_scale_factor)

        self.voxel_grid = (self.sx, self.sy, self.sz)

        self.trsf_vg_matrix = np.empty(shape=(4, 3), dtype=np.float_)
        for row in [1, 2, 3]:
            for col in [1, 2, 3]:
                value = self.trsf_Voxel2Global.Value(row, col)
                self.trsf_vg_matrix[row - 1, col - 1] = value
        coord = self.trsf_Voxel2Global.TranslationPart().Coord()
        self.trsf_vg_matrix[3:] = coord

        self.trsf_Global2Voxel = self.trsf_Voxel2Global.Inverted()

        self.trsf_gv_matrix = np.empty(shape=(3, 4), dtype=np.float_)
        for row in [1, 2, 3]:
            for col in [1, 2, 3]:
                value = self.trsf_Global2Voxel.Value(row, col)
                self.trsf_gv_matrix[row - 1, col - 1] = value
            self.trsf_gv_matrix[row - 1, 3] =\
                self.trsf_Global2Voxel.TranslationPart().Coord()[row - 1]

    def setup_from_corners(self,
                           corner1=None,
                           corner2=None,
                           x_dir=None,
                           z_dir=None,
                           voxel_scale_factor=None):

        self.process_setup_part1(x_dir, z_dir, voxel_scale_factor)

        pt1 = gp.gp_Pnt(*corner1)
        pt2 = gp.gp_Pnt(*corner2)
        pt1.Transform(self.trsf_Global2Voxel)
        pt2.Transform(self.trsf_Global2Voxel)

        pt1 = BRepBuilderAPI_MakeVertex(pt1).Vertex()
        pt2 = BRepBuilderAPI_MakeVertex(pt2).Vertex()
        bb = bnd.Bnd_Box()
        blib.brepbndlib_Add(pt1, bb)
        blib.brepbndlib_Add(pt2, bb)
        cornermin = bb.CornerMin()
        cornermax = bb.CornerMax()
        size = tuple(map(int, gp.gp_Vec(cornermin, cornermax).Coord()))

        self.process_setup_part2(corner1, size)

    def setup_from_array_size(self,
                              x_dir=None,
                              z_dir=None,
                              corner1=None,
                              size=None,
                              scale=None):
        if x_dir is None:
            x_dir = self.s.x_dir
        if z_dir is None:
            z_dir = self.s.z_dir
        if corner1 is None:
            corner1 = self.s.corner1
        if size is None:
            size = self.s.voxel_shape
        if scale is None:
            scale = self.s.voxel_scale_factor

        corner1 = tuple(map(operator.sub, corner1, self.offset))

        self.process_setup_part1(x_dir, z_dir, scale)
        self.process_setup_part2(corner1, size)

    def setup_from_features(self, x_dir=None, z_dir=None, scale=None):

        if x_dir is None:
            x_dir = self.s.x_dir
        if z_dir is None:
            z_dir = self.s.z_dir
        if scale is None:
            scale = self.s.voxel_scale_factor

        self.process_setup_part1(x_dir, z_dir, scale)

        bb = bnd.Bnd_Box()

        for each_feature in self.scene.features:

            shape = self.scene._elements[each_feature]['shape']
            transform = brep_Transform(self.trsf_Global2Voxel)
            transform.Perform(shape, True)
            shape1 = transform.ModifiedShape(shape)
            blib.brepbndlib_Add(shape1, bb)

        cornermin = bb.CornerMin()
        cornermax = bb.CornerMax()

        cornermin.Translate(gp.gp_Vec(-2 * scale, -2 * scale, -2 * scale))
        cornermax.Translate(gp.gp_Vec(3 * scale, 3 * scale, 3 *scale))

        corner1 = cornermin.Transformed(self.trsf_Voxel2Global)
        corner1 = corner1.Coord()

        size = tuple(map(int, gp.gp_Vec(cornermin, cornermax).Coord()))

        self.process_setup_part2(corner1, size)

    def setup_from_xdir_zdir_scale_corner1_size(self, x_dir, z_dir,
                                                scale, corner1, size):

        self.process_setup_part1(x_dir, z_dir, scale)
        self.process_setup_part2(corner1, size)

    def process_setup_part1(self, x_dir, z_dir, scale):

        self.voxel_origin = gp.gp_Pnt(0, 0, 0)

        self.x_dir = x_dir
        self.z_dir = z_dir

        self.voxel_x_dir = gp.gp_Dir(*self.x_dir)
        self.voxel_z_dir = gp.gp_Dir(*self.z_dir)

        self.voxel_axes = gp.gp_Ax3(self.voxel_origin,
                                    self.voxel_z_dir, self.voxel_x_dir)
        self.global_axes = gp.gp_Ax3()

        self.trsf_Voxel2Global = gp.gp_Trsf()

        self.trsf_Voxel2Global.SetTransformation(self.voxel_axes,
                                                 self.global_axes)

        self.voxel_scale_factor = float(scale)
        self.trsf_Voxel2Global.SetScaleFactor(self.voxel_scale_factor)

        self.process_transformations()

    def process_setup_part2(self, origin, size):

        self.voxel_origin = gp.gp_Pnt(*origin)
        self.origin = origin
        self.sx = int(size[0])
        self.sy = int(size[1])
        self.sz = int(size[2])
        self.voxel_grid = (self.sx, self.sy, self.sz)

        vec = gp.gp_Vec(self.voxel_origin.XYZ())
        self.trsf_Voxel2Global.SetTranslationPart(vec)

        self.process_transformations()
        log.info("origin - %.2f, %.2f, %.2f" % (origin[0], origin[1],
                                                 origin[2]))

        log.info("grid size - %s, %s, %s" % (self.sx, self.sy, self.sz))
        
        log.info("voxel size - %s" % (self.voxel_scale_factor))  

    def process_transformations(self):

        self.trsf_vg_matrix = np.empty(shape=(4, 3), dtype=np.float_)
        for row in [1, 2, 3]:
            for col in [1, 2, 3]:
                value = self.trsf_Voxel2Global.Value(row, col)
                self.trsf_vg_matrix[row - 1, col - 1] = value
        coord = self.trsf_Voxel2Global.TranslationPart().Coord()
        self.trsf_vg_matrix[3:] = coord

        self.trsf_Global2Voxel = self.trsf_Voxel2Global.Inverted()

        self.trsf_gv_matrix = np.empty(shape=(3, 4), dtype=np.float_)
        for row in [1, 2, 3]:
            for col in [1, 2, 3]:
                value = self.trsf_Global2Voxel.Value(row, col)
                self.trsf_gv_matrix[row - 1, col - 1] = value
            self.trsf_gv_matrix[row - 1, 3] =\
                self.trsf_Global2Voxel.TranslationPart().Coord()[row - 1]

    @property
    def shape(self):

        return (self.sx, self.sy, self.sz)


def test():
    global v
    v = VoxelGrid(update={"grid_method": 2,
                          "corner1": (0, 0, 0),
                          "corner2": (10, 10, 10),
                          "voxel_shape": (10, 10, 10),
                          "x_dir": (1, 0, 0),
                          "z_dir": (0, 0, 1),
                          "voxel_scale_factor": 2})

    pass


if __name__ == "__main__":

    test()
