# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import time
import json

import numpy as np
from scipy import ndimage as ndi
from scipy.ndimage import label as si_label

import OCC.Core.Aspect as asp
import OCC.Core.Graphic3d as gr
import OCC.Core.Prs3d as pr
import OCC.Core.Quantity as qu
from OCC.Core.TopoDS import TopoDS_Compound
from OCC.Core.BRep import BRep_Builder
from OCC.Core.gp import gp_Pnt

from OCC.Core.BRepPrimAPI import BRepPrimAPI_MakeSphere
import OCC.Core.AIS as ais

import voxel_creator as gen
from modules.thirtytwo_bit import t32_bit

from modules.triangulate import triangulate, transform, voxelise
import modules.add_mat as add_mat
from modules.grow_barrier_array import grow_barrier_array
from modules.form_solid_from_voxels import minimize, bigger

from utils.logging_env import setup_logging, logging, log_debug
log = logging.getLogger(__name__)
setup_logging(log)
if log.getEffectiveLevel() <= 10:
    log_debug = True


class VoxelArrayGroup:

    def __init__(self, vox_grid, progress, load=False, preload=False,
                 dist_max=99999999):

        self.progress = progress
        self.dist_max = dist_max
        self.vox_grid = vox_grid
        self.voxel_grid = self.vox_grid.voxel_grid
        try:
            self.start = self.scene.s.start
        except:
            log.warning("room_start not set - setting to 11")
            self.start = 11

        self.trsf_vg_matrix = self.vox_grid.trsf_vg_matrix

        if not load:

            self.array = np.full(self.voxel_grid, 0, dtype=bool)
            self.mask = np.full_like(self.array, 0)

            self.dft = np.full_like(self.array, dist_max, dtype=np.int32)
            self.raw_dft = None
            self.connect = np.full_like(self.array, 0)
            self.sources = {}

        else:
            self.load(load, preload)

        self.debug = np.full_like(self.array, 0)

        self.grid_drawn = False
        self.mask_drawn = False
        self.dft_drawn = False
        self.debug_drawn = False
        self.array_drawn = False
        self.connect_drawn = False
        self.source_drawn = False
        self.dft_by_val_drawn = False

        self.sx = self.array.shape[0]
        self.sy = self.array.shape[1]
        self.sz = self.array.shape[2]

        self.array_draw_height = int(self.sz / 2)
        self.dft_draw_height = int(self.sz / 2)
        self.mask_draw_height = int(self.sz / 2)

    def generate_mask(self, scene, mask_buffer=0):

        scene_elements = scene.elements
        hull_hash = scene.hull_hash

        # dft.dft(self.voxel_group,
        # self.mask,
        # np.full_like(self.mask, 0),
        # scene_elements,
        # voxel_scene_elements,
        # selected_elements,
        # True, #parallel,
        # True) #no_dft)

        node_array, face_list = triangulate(scene_elements[hull_hash]['shape'])
        transform(node_array, self.vox_grid.trsf_Global2Voxel)
        vx, voxels = voxelise(node_array, face_list, 1, True)

        offset = tuple(voxels.origin.astype(np.int32))

        hull = vx.copy()
        hull_offset = offset

        for n in range(mask_buffer):
            vx, offset = grow_barrier_array(vx,
                                            offset[0],
                                            offset[1],
                                            offset[2],
                                            self.voxel_grid[0],
                                            self.voxel_grid[1],
                                            self.voxel_grid[2],
                                            diag=True,
                                            hollow=False)

        mask = np.zeros(self.voxel_grid, dtype=np.bool)
        hull_buffer = np.zeros(self.voxel_grid, dtype=np.bool).copy()
        mask = add_mat.or_array(mask,
                                vx,
                                offset)

        self.buffer = mask & ~add_mat.or_array(hull_buffer, hull, hull_offset)

        log.info("Mask sum %s" % mask.sum())
        self.mask = (~mask)

    def perform_voxelise_trimesh(self, scene):

        time_start = time.time()
        array = np.zeros(self.voxel_grid, dtype=np.bool)

        interior = self.progress.s.trimesh_interior

        precision = self.progress.s.voxel_set_precision

        for each in scene.features:

            node_array, face_list = triangulate(scene._elements[each]['shape'],
                                                precision)

            transform(node_array, self.vox_grid.trsf_Global2Voxel)
            vx, voxels = voxelise(node_array,
                                  face_list,
                                  1,
                                  interior=interior)

            add_mat.or_array(array,
                             vx,
                             tuple(voxels.origin.astype(np.int32)))

            # TODO research effect of masking dft

            if False:
                add_mat.or_array(array,
                                 self.mask.astype(np.bool),
                                 (0, 0, 0))

        self.array = array

        log.info('time taken find objects %s' % (time.time() - time_start))

    def perform_dft(self, dft_factor=2):

        time_start = time.time()
        dft = ndi.distance_transform_edt(~self.array2)
        log.debug('time taken dft %s' % (time.time() - time_start))

        time_start = time.time()
        half_distance = dft / dft_factor
        ceil_half = np.ceil(half_distance)
        long_ceil_half = ceil_half.astype(t32_bit())
        self.dft = long_ceil_half  # * 2
        log.debug('time taken convert %s' % (time.time() - time_start))

        self.raw_dft = (dft * 1000).astype(t32_bit())
        self.raw_dft[self.mask] = 0
        self.dft[self.mask] = 0
        self.n = self.dft.max() + 1

    def save(self, folder_path):

        stem = folder_path.stem
        array_path = folder_path / (stem + "_array.npy")
        connect_path = folder_path / (stem + "_connect.npy")
        mask_path = folder_path / (stem + "_mask.npy")
        dft_path = folder_path / (stem + "_dft.npy")
        trsf_path = folder_path / (stem + "_trsf.npy")
        sources_path = folder_path / (stem + "_sources.json")

        np.save(array_path, self.array)
        np.save(connect_path, self.connect)
        np.save(mask_path, self.mask)
        np.save(dft_path, self.dft)
        np.save(trsf_path, self.trsf_vg_matrix)

        stringed_sources = {str(key): value
                            for key, value in self.sources.items()}

        for each_source in stringed_sources:
            ais = stringed_sources[each_source].get("ais", None)
            if ais is not None:
                stringed_sources[each_source]['ais'] = None

        with sources_path.open("w") as f:
            json.dump(stringed_sources, f, indent=2)

    def load(self, folder_path, preload):

        stem = folder_path.stem
        array_path = folder_path / (stem + "_array.npy")
        connect_path = folder_path / (stem + "_connect.npy")
        mask_path = folder_path / (stem + "_mask.npy")
        dft_path = folder_path / (stem + "_dft.npy")
        trsf_path = folder_path / (stem + "_trsf.npy")
        sources_path = folder_path / (stem + "_sources.json")

        log.info("Loading array")
        self.array = np.load(array_path)
        log.info("Loading mask")
        self.mask = np.load(mask_path)
        self.trsf_vg_matrix = np.load(trsf_path)
        self.n = self.array.max() + 1

        if preload:
            log.info("voxels only")
            try:
                self.dft = np.full_like(self.array, self.dist_max,
                                        dtype=t32_bit())
            except AttributeError:
                self.dft = np.full_like(self.array, 99, dtype=t32_bit())
                log.warning("************************")
                log.warning("scene.dist_max not saved")
                log.warning("************************")

            self.raw_dft = np.full_like(self.array, 0, dtype=np.float)
            self.connect = np.full_like(self.array, 0)
            self.sources = {}
        else:
            log.info("Loading connect")
            self.connect = np.load(connect_path)
            log.info("Loading dft")
            self.dft = np.load(dft_path)

            if sources_path.exists():
                log.info("Loading sources")
                with sources_path.open() as f:
                    stringed_sources = json.load(f)
                self.sources = {int(key): value
                                for key, value in stringed_sources.items()}

            else:
                self.sources = {}

    def display(self, array, display, maxval, height=None, minval=3):

        if height is None:
            height = self.sz

        SM = display._struc_mgr
        aPres = pr.Prs3d_Presentation(SM)
        group = pr.Prs3d_Root_CurrentGroup(aPres)

        aFaceAspect = gr.Graphic3d_AspectFillArea3d()
        aLineAspect = gr.Graphic3d_AspectLine3d()
        aWhite = qu.Quantity_Color(qu.Quantity_NOC_WHITE)
        aLineAspect.SetColor(aWhite)
        aBrassMaterial = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_BRASS)
        aFaceAspect.SetInteriorStyle(asp.Aspect_IS_SOLID)
        aBlue = qu.Quantity_Color(qu.Quantity_NOC_MATRABLUE)
        aFaceAspect.SuppressBackFace()
        aFaceAspect.SetFrontMaterial(aBrassMaterial)
        aFaceAspect.SetInteriorColor(aBlue)

        group.SetPrimitivesAspect(aFaceAspect)
        group.SetPrimitivesAspect(aLineAspect)

        array_ = array[:, :, :height].copy()

        gen.iterate2(array_, self.mask, group, self.trsf_vg_matrix,
                     minval, maxval)

        aPres.Display()

    def display_room(self, display):

        SM = display._struc_mgr
        aPres = pr.Prs3d_Presentation(SM)
        group = pr.Prs3d_Root_CurrentGroup(aPres)

        aFaceAspect = gr.Graphic3d_AspectFillArea3d()
        aLineAspect = gr.Graphic3d_AspectLine3d()
        aWhite = qu.Quantity_Color(qu.Quantity_NOC_WHITE)
        aLineAspect.SetColor(aWhite)
        aBrassMaterial = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_BRASS)
        aFaceAspect.SetInteriorStyle(asp.Aspect_IS_SOLID)
        # aFaceAspect.SetInteriorStyle(asp.Aspect_IS_HATCH)

        aBlue = qu.Quantity_Color(qu.Quantity_NOC_MATRABLUE)
        aFaceAspect.SuppressBackFace()

        aFaceAspect.SetFrontMaterial(aBrassMaterial)
        aFaceAspect.SetInteriorColor(aBlue)

        group.SetPrimitivesAspect(aFaceAspect.GetHandle())
        group.SetPrimitivesAspect(aLineAspect.GetHandle())

        array = self.connect.copy()

        gen.iterate2(array, self.mask, group, self.trsf_vg_matrix, 3, 998)
        aPres.Display()

    # def write_blob(self, filename):

        # n = self.array.max() + 1
        # palette = seaborn.color_palette("deep", n)

        # with open(filename, "w") as the_file:
            # the_file.write("Qubicle Exchange Format\nVersion 0.2\nwww.minddesk.com\n""")
            # the_file.write(str(self.sy) + " " + str(self.sz) + " " + str(self.sx) + "\n")

            # the_file.write(str(len(palette)) + "\n")
            # for each in palette:

                # the_file.write(str(each[0]) + " " + str(each[1]) + " " + str(each[2]) + "\n")

            # for x in range(self.sx):
                # for y in range(self.sy):
                    # for z in range(self.sz):
                        # if not np.isnan(self.connect[x, y, z]):
                            # if self.connect[x, y, z] != 0:
                                # the_file.write(str(y) + " " + str(z) + " " + str(x) + " " + str(self.connect[x, y, z]) + " 126\n")

    def display2(self, array, display, maxval, height=None, minval=3,
                 color="red"):

        if height is None:
            height = self.sz

        SM = display._struc_mgr
        aPres = pr.Prs3d_Presentation(SM)
        group = pr.Prs3d_Root_CurrentGroup(aPres)

        aBlack = qu.Quantity_Color(qu.Quantity_NOC_BLACK)

        aFaceAspect = gr.Graphic3d_AspectFillArea3d()
        aLineAspect = gr.Graphic3d_AspectLine3d()

        aLineAspect.SetColor(aBlack)
        aStoneMaterial = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_STONE)
        aFaceAspect.SetInteriorStyle(asp.Aspect_IS_SOLID)
        # aFaceAspect.SetInteriorStyle(asp.Aspect_IS_HATCH)
        aBlue = qu.Quantity_Color(qu.Quantity_NOC_MATRABLUE)
        aFaceAspect.SuppressBackFace()
        aFaceAspect.SetFrontMaterial(aStoneMaterial)
        aFaceAspect.SetInteriorColor(aBlue)

        group.SetPrimitivesAspect(aFaceAspect)
        group.SetPrimitivesAspect(aLineAspect)

        array_ = array[:, :, :height].astype(t32_bit())

        gen.iterate2(array_, np.full_like(array_, 0), group,
                     self.trsf_vg_matrix, minval, maxval)

        return aPres

    def make_grid_display(self):

        SM = self.display._struc_mgr
        self.gridPres = pr.Prs3d_Presentation(SM)
        group = pr.Prs3d_Root_CurrentGroup(self.gridPres)

        aLineAspect = gr.Graphic3d_AspectLine3d()
        aBlack = qu.Quantity_Color(qu.Quantity_NOC_BLACK)
        aLineAspect.SetColor(aBlack)
        group.SetPrimitivesAspect(aLineAspect)

        self.grid_drawn = True

        gen.display_grid(group,
                         self.vox_grid.trsf_vg_matrix,
                         self.sx,
                         self.sy,
                         self.sz)

        self.gridPres.Erase()
        self.display.Repaint()

    def add_grid_to_menu(self):

        self.disp['add_menu']('grid')
        self.disp['add_function_to_menu']('grid', self.show_grid)
        self.disp['add_function_to_menu']('grid', self.hide_grid)

    def show_grid(self):
        log.info("Displaying Grid")
        if not self.grid_drawn:
            self.make_grid_display()
            self.grid_drawn = True

        self.gridPres.Display()
        self.display.Repaint()

    def hide_grid(self):
        if self.grid_drawn:
            self.gridPres.Erase()
            self.display.Repaint()

    def add_debug_to_menu(self):

        self.disp['add_menu']('debug')
        self.disp['add_function_to_menu']('debug', self.show_debug)
        self.disp['add_function_to_menu']('debug', self.hide_debug)
        self.disp['add_function_to_menu']('debug', self.debug_increment)
        self.disp['add_function_to_menu']('debug', self.debug_decrement)

    def debug_decrement(self):

        if not self.debug_drawn:
            return

        self.debugPres.Clear()
        self.display.Repaint()

        self.debug_draw_height -= 1

        self.debugPres = self.display2(self.debug,
                                       self.display,
                                       99,
                                       self.debug_draw_height,
                                       1)
        self.debugPres.Display()
        self.display.Repaint()

    def debug_increment(self):

        if not self.debug_drawn:
            return

        self.debugPres.Clear()
        self.display.Repaint()

        self.debug_draw_height += 1

        self.debugPres = self.display2(self.debug,
                                       self.display,
                                       99,
                                       self.debug_draw_height,
                                       1)
        self.debugPres.Display()
        self.display.Repaint()

        print("increment")

    def make_debug_display(self):

        self.debug_drawn = True
        self.debug_draw_height = int(self.sz)

        self.debugPres = self.display2(self.debug,
                                       self.display,
                                       99,
                                       self.debug_draw_height,
                                       1)

        self.debugPres.Erase()
        self.display.Repaint()

    def show_debug(self):

        if not self.debug_drawn:
            self.make_debug_display()
            self.debug_drawn = True

        self.debugPres.Display()
        self.display.Repaint()

    def hide_debug(self):
        if self.debug_drawn:
            self.debugPres.Erase()
            self.display.Repaint()

    def make_mask_display(self):

        SM = self.display._struc_mgr
        pr.Prs3d_Presentation(SM)

        self.maskPres = self.display2(self.mask,
                                      self.display,
                                      99,
                                      self.mask_draw_height,
                                      1)

        group = pr.Prs3d_Root_CurrentGroup(self.maskPres)

        aLineAspect = gr.Graphic3d_AspectLine3d()
        aWhite = qu.Quantity_Color(qu.Quantity_NOC_WHITE)
        aLineAspect.SetColor(aWhite)
        group.SetPrimitivesAspect(aLineAspect)

        self.mask_drawn = True

        self.maskPres.Erase()
        self.display.Repaint()

    def add_mask_to_menu(self):

        self.disp['add_menu']('mask')
        self.disp['add_function_to_menu']('mask', self.show_mask)
        self.disp['add_function_to_menu']('mask', self.hide_mask)
        self.disp['add_function_to_menu']('mask', self.mask_inc)
        self.disp['add_function_to_menu']('mask', self.mask_dec)

    def show_mask(self):

        if not self.mask_drawn:
            self.make_mask_display()
            self.mask_drawn = True

        self.maskPres.Display()
        self.display.Repaint()

    def hide_mask(self):
        if self.mask_drawn:
            self.maskPres.Erase()
            self.display.Repaint()

    def mask_inc(self):

        if self.mask_draw_height != int(self.sz):
            self.mask_draw_height += 1

        if self.mask_drawn:

            self.maskPres.Clear()

            self.maskPres = self.display2(self.mask,
                                          self.display,
                                          99,
                                          self.mask_draw_height,
                                          1)
            self.maskPres.Display()
            self.display.Repaint()

    def mask_dec(self):

        if self.mask_draw_height != 0:
            self.mask_draw_height -= 1

        if self.mask_drawn:

            self.maskPres.Clear()

            self.maskPres = self.display2(self.mask,
                                          self.display,
                                          99,
                                          self.mask_draw_height,
                                          1)
            self.maskPres.Display()
            self.display.Repaint()

    def make_dft_display(self):

        SM = self.display._struc_mgr
        self.dftPres = pr.Prs3d_Presentation(SM)
        group = pr.Prs3d_Root_CurrentGroup(self.dftPres)

        aLineAspect = gr.Graphic3d_AspectLine3d()
        aWhite = qu.Quantity_Color(qu.Quantity_NOC_WHITE)
        aLineAspect.SetColor(aWhite)
        group.SetPrimitivesAspect(aLineAspect)

        self.dft_drawn = True

        self.dftPres = self.display2(self.dft,
                                     self.display,
                                     99,
                                     self.dft_draw_height,
                                     1)

        self.dftPres.Erase()
        self.display.Repaint()

    def add_dft_to_menu(self):

        self.disp['add_menu']('dft')
        self.disp['add_function_to_menu']('dft', self.show_dft)
        self.disp['add_function_to_menu']('dft', self.hide_dft)
        self.disp['add_function_to_menu']('dft', self.dft_inc)
        self.disp['add_function_to_menu']('dft', self.dft_dec)

    def show_dft(self):

        if not self.dft_drawn:
            self.make_dft_display()
            self.dft_drawn = True

        self.dftPres.Display()
        self.display.Repaint()

    def hide_dft(self):
        if self.dft_drawn:
            self.dftPres.Erase()
            self.display.Repaint()

    def dft_inc(self):

        if self.dft_draw_height != int(self.sz):
            self.dft_draw_height += 1

        if self.dft_drawn:

            self.dftPres.Clear()

            self.dftPres = self.display2(self.dft,
                                         self.display,
                                         99,
                                         self.dft_draw_height,
                                         1)
            self.dftPres.Display()
            self.display.Repaint()

    def dft_dec(self):

        if self.dft_draw_height != 0:
            self.dft_draw_height -= 1

        if self.dft_drawn:

            self.dftPres.Clear()
            self.dftPres = self.display2(self.dft,
                                         self.display,
                                         99,
                                         self.dft_draw_height,
                                         1)
            self.dftPres.Display()
            self.display.Repaint()

    def add_dft_by_val_to_menu(self):

        self.disp['add_function_to_menu']('dft', self.show_dft_by_val)
        self.disp['add_function_to_menu']('dft', self.hide_dft_by_val)
        self.disp['add_function_to_menu']('dft', self.dft_inc_by_val)
        self.disp['add_function_to_menu']('dft', self.dft_dec_by_val)

    def make_dft_by_val(self):

        self.max_dft = int(self.dft.max())
        self.dft_value = self.max_dft
        log.info(self.max_dft)
        self.dft_shells = {}
        self.dft_by_val_visible = None

    def make_dft_shell(self, val):

        log.info("Making dft_shell of value %s" % val)
        log.info(type(val))

        # ---------------
        # DEBUG
        # position_of_source = (0, 0, 0)
        # source_point = gp_Pnt(*position_of_source)
        # trs_point = source_point.Transformed(self.vox_grid.trsf_Voxel2Global)
        # dft_shell = BRepPrimAPI_MakeSphere(trs_point, val).Shell()

        # ---------------

        isolated_value_dft = np.zeros_like(self.dft)
        isolated_value_dft[self.dft == val] = 1
        isolated_value_dft, count = si_label(isolated_value_dft)

        log.info("Making dft for value %s - %s parts" % (val, count))
        log.info(isolated_value_dft.max())
        if count == 1:

            mini, origin = minimize(isolated_value_dft, 1)
            mini = mini.astype(np.int32)

            mini, origin = bigger(mini, origin)

            log.info(set(mini.flat))
            log.info(mini.sum())
            log.info(origin)
            shell = gen.make_faces(mini, origin, self.trsf_vg_matrix, 1)

        else:

            shell = TopoDS_Compound()
            builder = BRep_Builder()
            builder.MakeCompound(shell)

            for each in range(count):

                mini, origin = minimize(isolated_value_dft, each + 1)
                mini, origin = bigger(mini, origin)

                log.info(set(mini.flat))
                log.info(mini.sum())
                log.info(origin)

                mini = mini.astype(np.int32)

                part = gen.make_faces(mini, origin,
                                      self.trsf_vg_matrix, each + 1)
                builder.Add(shell, part)

        self.dft_shells[val] = ais.AIS_Shape(shell)

        self.dft_shells[val].SetMaterial(gr.Graphic3d_NOM_GLASS)

    def show_dft_by_val(self):

        if not self.dft_by_val_drawn:
            self.make_dft_by_val()
            self.dft_by_val_drawn = True

        self.dft_by_val_visible = True

        log.info("dft_value_being_shown = %s" % self.dft_value)

        shell = self.dft_shells.get(self.dft_value, None)
        if shell is None:
            self.make_dft_shell(self.dft_value)

        self.display.Context.Display(self.dft_shells[self.dft_value])
        self.dft_by_val_visible = self.dft_value

        self.display.Repaint()

    def hide_dft_by_val(self):

        if not self.dft_by_val_drawn:
            self.make_dft_by_val()
            self.dft_by_val_drawn = True

        if self.dft_by_val_visible is not None:
            shells = self.dft_shells[self.dft_by_val_visible]
            self.display.Context.Remove(shells)

        self.dft_by_val_visible = None
        self.display.Repaint()

    def dft_inc_by_val(self):

        if not self.dft_by_val_drawn:
            self.make_dft_by_val()
            self.dft_by_val_drawn = True

        if self.dft_value != self.max_dft:
            self.dft_value += 1

        if self.dft_by_val_visible is not None:
            shells = self.dft_shells[self.dft_by_val_visible]
            self.display.Context.Remove(shells)

            ais_shell = self.dft_shells.get(self.dft_value, None)
            if ais_shell is None:
                self.make_dft_shell(self.dft_value)
            self.display.Context.Display(self.dft_shells[self.dft_value])
            self.dft_by_val_visible = self.dft_value

    def dft_dec_by_val(self):

        if not self.dft_by_val_drawn:
            self.make_dft_by_val()
            self.dft_by_val_drawn = True

        if self.dft_value != 0:
            self.dft_value -= 1

        if self.dft_by_val_visible is not None:
            shells = self.dft_shells[self.dft_by_val_visible]
            self.display.Context.Remove(shells)

            ais_shell = self.dft_shells.get(self.dft_value, None)
            if ais_shell is None:
                self.make_dft_shell(self.dft_value)
            self.display.Context.Display(self.dft_shells[self.dft_value])
            self.dft_by_val_visible = self.dft_value

    def add_element_array_to_menu(self):

        self.disp['add_menu']('array')
        self.disp['add_function_to_menu']('array', self.show_array)
        self.disp['add_function_to_menu']('array', self.hide_array)
        self.disp['add_function_to_menu']('array', self.array_inc)
        self.disp['add_function_to_menu']('array', self.array_dec)

    def array_inc(self):

        if self.array_draw_height != int(self.sz):
            self.array_draw_height += 1

        if self.array_drawn:

            self.arrayPres.Clear()

            self.arrayPres = self.display2(self.array,
                                           self.display,
                                           99,
                                           self.array_draw_height,
                                           1)
            self.arrayPres.Display()
            self.display.Repaint()

    def array_dec(self):

        if self.array_draw_height != 0:
            self.array_draw_height -= 1

        if self.array_drawn:
            self.arrayPres.Clear()
            self.arrayPres = self.display2(self.array,
                                           self.display,
                                           99,
                                           self.array_draw_height,
                                           1)
            self.arrayPres.Display()
            self.display.Repaint()

    def make_array_display(self):

        self.array_drawn = True
        self.arrayPres = self.display2(self.array,
                                       self.display,
                                       99,
                                       self.array_draw_height,
                                       1)

        self.arrayPres.Erase()
        self.display.Repaint()

    def show_array(self):

        if not self.array_drawn:
            self.make_array_display()
            self.array_drawn = True

        self.arrayPres.Display()
        self.display.Repaint()

    def hide_array(self):
        if self.array_drawn:
            self.arrayPres.Erase()
            self.display.Repaint()

    def add_connect_to_menu(self):

        self.disp['add_menu']('connect')
        self.disp['add_function_to_menu']('connect', self.show_connect)
        self.disp['add_function_to_menu']('connect', self.hide_connect)

    def make_connect_display(self):

        self.connect_drawn = True

        self.connectPres = self.display2(self.connect,
                                         self.display,
                                         99,
                                         int(self.sz / 2),
                                         1)

        self.connectPres.Erase()
        self.display.Repaint()

    def show_connect(self):

        if not self.connect_drawn:
            self.make_connect_display()
            self.connect_drawn = True

        self.connectPres.Display()
        self.display.Repaint()

    def hide_connect(self):
        if self.connect_drawn:
            self.connectPres.Erase()
            self.display.Repaint()

    def add_sources_to_menu(self):

        self.disp['add_menu']('source')
        self.disp['add_function_to_menu']('source', self.show_sources)
        self.disp['add_function_to_menu']('source', self.hide_sources)

    def make_sources_display(self):

        self.source_drawn = True

        for each_source in self.sources:

            position_of_source = self.sources[each_source]['pos']

            src_point = gp_Pnt(*position_of_source)

            tr_point = src_point.Transformed(self.vox_grid.trsf_Voxel2Global)

            source_shape = BRepPrimAPI_MakeSphere(tr_point, 1).Shell()

            self.sources[each_source]['ais'] = ais.AIS_Shape(source_shape)

            self.display.Context.Display(self.sources[each_source]['ais'])

        self.display.Repaint()

    def show_sources(self):

        if not self.source_drawn:
            self.make_sources_display()

        for each_source in self.sources:
            ais_shape = self.sources[each_source].get('ais', None)

            if ais_shape is not None:
                self.display.Context.Display(ais_shape)

        self.display.Repaint()

    def hide_sources(self):

        if self.source_drawn:
            for each_source in self.sources:
                ais_shape = self.sources[each_source].get('ais', None)

                if ais_shape is not None:
                    self.display.Context.Remove(ais_shape)

        self.display.Repaint()
