# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import math
from pprint import pprint
import pickle
from collections import namedtuple
from itertools import product

import rtree.index as rt

import OCC.Core.AIS as ais
import OCC.Core.BRepBndLib as blib
import OCC.Core.Bnd as bnd
import OCC.Core.gp as gp
import OCC.Core.TopoDS as tp
import OCC.Core.Quantity as qu
import OCC.Core.Graphic3d as gr
from OCC.Core.BRepBuilderAPI import BRepBuilderAPI_Transform as brep_Transform
from OCC.Core.BRep import BRep_Builder
from OCC.Core.TopAbs import (TopAbs_COMPOUND,
                             TopAbs_COMPSOLID,
                             TopAbs_SOLID,
                             TopAbs_SHELL,
                             TopAbs_FACE,
                             TopAbs_WIRE,
                             TopAbs_EDGE,
                             TopAbs_VERTEX,
                             TopAbs_SHAPE)


from settings.load_settings import get_scene_settings
from topology import convertTopology
import modules.hull as hull
import modules.two_rooms as two_rooms
import modules.scene as scene_module
from modules.triangulate import triangulate
from trimesh import Trimesh

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)


def funky():

    pass


class Scene:

    def __init__(self, scene_no=None, update=None, load=None):

        if scene_no is None:
            scene_no = "_default"

        if not load:
            self.init(scene_no, update)
        else:
            self.load(load)

        self.features_drawn = False
        self.hull_drawn = False

    def init(self, scene_no, update=None):
        self.scene_no = scene_no
        self.s, _ = get_scene_settings(scene_no, update)
        self._elements = {}
        self._scene = None
        self.barriers = {}
        self.rooms = {}
        self.room_hash = {}
        self.barrier_list = []
        self.internals = []
        self.dist_max = 1
        self.offset = self.s.offset

    def print_scene_settings(self):
        pprint(dict(self.s._asdict()))

    def load_scene_elements(self,
                            limit=None,
                            guids=None,
                            tol=0.00005):

        if self.s.scene == 'TwinRooms':
            self._scene = two_rooms.Tworoomset(self.settings,
                                               self._elements)
            log.debug("2 room set")

        elif self.s.scene == "Broadgate":
            log.info("Broadgate Ticket Hall")
            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             offset=self.offset,
                                             guids=guids)

        elif self.s.scene == 'empty':
            self._scene = two_rooms.Empty(self._elements)
            log.info("Empty Scene")

        elif self.s.scene == "Illustration":
            log.info("Illustration")
            self._scene = two_rooms.Illustration(self._elements)

        elif self.s.scene == "Nagel_room":
            log.info("Nagel_room")
            self._scene = two_rooms.Nagel_room(self._elements)

        elif self.s.scene == "Two_spirited_penguin":
            log.info("Two spirited penguin")
            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             special="penguin",
                                             guids=["1uA34BXDX3HeiJ1MBuh_wi"])
            for n, each in enumerate(self._elements):
                if n == 1:
                    self._elements[each]['alpha'] = 0.7

        elif self.s.scene == "Dodgy_shaft":
            log.info("Dodgy_shaft")
            # self._scene = scene_module.Scene(self.settings,
            # self._elements,
            # limit = limit,
            # table = "A20191104_TEST",
            # guids = ["1X88wcVkn5bfRDjGz0x7aB"])
            where = ("WHERE a.FILENAME = "
                     "' C123-JUL-S-DMA-CR094_SH006-00004.dgn'")
            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20191104_MES",
                                             where=where)
            for n, each in enumerate(self._scene.elements):
                self._elements[each]['alpha'] = 0.90

            bad_elements = self._scene.add(table="A20191104_TEST",
                                           guids=["1X88wcVkn5bfRDjGz0x7aB"])
            for n, each in enumerate(bad_elements):
                self._elements[each]['alpha'] = 0.01
            log.info(len(bad_elements))
            self._scene.elements.append(bad_elements)

        elif self.s.scene == "MES_shaft":
            log.info("MES_shaft")
            log.info("tolerance: %s" % tol)
            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20200107_REPAIRED",
                                             tol=tol)

        elif self.s.scene == "cutaway":
            log.info("Cutaway")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             special="cutaway",
                                             table="A202008026_CUTAWAY",
                                             guids=["cutaway"])

        elif self.s.scene == "extruded":
            log.info("Displaying extruded rooms")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             special="extruded",
                                             table="B20200828_EXTRUDED",
                                             )

        elif self.s.scene == "extruded_and_MEP":
            log.info("Displaying extruded rooms")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20200108_BROADGATEMEP",
                                             )

            extruded = self._scene.add(table="B20200828_EXTRUDED")
            for n, each in enumerate(extruded):
                self._elements[each]['alpha'] = 0.9
            log.info(len(bad_elements))

        elif self.s.scene == "cutaway_voxels":
            log.info("Displaying extruded rooms")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="B20200904_CUTAVOX",
                                             guids=["0", "1", "2", "3"]
                                             )

        elif self.s.scene == "MEP":
            log.info("Displaying extruded rooms")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20200904_LIVMEPCLEAN",
                                             )

        elif self.s.scene == "MEP and cutaway":
            log.info("Displaying cutaway rooms and MEP")

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20200904_LIVMEPCLEAN",
                                             )

            self._scene.add(table="B20200904_CUTAVOX", multi_surface=True)

        elif self.s.scene == "filename":
            log.info("Displaying by filename")

            where = "WHERE a.FILENAME = '%s'" % self.s.filename
            table = self.s.table

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=self.s.limit,
                                             table=table,
                                             where=where,
                                             )

        elif self.s.scene == "spaces":
            log.info("Displaying Spaces")

            # where = "WHERE a.FILENAME = '%s'"% self.s.filename
            table = self.s.table

            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=self.s.limit,
                                             table=table,
                                             # where = where,
                                             )

        elif self.s.scene == "MES_shaft_transparent":
            log.info("MES_shaft")
            log.info("tolerance: %s" % tol)
            where = ("WHERE a.GUID != 'A1'"
                     " and a.GUID != 'd42f5dc9-0af9-4d0f-8cff-8f09227a90f8'"
                     " and a.GUID != '7e3f8596-37ec-40d2-bd05-66972d983b8e'")
            self._scene = scene_module.Scene(self.settings,
                                             self._elements,
                                             limit=limit,
                                             table="A20200107_REPAIRED",
                                             tol=tol)

            guid1 = 'd42f5dc9-0af9-4d0f-8cff-8f09227a90f8'
            guid2 = '7e3f8596-37ec-40d2-bd05-66972d983b8e'
            guid3 = "A1"
            transparent_elements = self._scene.add(table="A20200107_REPAIRED",
                                                   guids=[guid1])
            transparent_elements += self._scene.add(table="A20200107_REPAIRED",
                                                    guids=[guid2])
            transparent_elements += self._scene.add(table="A20200107_REPAIRED",
                                                    guids=[guid3])

            for n, each in enumerate(transparent_elements):
                self._elements[each]['alpha'] = 0.9
            log.info(len(transparent_elements))
            self._scene.elements.append(transparent_elements)

        else:
            log.info("Empty Scene")
            self._scene = two_rooms.Empty(self._elements)

        log.info("Number of scene elements: %s" % (len(self._elements)))

        self.selected_elements = []
        self.features = list(set(self._elements))
        log.info("Number of scene features: %s" % (len(self.features)))

    def generate_pickle(self):

        self._scene.dump(self._elements)

    def generate_voxel_scene(self, vox_grid):

        self.vox_grid = vox_grid
        self.vox_scene_elements = self.generate_voxel_scene1(self._elements,
                                                             self.vox_grid)

    def generate_voxel_scene1(self, scene_elements, voxel_group):

        voxel_scene = {}

        for each_element in scene_elements:
            if each_element in self.barriers:
                continue
            if each_element in self.rooms:
                continue
            shape = scene_elements[each_element]['shape']
            shape.ShapeType()
            bb1 = bnd.Bnd_Box()
            transform = brep_Transform(voxel_group.trsf_Global2Voxel)
            transform.Perform(shape, True)
            shape1 = transform.ModifiedShape(shape)

            if shape.ShapeType() == TopAbs_COMPOUND:
                solid = tp.topods_Compound(shape1)
            elif shape.ShapeType() == TopAbs_COMPSOLID:
                solid = tp.topods_CompSolid(shape1)
            elif shape.ShapeType() == TopAbs_SOLID:
                solid = tp.topods_Solid(shape1)
            elif shape.ShapeType() == TopAbs_SHELL:
                solid = tp.topods_Shell(shape1)
            elif shape.ShapeType() == TopAbs_FACE:
                solid = tp.topods_Face(shape1)
            else:
                raise Exception(shape.ShapeType())
            blib.brepbndlib_Add(shape1, bb1)
            cornermin = bb1.CornerMin().Coord()
            cornermax = bb1.CornerMax().Coord()
            corner_tuple = (*cornermin, *cornermax)

            voxel_scene[each_element] = {'shape': solid,
                                         'bb': bb1,
                                         'bbxyz': corner_tuple}

        return voxel_scene

    @property
    def elements(self):
        return self._elements

    @property
    def settings(self):
        return dict(self.s._asdict())

    def index(self, elements):

        p = rt.Property()
        p.dimension = 3
        idx = rt.Index(properties=p)

        for each in elements:
            if each in self.features:
                element = elements[each]
                corner_tuple = element['bbxyz']
                idx.insert(each, corner_tuple)

            # -------------
            # TODO - find which elements are not tagged with bbxyz
            # -------------

        return idx

    def index_elements(self):

        self.idx = self.index(self._elements)
        self.calc_scene_centre()

    def vox_index_elements(self):

        self.vox_idx = self.index(self.vox_scene_elements)

        # -------------
        # TODO - extend index beyond features for reloaded
        # -------------

    def bounding_box(self):

        (xmin, ymin, zmin, xmax, ymax, zmax) = self.idx.bounds

        voxel_group = self.vox_grid

        distance = int(0)

        log.debug(xmin, ymin, zmin, "-", xmax, ymax, zmax)

        if voxel_group is not None:
            if xmin > -0.5:
                xmin = -0.5

            if ymin > -0.5:
                ymin = -0.5

            if zmin > -0.5:
                zmin = -0.5

            if xmax < voxel_group.sx + 0.5:
                xmax = voxel_group.sx + 0.5

            if ymax < voxel_group.sy + 0.5:
                ymax = voxel_group.sy + 0.5

            if zmax < voxel_group.sz + 0.5:
                zmax = voxel_group.sz + 0.5

            log.debug("%s, %s, %s - %s, %s, %s)"
                      % (xmin, ymin, zmin, xmax, ymax, zmax))

            distance = int(math.ceil(math.sqrt((xmax - xmin)**2
                                               + (ymax - ymin)**2
                                               + (zmax - zmin)**2)))

            log.debug("Distance: %s" % distance)

        self.dist_max = distance

        return distance

    def gen_hull(self, voxelise=True):

        if self._elements:

            scene_elements = self._elements
            if voxelise:
                voxel_scene_elements = self.vox_scene_elements

            hull_element = hull.hull(scene_elements)
            hash_ = hash(hull_element)

            bb = bnd.Bnd_Box()
            v_bb = bnd.Bnd_Box()
            blib.brepbndlib_Add(hull_element, bb)

            if voxelise:
                inverted = self.vox_grid.trsf_Voxel2Global.Inverted()
                transform = brep_Transform(inverted)
                transform.Perform(hull_element, True)
                v_shape = transform.ModifiedShape(hull_element)
                v_solid = tp.topods_Solid(v_shape)
                blib.brepbndlib_Add(v_shape, v_bb)

            cornermin = bb.CornerMin().Coord()
            cornermax = bb.CornerMax().Coord()
            corner_tuple = (*cornermin, *cornermax)

            if voxelise:
                cornermin_v = v_bb.CornerMin().Coord()
                cornermax_v = v_bb.CornerMax().Coord()
                corner_tuple_v = (*cornermin_v, *cornermax_v)

            scene_elements[hash_] = {'name': 'hull',
                                     'ais': None,
                                     'shape': hull_element,
                                     'bb': bb,
                                     'bbxyz': corner_tuple}
            if voxelise:
                voxel_scene_elements[hash_] = {'shape': v_solid,
                                               'bb': v_bb,
                                               'bbxyz': corner_tuple_v}

            log.info("""Finished making hull""")

        else:
            hash_ = None
            log.info("""No elements no hull""")

        self.hull_hash = hash_

    def make_feature_display(self):

        features = self.features.copy()

        for each in self.features:
            if self._elements[each]['name'] == "ceiling":
                features.remove(each)

        self.display_walls(elements_to_display=features,
                           colour=None,
                           material=None)

        self.features_drawn = True
        self.display.FitAll()

    def add_features_to_menu(self):

        self.disp['add_menu']('features')
        self.disp['add_function_to_menu']('features', self.show_features)
        self.disp['add_function_to_menu']('features', self.hide_features)

    def show_features(self):

        if not self.features_drawn:
            self.make_feature_display()

        for each in self.features:
            ais_shape = self._elements[each]['ais']
            alpha = self._elements[each].get('alpha', 0)
            if ais_shape is not None:
                if alpha != 0:
                    ais_shape.SetTransparency(alpha)
                self.display.Context.Display(ais_shape)

    def hide_features(self):
        if self.features_drawn:
            for each in self.features:
                ais_shape = self._elements[each]['ais']
                log.info("ais_shape: %s" % repr(ais_shape) )
                self.display.Context.Remove(ais_shape)

    def transparent_features(self):
        if self.features_drawn:
            for each in self.features:
                ais_shape = self._elements[each]['ais']
                ais_shape.SetTransparency(0.9)
                self.display.default_drawer.SetFaceBoundaryDraw(False)
                self.display.Context.Redisplay(ais_shape)

    def make_hull_display(self):

        self.display_walls(elements_to_display=[self.hull_hash],
                           colour=None,
                           material=None)

        self.hull_drawn = True
        self.display.FitAll()

    def add_hull_to_menu(self):

        self.disp['add_menu']('hull')
        self.disp['add_function_to_menu']('hull', self.show_hull)
        self.disp['add_function_to_menu']('hull', self.hide_hull)

    def show_hull(self):

        if not self.hull_drawn:
            self.make_hull_display()

        else:
            each = self.hull_hash
            ais_shape = self._elements[each]['ais']
            self.display.Context.Display(ais_shape.GetHandle())

    def hide_hull(self):
        if self.hull_drawn:
            each = self.hull_hash
            ais_shape = self._elements[each]['ais']
            self.display.Context.Remove(ais_shape.GetHandle())

    def save(self, folder_path):

        elements_path = folder_path / (folder_path.stem + "_elems.pkl")
        elem_data_path = folder_path / (folder_path.stem + "_eldata.pkl")

        with elements_path.open('wb') as f:
            for each in self._elements:
                self._elements[each]['bb'] = None
                self._elements[each]['ais'] = None
            pickle.dump(self._elements, f)

        element_dict = {"features": self.features,
                        "barriers": self.barriers,
                        "barriers_list": self.barrier_list,
                        "hull_hash": self.hull_hash,
                        "rooms": self.rooms,
                        # "room_hash" :self.room_hash,
                        "settings": self.s._asdict(),
                        "dist_max": self.dist_max,
                        "centre": self.scene_centre,
                        "internals": self.internals,
                        "externals": self.externals}

        with elem_data_path.open('wb') as f:
            pickle.dump(element_dict, f)

    def load(self, folder_path):

        elements_path = folder_path / (folder_path.stem + "_elems.pkl")
        elem_data_path = folder_path / (folder_path.stem + "_eldata.pkl")

        with elements_path.open('rb') as f:
            elements = pickle.load(f)

        self._elements = {}
        new_element_lookup = {}

        for each in elements:
            try:
                new_hash = hash(elements[each]['shape'])
            except KeyError:
                new_hash = hash(elements[each]['solid'])

            new_element_lookup[each] = new_hash

            self._elements[new_hash] = elements[each]
            self._elements[new_hash]["hash"] = new_hash

        with elem_data_path.open('rb') as f:
            element_dict = pickle.load(f)

        features = element_dict["features"]
        barriers = element_dict["barriers"]
        barrier_list = element_dict["barriers_list"]
        hull_hash = element_dict["hull_hash"]
        rooms = element_dict["rooms"]
        self.dist_max = element_dict.get('dist_max', 99)
        self.internals = element_dict.get("internals", [])
        self.externals = element_dict.get("externals", [])
        self.scene_centre = element_dict.get("centre", (0, 0, 0))

        self.features = []
        for each in features:
            self.features.append(new_element_lookup[each])

        self.barrier_list = []
        for each in barrier_list:
            self.barrier_list.append(new_element_lookup[each])

        self.barriers = {}
        for each in barriers:
            self.barriers[new_element_lookup[each]] = barriers[each]

        self.rooms = {}
        self.room_hash = {}
        for each in rooms:

            self.rooms[each] = new_element_lookup[rooms[each]]
            self.room_hash[new_element_lookup[rooms[each]]] = each

        self.hull_hash = new_element_lookup[hull_hash]

        s = element_dict["settings"]
        self.s = namedtuple("scene_settings", s)(**s)

        self.offset = s.get("offset", (0, 0, 0))

    def return_shape_wires(self, shape):

        converted_shape = convertTopology(shape)

        wires = converted_shape.wires()

        compound = tp.TopoDS_Compound()
        builder = BRep_Builder()
        builder.MakeCompound(compound)

        for each_wire in wires:
            print(each_wire)
            builder.Add(compound, each_wire)

        return compound

    def init_display(self, display, disp, bounded_planes=True):

        self.display = display
        self.disp = disp
        self.bounded_planes = bounded_planes
        if bounded_planes:
            self.display.default_drawer.SetFaceBoundaryDraw(True)
        else:
            self.display.default_drawer.SetFaceBoundaryDraw(False)

    def add_view_point_to_menu(self, view_from, view_to):

        self.disp['add_menu']('views')
        self.disp['add_function_to_menu']('views', self.view_from_point)
        self.view_from = view_from
        self.view_to = view_to

    def view_from_point(self):

        if self.view_to == (0, 0, 0):
            view_from = gp.gp_Pnt(*tuple(map(lambda x, y: x + y,
                                             self.scene_centre,
                                             self.view_from)))
            view_to = gp.gp_Pnt(*self.scene_centre)
        else:
            view_from = gp.gp_Pnt(*self.view_from)
            view_to = gp.gp_Pnt(*self.view_to)

        self.display.ViewPoint(view_from, view_to)

    def display_walls(self,
                      elements_to_display=None,
                      colour=None,
                      material=None,
                      flung=False):

        if colour is None:
            colour = qu.Quantity_Color(qu.Quantity_NOC_LIGHTBLUE)
        black = qu.Quantity_Color(qu.Quantity_NOC_BLACK)
        if material is None:
            material = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_PEWTER)

        if elements_to_display is None:
            _elements_to_display = list(self._elements)
        else:
            _elements_to_display = elements_to_display

        for each in _elements_to_display:

            rgb = self._elements[each].get('rgb', None)
            if rgb:
                r, g, b = rgb
                color = qu.Quantity_Color(r, g, b, 0)
            else:
                color = colour

            if flung:
                s = 'flung'
            else:
                s = 'shape'
            shape = self._elements[each][s]
            ais_shape = ais.AIS_Shape(shape)

            alpha = self._elements[each].get('alpha', 0)
            ais_wire = tp.TopoDS_Shape()
            if not self.bounded_planes and alpha < 0.3:
                wires = self.return_shape_wires(shape)
                ais_wire = ais.AIS_Shape(wires)
                ais_wire.SetColor(black)
                ais_wire.SetWidth(0.01)
                self.display.Context.Display(ais_wire)
            else:
                ais_wire = ais.AIS_Shape(tp.TopoDS_Shape())

            ais_shape.SetMaterial(material)

            rgb = self._elements[each].get('rgb', None)
            if rgb:
                r, g, b = rgb
                color = qu.Quantity_Color(r, g, b, 0)
            else:
                color = colour
            ais_shape.SetColor(color)

            if alpha != 0:
                log.info(alpha)
                ais_shape.SetTransparency(alpha)
            log.info(ais_shape.Transparency())
            self.display.Context.Display(ais_shape)

            self.display.shapes[each] = {'ais': ais_shape,
                                         'name': "",
                                         'wire': ais_wire}
            self._elements[each]['ais'] = ais_shape

    def calc_scene_centre(self):

        (xmin, ymin, zmin, xmax, ymax, zmax) = self.idx.bounds

        xcen = (xmin + xmax) / 2
        ycen = (ymin + ymax) / 2
        zcen = (zmin + zmax) / 2

        self.scene_centre = (xcen, ycen, zcen)

    def calculate_centres(self):

        for each in self._elements:

            ct = self._elements[each]['bbxyz']

            mid = ((ct[0] + ct[3]) / 2,
                   (ct[1] + ct[4]) / 2,
                   (ct[2] + ct[5]) / 2)

            self._elements[each]['centre'] = mid

    def calculate_flung_objects(self):

        the_centre = gp.gp_Pnt(*self.scene_centre)

        for each in self._elements:

            shape = self._elements[each]['shape']

            cen_element = gp.gp_Pnt(*self._elements[each]['centre'])

            vector_from_centre = gp.gp_Vec(the_centre, cen_element)

            extended_vector = vector_from_centre.Scaled(1)

            translate = gp.gp_Trsf()
            translate.SetTranslation(extended_vector)

            transform = brep_Transform(translate)
            transform.Perform(shape, True)
            shape1 = transform.ModifiedShape(shape)
            self._elements[each]['flung'] = shape1

    def check_node_array(self, node_array):

        a = list(set([tuple(each) for each in list(node_array)]))

        min_dist = float("inf")
        for x, y in product(a, a):

            if x == y:
                continue
            dist = (x[0] - y[0])**2 + (x[1] - y[1])**2 + (x[2] - y[2])**2
            if dist < min_dist:
                min_dist = dist

        log.info(math.sqrt(min_dist))

        return math.sqrt(min_dist)

    def calculate_meshes(self, elements, precision=1e-7):

        log.info("meshing")

        for each in elements:

            node_array, face_list = triangulate(self._elements[each]['shape'],
                                                precision)

            self._elements[each]['mesh'] = Trimesh(node_array,
                                                   face_list,
                                                   validate=True)

    def set_colors(self):
        count = 0
        for each in self._elements:
            if not self._elements[each]['mesh'].is_watertight:
                self._elements[each]['rgb'] = (1, 0, 0)
                count += 1
        log.info("Non-watertight elements:%s" % count)

        count = 0
        for each in self._elements:
            if self._elements[each]['mesh'].contain_degenerate_faces():
                self._elements[each]['rgb'] = (1, 0, 1)
                count += 1
        log.info("Euler-not-2 elements:%s" % count)

