# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

from itertools import product

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

import shapely.geometry as sg
from descartes import PolygonPatch

from modules.form_solid_from_voxels import minimize, bigger


def rgb2hex(r, g, b):

    rgb = '#%02X%02X%02X' % (int(r * 255), int(g * 255), int(b * 255))
    return rgb


def format_axes(ax, sx, sy, ticks=True):

    # ax.grid(False, color="gray", linewidth = .3)
    ax.set_axisbelow(True)
    ax.set_aspect(1)

    if not ticks:
        ax.set_xlim([0, sx])
        ax.set_ylim([0, sy])
        ax.tick_params(
            axis='both',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,  # ticks along the top edge are off
            right=False,
            left=False,
            labelleft=False,
            labelbottom=False)

    else:
        ax.set_xticks(range(0, sx + 1))
        ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
        ax.xaxis.set_major_formatter(ticker.NullFormatter())
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(0.5))
        form = ticker.FuncFormatter(lambda x, pos: "" if pos % 2 else "%d" % x)
        ax.xaxis.set_minor_formatter(form)
        for tick in ax.xaxis.get_minor_ticks():
            tick.tick1line.set_markersize(0)
            tick.tick2line.set_markersize(0)
            tick.label1.set_horizontalalignment('center')

        ax.set_yticks(range(0, sy + 1))
        ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
        ax.yaxis.set_major_formatter(ticker.NullFormatter())
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(0.5))
        form = ticker.FuncFormatter(lambda y, pos: "" if pos % 2 else "%d" % y)
        ax.yaxis.set_minor_formatter(form)
        for tick in ax.yaxis.get_minor_ticks():
            tick.tick1line.set_markersize(0)
            tick.tick2line.set_markersize(0)
            tick.label1.set_verticalalignment('center')


def make_voxel_patch(x, y, kwargs_):

    corners = np.array(((0, 0),
                        (0, 1),
                        (1, 1),
                        (1, 0)))

    corners = (corners + (x, y))
    polygon = sg.Polygon(corners)
    polygon_patch = PolygonPatch(polygon, **kwargs_)
    return polygon_patch


def process_side_view(array, axis):

    if axis == 1:
        array = np.transpose(array, (0, 1, 2))
        sx, sy, sz = array.shape
        rng = range(sz - 1, -1, -1)
    if axis == 2:
        array = np.transpose(array, (0, 2, 1))
        sx, sy, sz = array.shape
        rng = range(sz)
    elif axis == 3:
        array = np.transpose(array, (2, 1, 0))
        sx, sy, sz = array.shape
        rng = range(sz - 1, -1, -1)

    array2 = np.zeros((sx, sy), dtype=array.dtype)
    for x, y in product(range(sx), range(sy)):
        for z in rng:
            val = array[x, y, z]
            if val != 0:
                array2[x, y] = val
                break

    return array2


def view_illust(scene, vox_grid, vox_group, filepath1=None, filepath2=None):

    fig1: plt.Figure = plt.figure(figsize=(10, 10))
    ax1: plt.Axes = fig1.add_subplot(1, 1, 1)
    fig2: plt.Figure = plt.figure(figsize=(10, 10))
    ax2: plt.Axes = fig2.add_subplot(1, 1, 1)

    # colors = [
    #     '#9D0142',
    #     '#D53D4E',
    #     '#F46C43',
    #     '#FCAD61',
    #     '#FDDF8B',
    #     '#E5F497',
    #     '#ABDDA3',
    #     '#66C1A4',
    #     '#3187BC',
    #     '#5D4EA2']

    # cmap = LinearSegmentedColormap.from_list("mine", colors, N=256)
    # norm = Normalize(vmin=s.vmin, vmax=s.vmax)
    # sm = cm.ScalarMappable(norm=norm, cmap=cmap)

    internals = vox_group.connect.copy()

    # Hide external spaces 
    for each in scene.externals:
        internals[internals == each] = 0

    # clip array to internal spaces and expand by one voxel
    internals, origin = minimize(internals)
    internals, origin = bigger(internals, origin)

    # process plan view
    array = process_side_view(internals, 1)
    sx, sy = array.shape

    format_axes(ax1, sx, sy, ticks=False)
    ax1.set_axis_off()

    if vox_grid.voxel_scale_factor == 1:
        linewidth = 0.25
    else:
        linewidth = 0.25

    for x, y in product(range(sx), range(sy)):
        val = array[x, y]

        if val == 0:
            color = "#FFFFFF"
            continue
        else:
            r, g, b = scene._elements[scene.rooms[val]].get('rgb', (0, 0, 0))
            color = rgb2hex(r, g, b)

        kwargs_ = {"facecolor": color,
                   "linewidth": linewidth,
                   }
        patch = make_voxel_patch(x, y, kwargs_)
        ax1.add_patch(patch)

    # process side view 
    array = process_side_view(internals, 2)
    sx, sy = array.shape

    format_axes(ax2, sx, sy, ticks=False)
    ax2.set_axis_off()

    for x, y in product(range(sx), range(sy)):
        val = array[x, y]
        if val == 0:
            color = "#FFFFFF"
            continue
        else:
            # color = colors[(array[x,y] % len(colors))-1]
            r, g, b = scene._elements[scene.rooms[val]].get('rgb', (0, 0, 0))
            color = rgb2hex(r, g, b)

        kwargs_ = {"facecolor": color,
                   'linewidth': linewidth}
        patch = make_voxel_patch(x, y, kwargs_)
        ax2.add_patch(patch)

    if filepath1 and filepath2 is not None:
        # .gca().xaxis.set_major_locator(plt.NullLocator())
        plt.gca().yaxis.set_major_locator(plt.NullLocator())
        pdf_path1 = str(filepath1.with_suffix(".pdf"))
        pdf_path2 = str(filepath2.with_suffix(".pdf"))
        svg_path1 = str(filepath1.with_suffix(".svg"))
        svg_path2 = str(filepath2.with_suffix(".svg"))
        png_path1 = str(filepath1.with_suffix(".png"))
        png_path2 = str(filepath2.with_suffix(".png"))
        fig1.savefig(pdf_path1,
                     format="pdf", bbox_inches="tight", pad_inches=0)
        fig2.savefig(pdf_path2,
                     format="pdf", bbox_inches="tight", pad_inches=0)
        fig1.savefig(svg_path1,
                     format="svg", bbox_inches="tight", pad_inches=0)
        fig2.savefig(svg_path2,
                     format="svg", bbox_inches="tight", pad_inches=0)
        fig1.savefig(png_path1,
                     format="png", bbox_inches="tight", pad_inches=0)
        fig2.savefig(png_path2,
                     format="png", bbox_inches="tight", pad_inches=0)
    plt.show()
