# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import random
import operator
from pprint import pprint
import pickle

import numpy as np
from scipy.ndimage.morphology import binary_dilation

from shapely.geometry import Polygon

from itertools import product

import networkx as nx

from OCC.Core.gp import gp_Pnt, gp_Dir, gp_Pln, gp_Ax2, gp_Vec, gp_Trsf
from OCC.Core.BRepLib import BRepLib_MakeFace
from OCC.Core.BRepBuilderAPI import (BRepBuilderAPI_Sewing,
                                     BRepBuilderAPI_MakeSolid,
                                     BRepBuilderAPI_MakeFace,
                                     BRepBuilderAPI_MakePolygon)
from OCC.Core.BRepBuilderAPI import BRepBuilderAPI_Transform as brep_Transform

from OCC.Core.BRepPrimAPI import (BRepPrimAPI_MakeSphere,
                                  BRepPrimAPI_MakeCylinder)

from OCC.Core.TopoDS import (TopoDS_Compound,
                             TopoDS_Iterator,
                             topods_Wire,
                             topods_Solid,
                             topods)

from OCC.Core.BRep import BRep_Builder
from OCCUtils.Topology import WireExplorer
from OCC.Core.BRep import BRep_Tool

import OCC.Core.Quantity as qu

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)

adjacencies = ((-1, 0, 0),
               (1, 0, 0),
               (0, -1, 0),
               (0, 1, 0),
               (0, 0, -1),
               (0, 0, 1))


def rgb(color):

    col = qu.Quantity_Color(color)
    r = col.Red()
    g = col.Green()
    b = col.Blue()

    return (r, g, b)


lightblue = rgb(qu.Quantity_NOC_LIGHTBLUE)
orange = rgb(qu.Quantity_NOC_ORANGE)
olive = rgb(qu.Quantity_NOC_DARKOLIVEGREEN)
purple = rgb(qu.Quantity_NOC_PURPLE)
red = rgb(qu.Quantity_NOC_RED3)
yellow = rgb(qu.Quantity_NOC_YELLOW)
blue = rgb(qu.Quantity_NOC_BLUE1)
black = rgb(qu.Quantity_NOC_BLACK)

colors = [olive, purple, red, yellow, blue]


def make_face(x, y, z, dx, dy, dz, quartered=False):

    if quartered:
        q = 4
        o = -1.5
    else:
        q = 1
        o = 0

    norm = gp_Dir(dx, dy, dz)
    pt = gp_Pnt((x + o + dx / 2) / q,
                (y + o + dy / 2) / q,
                (z + o + dz / 2) / q)
    pln = gp_Pln(pt, norm)
    face = BRepLib_MakeFace(pln, -0.5 / q, 0.5 / q, -0.5 / q, 0.5 / q).Face()
    return face


def make_verts(face,
               G: nx.MultiDiGraph,
               value,
               dx, dy, dz,
               internal_face,
               inner):

    face_dict = G.graph['faces']

    it = TopoDS_Iterator(face)
    wire = it.Value()
    wire_ = WireExplorer(topods_Wire(wire))
    coords = []
    for each in wire_.ordered_vertices():
        coord = BRep_Tool.Pnt(each).Coord()
        coords.append(coord)

    for n, each in enumerate(coords):
        start = coords[n]
        end = coords[(n + 1) % 4]
        nxt = coords[(n + 2) % 4]
        G.add_edge(start, end, value, nxt=nxt, internal=internal_face)
        G.nodes[start]['cat'] = 6
        G.nodes[start].setdefault('faces', []).append(hash(face))
        G.nodes[start]['pos'] = start
        G.nodes[start].setdefault("value", set([])).add(value)

    G.edges[coords[-1], coords[0], value]['last'] = True
    G.edges[coords[0], coords[1], value]['start'] = True

    face_dict[(*coords[0], *coords[1], value)] = {"value": value,
                                                  "internal": internal_face,
                                                  "normal_x": dx,
                                                  "normal_y": dy,
                                                  "normal_z": dz}
    return


def quartered(array):

    sx, sy, sz = array.shape
    qx = sx * 4
    qy = sy * 4
    qz = sz * 4

    q_array = np.zeros((qx, qy, qz), array.dtype)
    for x, y, z in product(range(sx), range(sy), range(sz)):
        q_array[x * 4:(x * 4 + 4),
                y * 4:(y * 4 + 4),
                z * 4:(z * 4 + 4)] = array[x, y, z]
    return q_array


def process(array, G, quartered=False):

    sx, sy, sz = array.shape

    compound = TopoDS_Compound()
    builder = BRep_Builder()
    builder.MakeCompound(compound)

    for x, y, z in product(range(sx), range(sy), range(sz)):
        for dx, dy, dz in adjacencies:
            if array[x, y, z] == 0:
                continue
            if x + dx in [-1, sx] or y + dy in [-1, sy] or z + dz in [-1, sz]:
                raise Exception("exceeding boundaries")
            elif array[x, y, z] == array[x + dx, y + dy, z + dz]:
                continue
            elif array[x + dx, y + dy, z + dz] > 0:
                internal_face = True
            else:
                internal_face = False

            face = make_face(x, y, z, dx, dy, dz, quartered)
            builder.Add(compound, face)
            value = array[x, y, z]
            make_verts(face, G, value, dx, dy, dz, internal_face,
                       not quartered)

    return compound


def from_axis(coord, axis):

    if axis == "z":
        return (coord[0], coord[1])
    elif axis == "x":
        return (coord[1], coord[2])
    elif axis == "y":
        return (coord[2], coord[0])
    else:
        raise Exception


def to_axis(coord, z, axis):

    if axis == "z":
        return (coord[0], coord[1], z)
    elif axis == "x":
        return (z, coord[0], coord[1])
    elif axis == "y":
        return (coord[1], z, coord[0])
    else:
        raise Exception


def add_edges(G, H):

    for each_node in list(G.nodes):
        x, y, z = each_node
        for dx, dy, dz in product(range(-1, 2), range(-1, 2), range(-1, 2)):
            if dx == dy == dz == 0:
                continue
            v = (x + dx / 4, y + dy / 4, z + dz / 4)
            if v in H.nodes:
                H.add_edge(each_node, v, 0)
                H.nodes[each_node]['cat'] = 7
                H.nodes[each_node].setdefault("parents", []).append(each_node)
                H.nodes[v]['value'] = G.nodes[each_node]['value']
                current_status = H.nodes[v].get('cat', 6)
                status = abs(dx) + abs(dy) + abs(dz)
                if status < current_status:
                    H.nodes[v]['cat'] = status
                    H.nodes[v]['parent'] = each_node


def make_cylinder(u, v):

    vec = gp_Vec(gp_Pnt(*u), gp_Pnt(*v))
    length = vec.Magnitude()
    try:
        z_direction = gp_Dir(vec)
    except RuntimeError:
        return None
    axis = gp_Ax2(gp_Pnt(*u), z_direction)
    radius = 0.02
    edge = BRepPrimAPI_MakeCylinder(axis, radius, length).Shell()

    return edge


def make_wires(G):

    compound = TopoDS_Compound()
    builder = BRep_Builder()
    builder.MakeCompound(compound)
    for u, v, val in G.edges:
        vec = gp_Vec(gp_Pnt(*u), gp_Pnt(*v))
        length = vec.Magnitude()
        z_direction = gp_Dir(vec)
        axis = gp_Ax2(gp_Pnt(*u), z_direction)
        radius = 0.01
        edge = BRepPrimAPI_MakeCylinder(axis, radius, length).Shell()
        builder.Add(compound, edge)
    return compound


def make_nodes(G, filter_="dummy", value=False):

    compound = TopoDS_Compound()
    builder = BRep_Builder()
    builder.MakeCompound(compound)
    count = 0
    for each_node in G.nodes:
        if G.nodes[each_node].get(filter_, False) == value:
            # node_pos = G.nodes[each_node]['pos']
            node = BRepPrimAPI_MakeSphere(gp_Pnt(*each_node), 0.04).Shell()
            builder.Add(compound, node)
            count += 1
    return compound


def node_type(H: nx.MultiDiGraph, node, debug=False):

    direct = np.zeros((3), np.float)
    count = 0
    for each_node in H.adj[node]:
        direct += each_node
        direct -= node
        count += 1

    if count == 6:
        return 3  # internal corner case

    elif count == 5:
        internal_node = H.nodes[node]['parent']
        int_vec = tuple(map(operator.sub, internal_node, node))
        if np.dot(int_vec, direct) > 0:
            return 4  # peculiar corner case 1
        else:
            H.nodes[node]['normal'] = direct
            return 5  # peculiar corner case 2

    elif np.all(direct == 0):
        return 6  # face case

    elif np.all(np.abs(direct)):
        return 3  # external corner case

    else:
        return 2  # edge case  (note 6 is an non-secondary edge)


def cat_nodes(H):

    for each_node in list(H.nodes):
        cat = H.nodes[each_node].get('cat', 6)
        if cat == 6:
            continue

        elif cat == 5:
            continue

        elif cat == 4:
            continue

        elif cat == 3:
            type_ = node_type(H, each_node)

            if type_ == 3:
                pass
            elif type_ == 2:
                H.nodes[each_node]['cat'] = 6
            else:
                H.nodes[each_node]['cat'] = type_
        elif cat == 2:
            type_ = node_type(H, each_node)
            if type_ != 2:
                H.nodes[each_node]['cat'] = type_


def make_face_from_nodes(M,
                         face_dict_M,
                         nodes,
                         value,
                         H,
                         triangulate=False,
                         add_pos=True):

    normal = np.cross(np.array(nodes[1]) - np.array(nodes[0]),
                      np.array(nodes[2]) - np.array(nodes[1]))
    normal = normal / np.linalg.norm(normal)

    if triangulate:
        reverse = False
        absolute = np.abs(normal)
        if absolute[0] > 0:
            plane = nodes[0][0]
            axis = "x"
            if normal[0] < 0:
                reverse = True
        if absolute[1] > 0:
            plane = nodes[0][1]
            axis = "y"
            if normal[1] < 0:
                reverse = True
        if absolute[2] > 0:
            plane = nodes[0][2]
            axis = "z"
            if normal[2] < 0:
                reverse = True

        polygon = Polygon([from_axis(each, axis) for each in nodes])
        triangles = triangulate(polygon)

        for n, each in enumerate(triangles):
            if not polygon.contains(each):
                continue
            coords = list(each.exterior.coords)
            first = to_axis(coords[0], plane, axis)
            second = to_axis(coords[1], plane, axis)
            third = to_axis(coords[2], plane, axis)

            if reverse:
                second, third = third, second

            M.add_edge(first, second, value, nxt=third, internal=False)
            M.add_edge(second, third, value, nxt=first, internal=False)
            M.add_edge(third, first, value, nxt=second, last=True,
                       internal=False)

            id_ = (*first, *second, value)
            face_dict_M[id_] = {"normal_x": normal[0],
                                "normal_y": normal[1],
                                "normal_z": normal[2],
                                "value": value}

    else:
        m = len(nodes)
        for n, each in enumerate(nodes[0:-1]):
            M.add_edge(each,
                       nodes[(n + 1) % m],
                       value,
                       nxt=nodes[(n + 2) % m],
                       internal=False)
        M.edges[nodes[-2], nodes[0], value]['last'] = True
        id_ = (*nodes[0], *nodes[1], value)

        face_dict_M[id_] = {"normal_x": normal[0],
                            "normal_y": normal[1],
                            "normal_z": normal[2],
                            "value": value}

    for each in nodes:
        if add_pos:
            M.nodes[each]['pos'] = each
        source_node = H.nodes.get(each, {})
        M.nodes[each]['parent'] = source_node.get('parent', None)
        M.nodes[each]['parents'] = source_node.get('parents', None)
        M.nodes[each]['value'] = source_node.get('value', None)
        M.nodes[each]['cat'] = source_node.get('cat', None)


def determine_edge(H, first_node, second_node, value, first=True):

    if len(second_node) == 2:
        second_node, next_node = second_node
        return second_node, next_node, True

    try:
        next_node = H.edges[first_node, second_node, value]['nxt']
    except KeyError:
        raise Exception("Oopsy Daisy")

    vec1 = tuple(map(operator.sub, second_node, first_node))
    vec2 = tuple(map(operator.sub, next_node, second_node))

    if not first:
        if H.nodes[next_node]['cat'] == 4:
            second_node = next_node
            next_node = tuple(map(operator.add, second_node, vec2))
            return second_node, next_node, False

    if H.edges[first_node, second_node, value].get('done', False) is True:
        return None, None, False

    H.edges[first_node, second_node, value]['done'] = True

    count = 1
    while count < 9:
        try:
            second_node_cat = H.nodes[second_node]['cat']
        except KeyError:
            raise Exception("oops, oops")

        if second_node_cat == 5:
            normal = H.nodes[second_node]['normal']
            cross = np.cross(vec1, vec2)
            if np.dot(cross, normal) != 0:
                next_node = tuple(map(operator.sub, second_node, vec2))
                return second_node, next_node, False
            else:
                break

        elif second_node_cat == 4:
            next_node = tuple(map(operator.add, second_node, vec2))
            check_node = tuple(map(operator.add, next_node, vec1))
            if H.nodes.get(check_node, None) is None:
                break
            else:
                return second_node, (check_node, next_node), False

        elif second_node_cat < 4:
            break

        second_node = tuple(map(operator.add, second_node, vec1))
        count += 1

    if count == 9:
        raise Exception("Hmmm!, more than 6 counts")

    next_node = tuple(map(operator.add, second_node, vec2))

    return second_node, next_node, False


def determine_nodes(H, first_node, second_node, value):

    this_node = first_node
    next_node = second_node
    count = 1
    nodes = [this_node]
    disable = True

    x = 13
    while (this_node != first_node or count == 1) and count < x:
        this_node, next_node, disable = determine_edge(H,
                                                       this_node,
                                                       next_node,
                                                       value,
                                                       first=disable)
        if this_node is None:
            return None

        nodes.append(this_node)

        count += 1

    if count == x:
        raise Exception("Exceeded edge count", )

    return nodes


def face_cull(H: nx.MultiDiGraph, M: nx.MultiDiGraph):

    # face_dict_H = H.graph['faces']
    face_dict_M = M.graph['faces']

    for each_node in H.nodes:
        if H.nodes[each_node]['cat'] in [1, 2, 3]:
            for each_adjacent_node in H.adj[each_node]:
                for value in H.adj[each_node][each_adjacent_node]:
                    if H.nodes[each_adjacent_node] == 7:
                        continue
                    # normal = np.cross(each_node, each_adjacent_node)
                    nodes = determine_nodes(H,
                                            each_node,
                                            each_adjacent_node,
                                            value)

                    if nodes is not None:
                        values = H.nodes[nodes[0]]['value'].copy()
                        for n in range(len(nodes) - 1):
                            t = H.nodes[nodes[n + 1]]['value']
                            values.intersection_update(t)
                        value = values.pop()
                        make_face_from_nodes(M, face_dict_M, nodes, value, H)

    return


def expand(M, f):

    for each in M.nodes:

        pos = M.nodes[each]['pos']
        parent = M.nodes[each]['parent']
        vec = tuple(map(operator.sub, pos, parent))
        vec = tuple(map(operator.mul, vec, (f, f, f)))
        new_pos = tuple(map(operator.add, parent, vec))
        M.nodes[each]['pos'] = new_pos


def make_faces(M, scene=None, filter_="dummy", val=False, negate=False,
               value_list=None,
               debug=False):

    rooms = {}
    room_hash = {}
    elements = {}

    faces_M = M.graph["faces"]

    if value_list is None:
        value_list = []

    value_list.append(0)

    value_collection = {}

    log.info("NUmber of faces %s" % len(faces_M))

    for each_face in faces_M:

        value = faces_M[each_face]['value']
        if not isinstance(value, (int, np.int64, np.int32)):
            value = 0
        if not value:
            continue
        value_collection.setdefault(value, []).append(each_face)

    for each_value in value_collection:
        log.info("room no %s" % each_value)
        if not debug:
            sew = BRepBuilderAPI_Sewing()
        for each_face in value_collection[each_value]:

            color = colors[each_value % len(colors)]
            poly = BRepBuilderAPI_MakePolygon()
            ux, uy, uz, vx, vy, vz, value = each_face

            u = (ux, uy, uz)
            v = (vx, vy, vz)

            poly.Add(gp_Pnt(*M.nodes[u]['pos']))
            poly.Add(gp_Pnt(*M.nodes[v]['pos']))
            count = 2
            try:
                while M.edges[u, v, value].get("last", False) is False:
                    next_node = M.edges[u, v, value]['nxt']
                    try:
                        pos = M.nodes[next_node]['pos']
                        poly.Add(gp_Pnt(*pos))
                    except KeyError:
                        poly.Add(gp_Pnt(*next_node))
                        # dud_node = BRepPrimAPI_MakeSphere(gp_Pnt(*next_node),
                        #                                   0.04).Shell()
                        # scene._elements[next_node] = {"shape": dud_node,
                        #                               "rgb": red}
                        # scene.features.append(next_node)
                    u = v
                    v = next_node
                    count += 1
            except KeyError:
                log.info(M.get_edge_data(u, v))
                continue
            except:
                log.info(M.get_edge_data(u, v))
                continue

            try:
                poly.Close()
            except RuntimeError:
                log.info("Unable to close polygon")
                continue
            face = BRepBuilderAPI_MakeFace(poly.Wire()).Face()
            if debug:
                scene._elements[each_face] = {"shape": face, "rgb": color}
                scene.features.append(each_face)
            else:
                sew.Add(face)

        if not debug:
            sew.Perform()
            shape = sew.SewedShape()
            shell = topods.Shell(shape)
            solid = BRepBuilderAPI_MakeSolid(shell).Solid()

            elements[hash(solid)] = {"shape": solid,
                                     "rgb": color,
                                     "hash": hash(solid)}
            rooms[each_value] = hash(solid)
            room_hash[hash(solid)] = each_value

    return rooms, room_hash, elements


def make_border_nodes(G: nx.MultiDiGraph):

    intrnal = "internal"

    G.graph['border_nodes'] = {}
    G.graph['values'] = set([])

    for each_node in G.nodes:
        if len(G.nodes[each_node].get('value', [])) > 1:
            border_nodes = {key: {} for key in G.nodes[each_node]['value']}
            G.graph['border_nodes'][each_node] = border_nodes
            G.nodes[each_node]['border'] = True
            G.graph['values'].update(G.nodes[each_node]['value'])

    H = {}
    for each in G.graph['values']:
        H[each] = nx.DiGraph()

    for each_node in G.graph['border_nodes']:
        for each_value in G.graph['border_nodes'][each_node]:
            for each_adj in G.adj[each_node]:
                if G.get_edge_data(each_node, each_adj, each_value):
                    reciprical = G.get_edge_data(each_adj, each_node)
                    if reciprical:
                        if len(reciprical) > 1:
                            outer = False
                            for each_recip in reciprical:
                                if reciprical[each_recip].get(intrnal, False):
                                    pass
                                else:
                                    outer = True
                            if not outer:
                                continue
                        elif each_value == list(reciprical)[0]:
                            continue
                        next_ = {'next': each_adj}
                        G.graph['border_nodes'][each_node][each_value] = next_
                        H[each_value].add_edge(each_node, each_adj)

    G.graph["collect"] = H


def find_border_starts(G: nx.MultiDiGraph):

    H = G.graph["collect"]

    for each_value in H:
        M: nx.DiGraph = H[each_value]
        M.graph['start_nodes'] = []
        for each in nx.simple_cycles(M):
            M.graph['start_nodes'].append(each[0])


def make_gap_faces(G, K, R):

    face_dict_R = R.graph['faces']

    make_border_nodes(G)
    make_border_nodes(K)
    find_border_starts(K)

    for each_value in K.graph['collect']:
        M: nx.DiGraph = K.graph['collect'][each_value]
        # N = G.graph['collect'][each_value]
        count = 0
        for each_start in M.graph['start_nodes']:
            first_node = each_start
            while (count < 1000):
                nodes = [first_node]
                second_node = next(M.predecessors(first_node))
                nodes.append(second_node)
                last_node = K.nodes[first_node]['parent']
                third_node = K.nodes[second_node]['parent']
                nodes.append(third_node)

                if third_node != last_node:
                    nodes.append(last_node)

                nodes.append(first_node)
                make_face_from_nodes(R, face_dict_R, nodes, each_value, K,
                                     add_pos=False)
                R.nodes[third_node]['pos'] = third_node
                R.nodes[last_node]['pos'] = last_node

                first_node = second_node
                if first_node == each_start:
                    break
                count += 1
            if count == 1000:
                raise Exception("Probably not manifold")


def make_border_edges_(G: nx.MultiDiGraph, scene):

    make_border_nodes(G)

    H = G.graph["collect"]
    n = random.randint(0, 1000)
    for each_value in H:
        compound = TopoDS_Compound()
        builder = BRep_Builder()
        builder.MakeCompound(compound)

        M = H[each_value]

        for each_edge in M.edges:
            u, v = each_edge
            cyl = make_cylinder(G.nodes[u]['pos'],
                                G.nodes[v]['pos'])
            if cyl is not None:
                builder.Add(compound, cyl)

        scene._elements["border%s_%s" % (each_value, n)] = {"shape": compound,
                                                            "rgb": black}
        scene.features.append("border%s_%s" % (each_value, n))


def copy_faces(G: nx.MultiDiGraph,
               R: nx.MultiDiGraph,
               K: nx.MultiDiGraph):

    face_dict_G = G.graph['faces'].copy()
    face_dict_R = R.graph['faces']
    face_dict_K = K.graph['faces'].copy()

    for u, v, val in G.edges:
        x = G.get_edge_data(u, v, val)
        if x.get('internal', False):
            R.add_edge(u, v, val, **x)
            R.nodes[u]['pos'] = G.nodes[u]['pos']
            R.nodes[v]['pos'] = G.nodes[v]['pos']

            if x.get('start', False):
                face_dict_R[(*u, *v, val)] = face_dict_G[(*u, *v, val)]

    R = nx.compose(K, R)
    R.graph['faces'].update(face_dict_K)

    return R


class Process:

    def __init__(self, array, filepath=None):

        if filepath is None:

            self.G = nx.MultiDiGraph()
            self.H = nx.MultiDiGraph()
            self.K = nx.MultiDiGraph()
            self.R = nx.MultiDiGraph()

            self.G.graph['faces'] = {}
            self.H.graph['faces'] = {}
            self.K.graph['faces'] = {}
            self.R.graph['faces'] = {}

            self.array = array
        else:
            with filepath.open("rb") as f:
                self.R = pickle.load(f)

    def process(self, expansion=1):

        log.info("copy the array into an array with 4 times the resolution")
        self.q_array = quartered(self.array)

        log.info("dilate the quartered array by one voxel")
        d = np.ones((3, 3, 3), bool)
        self.dilated = binary_dilation(self.q_array, d)

        log.info("convert the original array into a graph structure")
        process(self.array, self.G)

        log.info("convert the dilated array into a graph structure")
        process(self.dilated, self.H, quartered=True)

        log.info("add edges between nodes in original voxel graph and "
                 "dilated voxel graph")
        add_edges(self.G, self.H)

        log.info("classify nodes in dilated voxel graph")
        cat_nodes(self.H)

        log.info("identify faces in dilated voxel graph and "
                 "make a new voxel graph")
        face_cull(self.H, self.K)

        # Original vector is 0.25 of voxel cube side
        # Expansion factor multiplies this 0.25 vector
        # Maximum safe factor is 1.99
        expand(self.K, expansion)

        self.R = copy_faces(self.G, self.R, self.K)

        # log.info("identify boundaries in dilated voxel_graph")
        # make_border_edges_(self.G, self.scene)
        # make_border_edges_(self.K, self.scene)

        log.info("make faces that fill gap between inner and outer shells")
        make_gap_faces(self.G, self.K, self.R)

    def save(self, filepath):

        with filepath.open("wb") as f:
            pickle.dump(self.R, f)

    def make_faces(self):
        solids, solid_hash, elements = make_faces(self.R)

        return solids, solid_hash, elements


def solidify(mini, origin, trsf_VG, expansion, filepath=None, load=False):

    if filepath is not None and load is False:
        save = True
    else:
        save = False

    if filepath is not None and load is True:
        load = True
    else:
        load = False

    if load is False:
        p = Process(mini)
        log.info(p.G.nodes)
        p.process(expansion)
    else:
        p = Process(None, filepath)

    if save:
        p.save(filepath)

    rooms, room_hash, elements = p.make_faces()

    log.info(rooms)
    log.info(room_hash)

    pprint(elements)
    # log.info("elements %s"%repr(elements))

    hash_ = list(elements)[0]
    solid = elements[hash_]['shape']
    # log.info("type %s"%(type(solid)))

    vec = gp_Vec(*origin)
    # print(vec, flush=True)
    translate = gp_Trsf()
    translate.SetTranslation(vec)

    # print(translate, flush=True)

    for each in rooms:
        log.info("room: %s" % each)
        shape = elements[rooms[each]]['shape']
        first_transform = brep_Transform(translate)
        # print(first_transform, flush=True)
        first_transform.Perform(shape, True)
        shape1 = first_transform.ModifiedShape(shape)
        # print(shape1, flush=True)
        transform = brep_Transform(trsf_VG)

        # print(transform, flush=True)
        transform.Perform(shape1, True)
        shape2 = transform.ModifiedShape(shape1)
        # print(shape2, flush=True)

        solid = topods_Solid(shape2)
        # print(solid, flush=True)
        elements[rooms[each]]['shape'] = solid

    return rooms, room_hash, elements
