# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import numpy as np
import OCC.Core.gp as gp

# Local application imports

import OCC.Core.Quantity as qu
import OCC.Core.Graphic3d as gr

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)


def rgb2hex(r, g, b):

    rgb = '#%02X%02X%02X' % (int(r * 255), int(g * 255), int(b * 255))
    return rgb


color_pallete = np.array(
    [[0.61960784, 0.00392157, 0.25882353],
     [0.83529412, 0.24313725, 0.30980392],
     [0.95686275, 0.42745098, 0.2627451],
     [0.99215686, 0.68235294, 0.38039216],
     [0.99607843, 0.87843137, 0.54509804],
     [0.90196078, 0.96078431, 0.59607843],
     [0.67058824, 0.86666667, 0.64313725],
     [0.4, 0.76078431, 0.64705882],
     [0.19607843, 0.53333333, 0.74117647],
     [0.36862745, 0.30980392, 0.63529412]])

def apply_colour(elements, rooms, special_cases=None):

    for n, each_room in enumerate(rooms):
        r, g, b = color_pallete[n % color_pallete.shape[0]]
        elements[rooms[each_room]].update({'rgb': (r, g, b)})

    if special_cases is not None:
        for each_room in special_cases:
            room_hash = rooms.get(each_room, None)
            if room_hash is not None:
                elements[room_hash].update({'rgb': special_cases[each_room]})


class Displayed_Rooms(object):

    def __init__(self, display, disp, scene_elements, rooms, barriers=None):

        self.display = display
        self.disp = disp
        self.elements = scene_elements
        self.rooms = rooms
        if barriers is None:
            barriers = []
        self.barriers = barriers
        self.display.register_select_callback(self.funcky)
        self.qu_colors = []
        self.palette = []

        for each in color_pallete[:, :]:

            self.qu_colors.append(qu.Quantity_Color(each[0],
                                                    each[1],
                                                    each[2],
                                                    qu.Quantity_TOC_RGB))
            self.palette.append(rgb2hex(each[0], each[1], each[2]))

    def funcky(self, a, x, y):

        for each in a:
            if hash(each) in self.hash_table:
                row_id = self.hash_table[hash(each)]['row']
                self.disp['win'].Widg.toggleRowId(row_id)
            elif hash(each) in self.barriers:
                self.remove_barrier(hash(each))

    def display_rooms(self):

        self.table = {}
        self.hash_table = {}
        for n, each_room in enumerate(self.rooms):

            row = n

            # if each_room == 29:
            # n = 2
            # if each_room == 30:
            # n = 1

            r, g, b = self.elements[self.rooms[each_room]]['rgb']

            each_color = qu.Quantity_Color(r, g, b, qu.Quantity_TOC_RGB)
            each_hex_color = rgb2hex(r, g, b)

            stone = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_SHINY_PLASTIC)
            stone = gr.Graphic3d_MaterialAspect(gr.Graphic3d_NOM_BRASS)

            shape = self.elements[self.rooms[each_room]]['shape']
            ais = self.display.DisplayShape(shape,
                                            color=each_color,
                                            material=stone)

            self.table[row] = {"string": str(each_room),
                               "color": each_hex_color}
            self.elements[self.rooms[each_room]].update({'ais': ais,
                                                         'row': row,
                                                         })
            element_hash = self.elements[self.rooms[each_room]]['hash']
            self.hash_table[element_hash] = {'room': each_room,
                                             'row': row}

        self.disp['add_table'](None, None, self)

        for n, each_room in enumerate(self.rooms):

            status = self.elements[self.rooms[each_room]].get('visible', False)
            if status is False:
                self.remove_room(each_room)
                row_id = self.elements[self.rooms[each_room]]['row']
                self.disp['win'].Widg.toggleRowId(row_id, True)

        self.display.FitAll()

    def toggle_room(self, room):

        state = self.elements[self.rooms[room]].get('visible', None)
        if state is None:
            log.info("None")
            self.elements[self.rooms[room]]['visible'] = True
            self.remove_room(room)
        elif state is True:
            log.info("True")
            self.remove_room(room)
        elif state is False:
            log.info("False")
            self.show_room(room)

    def remove_barrier(self, barrier):
        ais_shape = self.elements[barrier]['ais']
        self.display.Context.Remove(ais_shape)

    def remove_room(self, room):

        ais_shape = self.elements[self.rooms[room]]['ais']
        self.elements[self.rooms[room]]['visible'] = False
        self.display.Context.Remove(ais_shape)

    def show_room(self, room):

        ais_shape = self.elements[self.rooms[room]]['ais']
        self.display.Context.Display(ais_shape)
        self.elements[self.rooms[room]]['visible'] = True


def main():
    pass


if __name__ == "__main__":
    main()

    import display.LegendGui as gui

    from OCC.Core.BRepPrimAPI import BRepPrimAPI_MakeBox
    rooms = {}
    for n in range(12):
        solid = BRepPrimAPI_MakeBox(gp.gp_Pnt(n, n, n), 1, 1, 1).Solid()
        rooms[n + 11] = {'solid': solid,
                         'hash': hash(solid)}

    disp = {}
    display = gui.init_display(disp, size=(600, 400), table=[])

    room_display = Displayed_Rooms(display, disp, rooms)
    room_display.display_rooms()

    disp['start']()
