# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import numpy as np
from skimage.morphology import binary_dilation

import OCC.Core.TopTools as tt
import OCC.Core.BRepAlgoAPI as bapi
import OCC.Core.TopoDS as tp
import OCC.Core.GProp as gprop
import OCC.Core.BRepGProp as bg
from OCC.Core.BRepBuilderAPI import BRepBuilderAPI_Sewing
import trimesh

from modules.triangulate import make_list_of_faces

import voxel_creator as gen


# Logging
from utils.logging_env import setup_logging, logging, log_debug
log = logging.getLogger(__name__)
setup_logging(log)
if log.getEffectiveLevel() <= 10:
    log_debug = True


def minimize(array, val=None, convert_zeros=False, minmax=None):

    a = array.copy()

    if val is not None and convert_zeros is False:
        a[a != val] = 0

    coords = np.transpose(np.nonzero(a))
    try:
        min_ = coords.min(axis=0)
    except ValueError:
        log.info(val)
        log.info(coords)
        return None, None
    ox, oy, oz = min_
    max_ = coords.max(axis=0) + 1

    if not convert_zeros:
        c = a[min_[0]:max_[0], min_[1]:max_[1], min_[2]: max_[2]].copy()

        if minmax is not None:
            minmax[0] = min_[0]
            minmax[1] = max_[0]
            minmax[2] = min_[1]
            minmax[3] = max_[1]
            minmax[4] = min_[2]
            minmax[5] = max_[2]

        return c, (float(ox), float(oy), float(oz))

    else:
        a = array.copy()
        b = array.copy()

        a[a != val] = 0

        b[b == 0] = val
        b[b != val] = 0

        c = a[min_[0]:max_[0], min_[1]:max_[1], min_[2]: max_[2]].copy()
        d = b[min_[0]:max_[0], min_[1]:max_[1], min_[2]: max_[2]].copy()

        return c, d, (float(ox), float(oy), float(oz))

    return c, (float(ox), float(oy), float(oz))


def bigger(array, origin, value=0):

    ox, oy, oz = origin

    nox = ox - 1
    noy = oy - 1
    noz = oz - 1

    sx, sy, sz = array.shape
    sx += 2
    sy += 2
    sz += 2
    bigger_array = np.full((sx, sy, sz), value, dtype=array.dtype)
    bigger_array[1:sx - 1, 1:sy - 1, 1:sz - 1] = array.copy()
    new_origin = (nox, noy, noz)

    return bigger_array, new_origin


def solidify_rooms(vox_group,
                   scene,
                   faces=False
                   ):

    rooms = {}
    room_hash = {}

    f = vox_group.vox_grid.voxel_scale_factor
    trsf = vox_group.trsf_vg_matrix

    # ---------------
    # TODO
    # Add code to not solidify non-internal rooms
    # but solidify them when toggled.
    # ---------------

    for each_source in vox_group.sources:

        room_array = np.zeros_like(vox_group.array)
        room_array[vox_group.connect == each_source] = 1

        visible = vox_group.sources[each_source].get('internal', False)

        new_elements, new_rooms, new_room_hash = solidify_room(room_array,
                                                               each_source,
                                                               f,
                                                               trsf,
                                                               visible,
                                                               faces)

        scene._elements.update(new_elements)
        rooms.update(new_rooms)
        room_hash.update(new_room_hash)

    return rooms, room_hash


def solidify_room(input_array, room_no, f, trsf, visible=True, faces=False):

    rooms = {}
    room_hash = {}
    elements = {}

    mini, origin = minimize(input_array, True)
    if mini is None:
        return elements, rooms, room_hash
    mini = mini.astype(np.int32)

    vox_volume = input_array.sum() * f**3
    if faces:
        shape = gen.make_faces(mini,
                               origin,
                               trsf,
                               1)
    else:
        shape = gen.iterate3(mini,
                             origin,
                             trsf,
                             1)
    elements = {}
    elements[hash(shape)] = {'shape': shape,
                             'hash': hash(shape),
                             "number": room_no,
                             "origin": origin,
                             "size": mini.shape,
                             "visible": visible,
                             "vox_volume": vox_volume}

    rooms[room_no] = hash(shape)
    room_hash[hash(shape)] = room_no

    return elements, rooms, room_hash


def dilate_rooms(vox_group,
                 scene,
                 ):

    rooms = {}
    room_hash = {}
    # dilated_rooms = np.zeros_like(vox_group.connect)

    for each_room in scene.internals:

        dilated_room = np.zeros_like(vox_group.array)
        dilated_room[vox_group.connect == each_room] = 1

        # vox_volume = (dilated_room.sum()
        #               * (scene.vox_grid.voxel_scale_factor) ** 3)

        d = np.ones((3, 3, 3), bool)
        dilated_room = binary_dilation(dilated_room, d)

        for each_adjacent in scene.adjacency_graph._adj[each_room]:
            log.info("adjacent to %s: %s" % (each_room, each_adjacent))
            if each_adjacent in [0, 1, 2, "E"]:
                continue
            log.info("adjacent to %s: %s" % (each_room, each_adjacent))
            dilated_room[vox_group.connect == each_adjacent] = 0

        dilated_room[vox_group.buffer] = 0

        new_elements, new_rooms, new_room_hash = solidify_room(dilated_room,
                                                               each_room)

        scene._elements.update(new_elements)
        rooms.update(new_rooms)
        room_hash.update(new_room_hash)

    return rooms, room_hash


def remove_features(vox_group, scene, room, mesh=True):

    hash_ = scene.rooms[room]
    origin = scene._elements[hash_]['origin']
    size = scene._elements[hash_]['size']
    # vox_volume = scene._elements[hash_].get('vox_volume', 0)

    minmax = (origin[0] - 0.5,
              origin[1] - 0.5,
              origin[2] - 0.5,
              origin[0] + size[0] - 0.5,
              origin[1] + size[1] - 0.5,
              origin[2] + size[2] - 0.5)

    features = list(scene.vox_idx.intersection(minmax))

    room_shape = scene._elements[hash_]['shape']

    vox_vol = scene._elements[hash_].get('vox_volume', 0)

    trimmed_room, bad_boys = cut_to_compound(room_shape, features,
                                             scene._elements, vox_vol)

    scene._elements[hash(trimmed_room)] = {'shape': trimmed_room,
                                           'hash': hash(trimmed_room),
                                           "number": room,
                                           'visible': True,
                                           'vox_volume': vox_vol}

    for each_bad in bad_boys:

        scene._elements[each_bad]['tag'] = True

    scene.rooms[room] = hash(trimmed_room)
    scene.room_hash[hash(trimmed_room)] = room

    log.info(features)


def cut_to_mesh(to_be_cut, cutters, scene):

    elements = scene._elements

    scene.calculate_meshes([to_be_cut])
    scene.calculate_meshes(cutters)

    mesh_list = [elements[to_be_cut]['mesh']]

    for each in cutters:
        mesh_list.append(elements[each]['mesh'])

    diff = trimesh.boolean.minkowski(mesh_list,
                                     engine='scad',
                                     dir_="/Users/garethboyes/Downloads/temp/")

    if type(diff) is list:
        list_of_faces = []
    else:
        list_of_faces = make_list_of_faces(diff.vertices, diff.faces)

    mkShell = BRepBuilderAPI_Sewing()
    for each in list_of_faces:
        mkShell.Add(each)
    mkShell.Perform()

    sew = mkShell.SewedShape()

    return sew


def volume(shape):

    volume = gprop.GProp_GProps()
    bg.brepgprop_VolumeProperties(shape, volume, 1e-5)

    return volume.Mass()


def cut_to_compound(to_be_cut, cutters, elements, min_vol):

    remaining_parts = [to_be_cut]
    number_of_cutters = len(cutters)

    log.info([volume(each) for each in remaining_parts])
    log.info(min_vol)

    bad_boys = []

    for n, each_cutter in enumerate(cutters):

        cutter_shape = elements[each_cutter]['shape']

        log.info("Cutting %s of %s - vol:%s " % (n, number_of_cutters,
                                                 volume(cutter_shape)))
        tools = tt.TopTools_ListOfShape()
        tools.Append(cutter_shape)

        new_parts = []

        for each_part in remaining_parts:
            args = tt.TopTools_ListOfShape()
            args.Append(each_part)
            cutter = bapi.BRepAlgoAPI_Cut()
            cutter.SetTools(tools)
            cutter.SetArguments(args)
            cutter.SetRunParallel(True)
            cutter.SetFuzzyValue(1e-5)
            cutter.Build()

            if cutter.IsDeleted(each_part):
                log.info('deleted')
                continue

            if not cutter.HasModified():
                new_parts.append(each_part)
                log.info('not_modified')
                continue

            mods = cutter.Modified(each_part)
            listit = tt.TopTools_ListIteratorOfListOfShape(mods)

            if mods.Extent():

                while listit.More():
                    new_parts.append(listit.Value())
                    listit.Next()
            else:
                new_parts.append(each_part)

        new_parts.sort(key=lambda x: volume(x), reverse=True)

        log.info([volume(each) for each in new_parts])

        if ((new_parts and (volume(new_parts[0]) <= min_vol))
            or len(new_parts)) == 0:

            log.info("erroneous cut detected")
            bad_boys.append(each_cutter)
            continue

        remaining_parts = new_parts

        remaining_parts = [remaining_parts[0]]

    log.info(remaining_parts[0])
    return tp.topods.Solid(remaining_parts[0]), bad_boys


def form_solid_from_voxels(voxel_rooms,
                           scene_elements,
                           scene_internals,
                           trsf_vg_matrix,
                           room_start=11):
    rooms = {}
    room_hash = {}
    log.info(room_start)
    log.info(voxel_rooms.max() + 1)

    for each_room in range(room_start, voxel_rooms.max() + 1):
        # for each_room in scene_internals:
        '''Solidify voxels'''

        if each_room in scene_internals:
            log.info("Skipping %s" % each_room)
            continue
        solid = gen.iterate3(voxel_rooms, [0, 0, 0], trsf_vg_matrix, each_room)
        scene_elements[hash(solid)] = {'shape': solid,
                                       'hash': hash(solid),
                                       "number": each_room,
                                       "visible": False}
        rooms[each_room] = hash(solid)
        room_hash[hash(solid)] = each_room

    return rooms, room_hash
