# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import sys
import getpass
import _pickle as cPickle
import numpy as np

import cx_Oracle as cx

import OCC.Core.BRepBuilderAPI as brep
import OCC.Core.BRepBndLib as blib
import OCC.Core.Bnd as bnd

if getpass.getuser() == 'gareth' and sys.platform == 'win32':
    import Py3DGeom_py as Py3DGeom
elif getpass.getuser() == 'garethboyes' and sys.platform == "darwin":
    import Py3DGeom as Py3DGeom
elif getpass.getuser() == 'gareth' and sys.platform == "darwin":
    import Py3DGeom_py as Py3DGeom

# Logging
from utils.logging_env import setup_logging, logging, log_debug
log = logging.getLogger(__name__)
setup_logging(log)
if log.getEffectiveLevel() <= 10:
    log_debug = True


class Scene(object):

    def __init__(self, settings, scene_elements, limit=None, offset=(0, 0, 0),
                 guids=None,
                 special=False,
                 table="A20170528_ELEMENTS",
                 where=None,
                 tol=0.00005,
                 multi_surface=False):

        self.offset = offset
        self.condition1 = ""

        self.settings = settings

        if special == "penguin":
            table = "A20170528_ELEMENTS"
            guids = ['1uA34BXDX3HeiJ1MBuh_wi']
            self.conditon1 = "WHERE a.GUID = '%s'" % guids[0]

        elif special == 'cutaway':
            table = "A20200826_CUTAWAY"
            guids = ['cutaway']
            self.conditon1 = "WHERE a.GUID = '%s'" % guids[0]

        elif special == 'extruded':

            self.condition1 = ""

        elif special == 'extruded_and_MEP':

            self.condition1 = ""

        elif guids is not None:
            if len(guids) == 1:
                self.condition1 = "WHERE a.GUID = '%s'" % guids[0]
            else:
                self.condition1 = "WHERE a.GUID IN %s" % str(tuple(guids))

        elif where is not None:
            self.condition1 = where

        if getpass.getuser() == 'gareth' and sys.platform == 'win32':
            self.con_tuple = ('gareth', 'BimGis32', 'localhost:1521/orcl')
            self.table = table
        elif getpass.getuser() == 'garethboyes' and sys.platform == "darwin":
            self.con_tuple = ('scott', 'tiger', 'localhost:1521/orcl')
            self.table = table

        self.elements = []
        self.n = 0
        self.scene_elements = scene_elements

        self.query1 = ("SELECT a.GEOM.SDO_GTYPE, a.GEOM.SDO_SRID, "
                       "a.GEOM.SDO_POINT, a.GEOM.SDO_ELEM_INFO, "
                       "a.GEOM.SDO_ORDINATES, a.RED, a.GREEN, a.BLUE, "
                       "a.GUID, a.SUBID FROM " + self.table + " a ")
        self.query2 = ("SELECT a.PICKLE, a.RED, a.GREEN, a.BLUE, a.GUID, "
                       "a.SUBID, a.CLOSED FROM " + self.table + " a ")

        if settings.get("multi_surface", False) is True:
            log.info("Loading as multi-surface")
            multi_surface = True

        if settings.get('load_shape', False) is True:
            log.info("Loading pickled geometry")
            self.load(scene_elements, limit=limit)
        else:
            log.info("Loading SDO_GEOMETRY")
            self.query(scene_elements, limit=limit, offset=offset, tol=tol,
                       multi_surface=multi_surface)

            if settings.get('pickle', False) is True:
                log.info("Pickling SDO_GEOMETRY")
                self.dump(scene_elements)  # TODO manage offset

    def add(self,
            pickle=None,
            limit=None,
            guids=None,
            penguin=False,
            table="A20170528_ELEMENTS",
            multi_surface=False,
            where=None):

        self.condition1 = ""

        if penguin:
            table = "A20170528_ELEMENTS"
            guids = ['1uA34BXDX3HeiJ1MBuh_wi']
            self.conditon1 = "WHERE a.GUID = '%s'" % guids[0]

        elif guids is not None:
            if len(guids) == 1:
                self.condition1 = "WHERE a.GUID = '%s'" % guids[0]
            else:
                self.condition1 = "WHERE a.GUID IN %s" % str(tuple(guids))

        elif where is not None:
            self.condition1 = where

        self.query1 = ("SELECT a.GEOM.SDO_GTYPE, a.GEOM.SDO_SRID, "
                       "a.GEOM.SDO_POINT, a.GEOM.SDO_ELEM_INFO, "
                       "a.GEOM.SDO_ORDINATES, a.RED, a.GREEN, a.BLUE, a.GUID, "
                       "a.SUBID FROM " + table + " a ")
        self.query2 = ("SELECT a.PICKLE, a.RED, a.GREEN, a.BLUE, a.GUID, "
                       "a.SUBID, a.CLOSED FROM " + table + " a ")

        if ((pickle is None and self.settings.get('load_shape', False) is True)
            or pickle is True):
            log.info("Loading pickled geometry")
            elements = self.load(self.scene_elements, query=self.query2,
                                 limit=limit, offset=self.offset)
        else:
            log.info("Loading SDO_GEOMETRY")
            elements = self.query(self.scene_elements, limit=limit,
                                  multi_surface=multi_surface)

        return elements

    def query(self,
              scene_elements,
              query_="",
              limit=None,
              offset=(0, 0, 0),
              multi_surface=False,
              tol=0.00005):

        if not query_:
            query_ = self.query1

        if limit is None:
            limit = 99999999999999

        offset = np.array(offset)

        con = cx.connect(*self.con_tuple)
        cur = con.cursor()
        log.info(query_ + self.condition1)
        cur.execute(query_ + self.condition1)
        elements = {}
        element_list = []
        n = 0

        bb = bnd.Bnd_Box()

        keepgoing = True

        while keepgoing and limit:

            element = cur.fetchone()
            if element:
                ordinates = np.array(element[4].aslist())
                reshaped = np.reshape(ordinates, (len(ordinates) // 3, 3))

                ordinates = reshaped - offset
                ordinates = list(ordinates.ravel())

                sdo_geom = Py3DGeom.SDO_Geometry(None,
                                                 element[0],
                                                 element[1],
                                                 None,
                                                 element[3].aslist(),
                                                 ordinates,
                                                 debug=True)
                r = element[5]
                g = element[6]
                b = element[7]

                sdo_geom.geom.set_colour((r, g, b))

                if not multi_surface:
                    shape = sdo_geom.geom.shape(tol)
                else:
                    shape = sdo_geom.geom.multi_surface

                bb = bnd.Bnd_Box()

                blib.brepbndlib_AddClose(shape, bb)
                cornermin = bb.CornerMin().Coord()
                cornermax = bb.CornerMax().Coord()
                corner_tuple = (*cornermin, *cornermax)
                del bb
                alpha = 0
                if not multi_surface:

                    solid = brep.BRepBuilderAPI_MakeSolid(shape).Solid()
                    hash_ = solid.__hash__()
                    elements[solid.__hash__()] = {'guid': element[8],
                                                  'subid': element[9],
                                                  'shape': solid,
                                                  'name': '',
                                                  'rgb': (r, g, b),
                                                  'alpha': alpha,
                                                  'closed': True,
                                                  'bbxyz': corner_tuple}
                    element_list.append(hash_)
                    n += 1

                else:
                    log.info("Adding Multi-surface")
                    hash_ = shape.__hash__()
                    elements[shape.__hash__()] = {'guid': element[8],
                                                  'subid': element[9],
                                                  'shape': shape,
                                                  'name': '',
                                                  'rgb': (r, g, b),
                                                  'alpha': alpha,
                                                  'closed': False,
                                                  'bbxyz': corner_tuple}

                    element_list.append(hash_)
                    n += 1

                limit -= 1

            else:
                keepgoing = False

        cur.close()
        con.close()
        self.scene_elements.update(elements)
        self.elements.extend(element_list.copy())
        self.n += n

        return element_list

    def dump(self, scene_elements):

        con = cx.connect(*self.con_tuple)
        cur = con.cursor()

        for each in scene_elements:
            string = cPickle.dumps(scene_elements[each]['shape'])

            blobvar = cur.var(cx.BLOB)
            blobvar.setvalue(0, string)
            sqlStr = ("update "
                      + self.table
                      + (" set PICKLE=:blobData where GUID = '%s' "
                         % scene_elements[each]['guid'])
                      + "and SUBID = '%s'" % scene_elements[each]['subid'])
            cur.setinputsizes(blobData=cx.BLOB)
            cur.execute(sqlStr, {'blobData': blobvar,
                                 })
            del blobvar
        cur.close()
        con.commit()
        con.close()

    def load(self, scene_elements, query=None, limit=None, offset=(0, 0, 0)):

        # TODO - Handle offset

        if not query:
            query = self.query2

        if limit is None:
            limit = 99999999999999

        con = cx.connect(*self.con_tuple)
        cur = con.cursor()
        log.info(query + self.condition1)
        cur.execute(query + self.condition1)
        elements = {}
        element_list = []
        n = 0
        global element
        keepgoing = True

        while keepgoing and limit:
            element = cur.fetchone()
            if element:
                if not element[4]:
                    log.warning("missing guid")
                    continue
                if element[5] is None:
                    log.warning("missing subid")
                    continue
                shape = cPickle.loads(element[0].read())
                assert(not shape.IsNull())

                r = element[1]
                g = element[2]
                b = element[3]

                bb = bnd.Bnd_Box()
                blib.brepbndlib_AddClose(shape, bb)

                cornermin = bb.CornerMin().Coord()
                cornermax = bb.CornerMax().Coord()
                corner_tuple = (*cornermin, *cornermax)

                hash_ = shape.__hash__()
                elements[hash_] = {'guid': element[4],
                                   'subid': element[5],
                                   'name': "",
                                   'shape': shape,
                                   'rgb': (r, g, b),
                                   'alpha': 0,
                                   'ais': None,
                                   'bbxyz': corner_tuple,
                                   'closed': True}

                element_list.append(hash_)
                limit -= 1
                n += 1

            else:
                keepgoing = False

        cur.close()
        con.close()
        self.scene_elements.update(elements)
        self.elements.extend(element_list.copy())
        self.n += n

        return element_list
