# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021
import numpy as np
import scipy.ndimage as si


from OCC.Core.BRepPrimAPI import BRepPrimAPI_MakeBox
from OCC.Core.BRepBuilderAPI import (BRepBuilderAPI_MakePolygon,
                                     BRepBuilderAPI_MakeFace,
                                     BRepBuilderAPI_Sewing)
from OCC.Core.gp import gp_Trsf, gp_Vec, gp_Pnt
from OCC.Core.TopLoc import TopLoc_Location
from OCC.Core.BRepAlgoAPI import BRepAlgoAPI_Cut
from OCC.Core.BRep import BRep_Tool
from OCC.Core.BRepMesh import BRepMesh_IncrementalMesh
from OCC.Core.TopAbs import TopAbs_REVERSED
from OCC.Core.TopoDS import topods
import OCC.Core.AIS as ais
import OCC.Core.BRepBuilderAPI as brep
import OCC.Core.gp as gp
import OCCUtils

import trimesh

from pythonocc.Display import SimpleGui

# Logging
from utils.logging_env import setup_logging, logging, log_debug
log = logging.getLogger(__name__)
setup_logging(log)
if log.getEffectiveLevel() <= 10:
    log_debug = True


def transform(node_array, trsf):

    for each_row in range(node_array.shape[0]):
        node = tuple(node_array[each_row])
        pt = gp_Pnt(*node)
        pt.Transform(trsf)

        node_array[each_row, :] = pt.Coord()


def make_list_of_faces(node_array, face_list):

    f_list = []

    for each_face in face_list:
        tup1 = tuple(node_array[each_face[0]])
        tup2 = tuple(node_array[each_face[1]])
        tup3 = tuple(node_array[each_face[2]])
        f_list.append(make_triangle(tup1, tup2, tup3))

    return f_list


def make_shell(node_array, face_list):

    l_of_faces = make_list_of_faces(node_array, face_list)
    mkShell = BRepBuilderAPI_Sewing()

    for each in l_of_faces:
        mkShell.Add(each)
    mkShell.Perform()

    sew = mkShell.SewedShape()
    # sew.Reverse()
    shell = topods.Shell(sew)

    return shell


def make_triangle(p1, p2, p3):

    wire = BRepBuilderAPI_MakePolygon(gp_Pnt(*p1),
                                      gp_Pnt(*p2), gp_Pnt(*p3), True).Wire()
    face = BRepBuilderAPI_MakeFace(wire).Face()

    return face


def triangulate_face(face, face_list, node_dict, node_id_dict, e=0.0000001):

    faceLocation = face.Location()
    faceTransformation = faceLocation.Transformation()
    mesher = BRepMesh_IncrementalMesh(face, 0.1)
    mesher.Perform()
    tool = BRep_Tool()
    tri = tool.Triangulation(face, faceLocation)
    reversed_ = (face.Orientation() == TopAbs_REVERSED)

    if tri:
        triangles = tri.Triangles()
        nodes = tri.Nodes()
        len_nodes = nodes.Length()

        local_node_dict = {}

        factor = 1 / e

        for each_node in range(1, len_nodes + 1):
            every_node = nodes.Value(each_node)
            x, y, z = every_node.Transformed(faceTransformation).Coord()
            x_ = int(round(x * factor))
            y_ = int(round(y * factor))
            z_ = int(round(z * factor))

            if (x_, y_, z_) in node_dict:
                node_id = node_dict[(x_, y_, z_)]
            else:
                node_id = len(node_id_dict)
                node_dict[(x_, y_, z_)] = node_id
                node_id_dict[node_id] = (x, y, z)

            local_node_dict[each_node] = node_id

        if triangles and triangles.Length() > 0:
            for i in range(1, triangles.Length() + 1):

                triangle = triangles.Value(i)
                n1, n2, n3 = triangle.Get()
                if (reversed_):
                    n1, n2 = n2, n1

                face_list.append((local_node_dict[n1],
                                  local_node_dict[n2],
                                  local_node_dict[n3]))


def triangulate(shape, e=1e-7):

    faces = OCCUtils.Topo(shape).faces()

    node_list = []
    face_list = []

    node_dict = {}
    node_id_dict = {}

    for n, each_face in enumerate(faces):
        triangulate_face(each_face, face_list, node_dict, node_id_dict, e)

    node_list_sorted = list(node_id_dict.keys())
    node_list_sorted.sort()

    for each in node_list_sorted:
        node_list.append(node_id_dict[each])

    node_array = np.array(node_list)

    return node_array, face_list


def process_all(list_of_shapes):

    corner1 = (83300, 36250, 90)
    corner2 = (83500, 36350, 120)

    gp1 = gp_Pnt(*corner1)
    gp2 = gp_Pnt(*corner2)

    box = BRepPrimAPI_MakeBox(gp1, gp2).Solid()

    node_array, face_list = triangulate(box)
    node_array = node_array - corner1

    new_list = []
    for each in list_of_shapes:
        node_array, face_list = triangulate(each)

        node_array = node_array - corner1

        m = trimesh.Trimesh(node_array, face_list)
        if not m.is_watertight:
            continue

        if len(node_array) / m.volume > 200:
            new_list.append(each)

    return new_list

def voxelise(node_array, face_list, scale_factor=1, interior=True):

    nod_arr = node_array  # MAC
    # nod_arr = node_array + 0.5   PC

    # TODO - Why does this perform differently on PC/MAC

    mesh = trimesh.Trimesh(nod_arr, face_list)

    voxels = mesh.voxelized(scale_factor)

    vx = voxels.matrix_surface.astype(np.bool)
    if interior:

        # Grow array by 1 zero voxel in each direction
        sx, sy, sz = vx.shape
        b = np.zeros((sx + 2, sy + 2, sz + 2), np.bool)
        b[1:sx + 1, 1:sy + 1, 1:sz + 1] = vx

        # Label inverted array
        b = (~b)
        c = si.label(b)[0]

        # Identify interior voxels and shrink array to original size
        d = (c > 1)
        d = d[1:sx + 1, 1:sy + 1, 1:sz + 1]

        # Convert matrix to points
        e = trimesh.voxel.matrix_to_points(d, scale_factor,
                                           voxels.origin + 0.5)
        g = trimesh.voxel.matrix_to_points(d, scale_factor, (0.5, 0.5, 0.5))

        e = e.astype(np.float)
        g = g.astype(np.int)

        if len(e):
            a = trimesh.ray.ray_pyembree.RayMeshIntersector(mesh)
            f = a.contains_points(e)

            for n, each in enumerate(g):
                if f[n]:
                    vx[each[0], each[1], each[2]] = True

    return vx, voxels


def Pt(a):
    return gp.gp_Pnt(a[0], a[1], a[2])


class Face(object):

    def __init__(self, *args):
        if len(args) == 3:
            self.init(*args)
        pass

    def init(self, a, b, c):

        pntList = [Pt(a), Pt(b), Pt(c)]
        poly = brep.BRepBuilderAPI_MakePolygon()

        for each in pntList:
            poly.Add(each)

        poly.Close()

        face = brep.BRepBuilderAPI_MakeFace(poly.Wire())
        self._face = face.Face()
        self.ais_face = ais.AIS_Shape(face.Shape())

    def from_plane(self, plane, a, b, c, d, a1, b1, c1, d1):

        pntList = [Pt(a), Pt(b), Pt(c), Pt(d)]
        pntList1 = [Pt(a1), Pt(b1), Pt(c1), Pt(d1)]

        poly = brep.BRepBuilderAPI_MakePolygon()

        for each in pntList:
            poly.Add(each)

        poly.Close()

        poly1 = brep.BRepBuilderAPI_MakePolygon()

        for each in pntList1:
            poly1.Add(each)

        poly1.Close()

        face = brep.BRepBuilderAPI_MakeFace(plane, poly.Wire(), True)

        face.Add(poly1.Wire())

        self.ais_face = ais.AIS_Shape(face.Shape())

    def from_pln(self, plane):

        face = brep.BRepBuilderAPI_MakeFace(plane, -20, 20, -20, 20, 0.00001)
        self._face = face.Face()

        self.ais_face = ais.AIS_Shape(face.Shape())

    @property
    def face_handle(self):

        return self.ais_face.GetHandle()

    @property
    def face(self):
        return self._face


def make_solid(node_array, face_array):

    mkShell = brep.BRepBuilderAPI_Sewing()

    for a, b, c in face_array:

        face = Face(node_array[a],
                    node_array[b],
                    node_array[c])

        mkShell.Add(face.face)
    mkShell.Perform()

    sew = mkShell.SewedShape()
    shell = topods.Shell(sew)

    solid = brep.BRepBuilderAPI_MakeSolid(shell).Solid()

    return solid


def main():

    global e

    MoveX = 1
    MoveY = 1
    MoveZ = 1
    DX = 2
    DY = 2
    DZ = 2

    box = BRepPrimAPI_MakeBox(DX, DY, DZ).Shape()

    tr = gp_Trsf()
    tr.SetTranslation(gp_Vec(MoveX, MoveY, MoveZ))
    tr1 = gp_Trsf()
    tr1.SetTranslation(gp_Vec(-1, -1, -1))

    box2 = box.Located(TopLoc_Location(tr))

    cutShape = BRepAlgoAPI_Cut(box, box2).Shape()

    global node_array, face_list
    node_array, face_list = triangulate(cutShape)

    solid = make_solid(node_array, face_list)

    disp = {}
    display = SimpleGui.init_display(disp=disp)
    display.DisplayShape(solid)
    disp['start']()
