# python3
# -*- coding: utf-8 -*-

# Copyright Gareth Boyes, 2021

import sys
import warnings
from itertools import product, count
import networkx as nx
import cython as cy
import igraph as ig
from pprint import pprint

# Logging
from utils.logging_env import setup_logging, logging
log = logging.getLogger(__name__)
setup_logging(log)

SQRT2 = 1.4142135623730951
SQRT3 = 1.7320508075688772

adjacency_6 : tuple =   ((1,0,0,False,1),
                        (0,1,0,False,1),
                        (0,0,1,False,1))
                 
                        ##(-1,0,0),
                        ##(0,-1,0),
                        ##(0,0,-1)]                           
                 
adjacency_18 : tuple = adjacency_6 + ((1,1,0,True,SQRT2),
                                    (1,0,1,True,SQRT2),
                                    (0,1,1,True,SQRT2),
                                    (1,-1,0,True,SQRT2),
                                    (1,0,-1,True,SQRT2),
                                    (0,1,-1,True,SQRT2))
                 
                                    ##(-1,-1,0),
                                    ##(-1,0,-1), 
                                    ##(0,-1,-1), 
                                    ##(-1,1,0),
                                    ##(-1,0,1), 
                                    ##(0,-1,1),                              
                 
adjacency_26 :tuple = adjacency_18 +((1,1,1,True,SQRT3), 
                                    (1,1,-1,True,SQRT3),                             
                                    (1,-1,-1,True,SQRT3), 
                                    (1,-1,1,True,SQRT3))
                                                                 
                                    ##(-1,-1,-1), 
                                    ##(-1,-1,1),                             
                                    ##(-1,1,1), 
                                    ##(-1,1,-1),


@cy.cfunc
def query(x2 : cy.int,
          y2 : cy.int,
          z2 : cy.int,
          sx : cy.int,
          sy : cy.int,
          sz : cy.int,
          dft: cy.int[:,:,::1],
          mask : cy.int[:,:,::1],
          val : cy.int,
          diag : cy.bint) -> cy.int:
    
    if x2 == sx or y2 == sy or z2 == sz:
        return 0    

    if x2 == -1 or y2 == -1 or z2 == -1:
        return 0    

    if mask[x2, y2, z2] == 1:
        return 0
    
    if dft[x2,y2,z2] == 0:
        return 0
    
    if dft[x2,y2,z2] == 1 and diag:
        return 0    
    
    if dft[x2, y2, z2] == val:
        return 1
    
#    if dft[x2,y2,z2] == val + 1:
    if dft[x2,y2,z2] > val:

        return 2

#    if dft[x2,y2,z2] == val - 1:
    if dft[x2,y2,z2] < val:

        return 3
    
    return 0
    

@cy.cfunc
@cy.boundscheck(False)
@cy.wraparound(False)
def populate_graphs(sx : cy.int,
                    sy : cy.int,
                    sz : cy.int,
                    mask : cy.int[:,:,::1],
                    dft : cy.int[:,:,::1],
                    graph_edges : dict,
                    g_vertices : list,
                    g_edges : list,
                    use_only_6: cy.bint
                    ) -> cy.void:
    
    sysz : cy.int = sy * sz

    x : cy.int
    y : cy.int
    z : cy.int
    dx : cy.int
    dy : cy.int
    dz : cy.int
    diag : cy.bint
    dist : cy.float
    x2 : cy.int
    y2 : cy.int
    z2 : cy.int
    val : cy.int
    ret : cy.int

    for x, y, z in product(range(sx), range(sy), range(sz)):

        if mask[x, y, z] != 1 and dft[x, y, z] != 0:
            
            val = dft[x, y, z]
            
            if val == 1 or use_only_6:
                adjacencies = adjacency_6
            else:
                adjacencies = adjacency_26
                             
            for dx,dy,dz,diag,dist in adjacencies:
                                                                        
                x2 = x + dx
                y2 = y + dy
                z2 = z + dz
                
                val = dft[x, y, z]
                
                                                  
                ret = query(x2, y2, z2, sx, sy, sz, dft, mask, val, diag)
                
                if ret == 0:
                    continue
                
                val2 = dft[x2, y2, z2]
                
                if ret == 1:
                    #graph_edges[dft[x, y, z]].append((pos_(sysz,sz,(x, y, z)),
                                                  #pos_(sysz,sz,(x2, y2, z2))))
                    
                    graph_edges[val].append((str(pos_(sysz,sz,(x, y, z))),
                                             str(pos_(sysz,sz,(x2, y2, z2)))))
                    
                    #g_edges.append({"source" : str(pos_(sysz,sz,(x, y, z))),
                                    #"target" : str(pos_(sysz,sz,(x2, y2, z2))),
                                    #"source_val" : val,
                                    #"target_val" : val,
                                    #"weight" : dist,
                                    #"diag" : diag})
                    
                    continue
                
                if ret == 2:     
                    #graph_edges[val, val+1].append((pos_(sysz,sz,(x, y, z)),
                                                #pos_(sysz,sz,(x2, y2, z2))))
                    
                    graph_edges[val].append((str(pos_(sysz,sz,(x, y, z))),
                                              str(pos_(sysz,sz,(x2, y2, z2)))))                      
                    
                    #g_edges.append({"source" : str(pos_(sysz,sz,(x, y, z))),
                                    #"target" : str(pos_(sysz,sz,(x2, y2, z2))),
                                    #"source_val" : val,
                                    #"target_val" : val2,
                                    #"weight" : dist,
                                    #"diag" : diag})
                    
                    continue
                        
                if ret == 3:     
                    #graph_edges[val - 1, val].append((pos_(sysz,sz,(x, y, z)),
                                                     #pos_(sysz,sz,(x2, y2, z2))))
                    
                    graph_edges[val2].append((str(pos_(sysz,sz,(x, y, z))),
                                              str(pos_(sysz,sz,(x2, y2, z2)))))                    
                    
                    #g_edges.append({"source" : str(pos_(sysz,sz,(x, y, z))),
                                    #"target" : str(pos_(sysz,sz,(x2, y2, z2))),
                                    #"source_val" : val2,
                                    #"target_val" : val,
                                    #"weight" : dist,
                                    #"diag" : diag})                  
                    continue                
                
                #if mask[x, y+1, z] != 1 and dft[x, y+1, z] != 0:
                    
                    #if dft[x, y, z] == dft[x, y+1, z]:
                        #graphs[dft[x, y, z]].add_edge((x, y, z),(x, y+1, z))
        
                    #if dft[x, y, z] == dft[x, y+1, z] - 1:
                        #graphs[(dft[x, y, z], dft[x, y+1, z])].add_edge((x, y, z),(x, y+1, z))
                    
                
                #if mask[x, y, z+1] != 1 and dft[x, y, z+1] != 0:
                    
                    #if dft[x, y, z] == dft[x, y, z+1] - 1:
                        #graphs[(dft[x, y, z], dft[x, y, z+1])].add_edge((x, y, z),(x, y, z+1))
                    
                    #if dft[x, y, z] == dft[x, y, z+1]:
                        #graphs[dft[x, y, z]].add_edge((x, y, z),(x, y, z+1))                
                


          
@cy.ccall
def connect_up(val : cy.int,
               #connect,
               sources : dict,
               dft : cy.int[:,:,::1],
               raw_dft : cy.int[:,:,::1],
               mask : cy.int[:,:,::1],
               log_debug : cy.bint,
               start : cy.int,
               no_barriers: cy.bint,
               use_only_6: cy.bint):
               #restricted = False):
    
    
    next_source : cy.int = start
    each_source : cy.int
    #sources_by_value : dict = {}
    
    sx : cy.int = dft.shape[0]
    sy : cy.int = dft.shape[1]
    sz : cy.int = dft.shape[2]
    sysz : cy.int
    max_d  : cy.float
    a : cy.int
    b : cy.int
    source_a : cy.int
    source_b : cy.int
    source : cy.long
    target : cy.long 
    
    d : cy.float
    
    pos_x : cy.int
    pos_y : cy.int
    pos_z : cy.int
    
    #sx,sy, sz = dft.shape
    sysz = sy * sz
    
    touching : cy.bint = False
    
    graph_edges : dict = {}
    g_edges : list =[]
    g_vertices : list = []
    
    each_val : cy.int
    last_val  : cy.int
    
    max_node : tuple
    
    composed_graph : ig.Graph
    each_subgraph : ig.Graph
    connected_to_a_node : ig.Graph
    G : ig.Graph
    
    a_node : str
    
    sources_touching : list
    sources_by_pos : dict = {}
    
    verts : set
    edges: list
    each_edge : tuple
    
    log.info("Hello")
    
    if val == 0:
        raise ValueError("val must be more than zero")
    
    for each_val in range(val+1):
        graph_edges[each_val] = []
        #graph_edges[(each_val, each_val+1)] = []    
    
    populate_graphs(sx, sy, sz, mask, dft, graph_edges, g_vertices, g_edges, use_only_6)
    
    g_vertices = [{"pos":v, "name":str(v)} for v in range(sx*sy*sz)]
    
    #complete_graph = ig.Graph.DictList(vertices=g_vertices, edges=g_edges)
    
    G = ig.Graph.DictList(vertices=g_vertices, edges=[])

    sources_touching= []
    
    # ---------------
    # TODO
    #Investigate the effect of not scanning for sources in value=1
    # ---------------
    
    
    for each_val in range(val, 0, -1):
    #for each_val in range(val, 1, -1):
        log.info("looking at %s" % each_val)
        
        #if len(graph_edges[each_val]) == 0:
            ##sources_by_value[each_val] = []
            #continue
        
        #graph = ig.Graph()
        #verts = set([])
        #edges = []
        
        #for each_edge in graph_edges[each_val]:
            #verts.add(str(each_edge[0]))
            #verts.add(str(each_edge[1]))
            #edges.append((str(each_edge[0]), str(each_edge[1])))
        #graph.add_vertices(list(verts))
        #graph.add_edges(edges)
        
        #for each_edge in graph_edges[(each_val,each_val+1)]:
            #verts.add(str(each_edge[0]))
            #verts.add(str(each_edge[1]))
            #edges.append((str(each_edge[0]), str(each_edge[1])))        
        
        #if len(composed_graph.vs):
            #existing_verts = set(composed_graph.vs['name'])
            #verts.difference_update(existing_verts)
        #composed_graph.add_vertices(list(verts))
        #composed_graph.add_edges(edges)   
        
        #x = complete_graph.es(source_val_ge=each_val).subgraph()
            
        #log.info(len(graph_edges[each_val]))
        
            
        G.add_edges(graph_edges[each_val])
    
        
        #log.info("edges added")    
        
        if G.ecount() != 0:
            #pprint(list(complete_graph.vs))
            #pprint({edge.index :(G.vs[edge.source]['name'],G.vs[edge.target]['name']) for edge in G.es})
            sys.stdout.flush()
            x = G.es.subgraph()
            list_of_subgraphs : list = x.decompose()
        else:
            list_of_subgraphs = []
        #log.info("subgraphs decomposed")    

        
        if len(sources) == 0:
            sources_touching = []
            for each_subgraph in list_of_subgraphs:
                max_node = (0, 0)
                for each_node in each_subgraph.vs:
                    #log.debug(each_node)
                    each_node_name = int(each_node["name"])
                    pos_x, pos_y, pos_z = inv_pos(sysz,sz, each_node_name)
                    this_node_val = raw_dft[pos_x, pos_y, pos_z]
                    if this_node_val > max_node[0]:
                        max_node = (this_node_val, each_node_name)
                
                
                #a_node = list(each_subgraph.nodes)[0]
                sources[next_source] = {"value":each_val,
                                        "pos": inv_pos(sysz,sz,max_node[1]),
                                        "touching":False,
                                        "pos_id" : str(max_node[1]),
                                        "internal" : True}
                
                sources_by_pos[str(max_node[1])] = next_source
                log.info("Source %s identifed positioned by %s " % (next_source, max_node[1]))
                next_source += 1
        
        else:
            sources_touching = []
            
            # ---------------
            # TODO Work out why list_of_subgraphs might be empty
            # ---------------
            
            sources_set = set(sources_by_pos)
            
            for each_subgraph in list_of_subgraphs:
                log.debug("subgraph %s" % len(each_subgraph.vs))
                
                sub_graph_verts = set(each_subgraph.vs['name'])
                log.debug(list(sub_graph_verts)[:10])
                log.debug([inv_pos(sysz, sz, int(each)) for each in list(sub_graph_verts)[:10]])
                sources_touching_pos = sources_set.intersection(sub_graph_verts)
                log.debug("Intersection: %s" % str(list(sources_touching_pos)))
                #print([inv_pos(sysz, sz, int(each)) for each in sources_touching_pos])
                
                ##a_node = list(each_subgraph.vs)[0]['name']

                #print(a_node)
                #a_node = composed_graph.vs.find(a_node)[0]
                #print([v for v in composed_graph.vs])
                
                ##connected_to_a_node = composed_graph.neighborhood(a_node, (sx+sy+sz))
                
                #print(connected_to_a_node)
                
                
                ##sources_touching = []
                
                ##for each_source in sources:
                    ##if pos_(sysz,sz,sources[each_source]['pos']) in [int(composed_graph.vs[v]['name']) for v in connected_to_a_node]:
                        ###sources[each_source]
                        ##sources_touching.append(each_source)
                        
                if len(sources_touching_pos) == 0: # and each_val != 1:
                    max_node = (0, 0)
                    for each_node in each_subgraph.vs:
                        each_node_name = int(each_node["name"])
                        pos_x, pos_y, pos_z = inv_pos(sysz,sz, each_node_name)
                        this_node_val = raw_dft[pos_x, pos_y, pos_z]
                        if this_node_val > max_node[0]:
                            max_node = (this_node_val, each_node_name)
                                        
                    
                    sources[next_source] = {"value":each_val,
                                            "pos": inv_pos(sysz,sz,max_node[1]),
                                            "pos_id" : str(max_node[1]),
                                            "touching":False,
                                            "internal": True}
                    
                    sources_by_pos[str(max_node[1])] = next_source
                    sources_set = set(sources_by_pos)
                    
                    log.info("Source %s identifed positioned by %s " %  (next_source, max_node[1]))
                    next_source += 1
                    
                elif len(sources_touching_pos) == 1:
                    pass
                
                else:
                    touching = True
                    sources_touching = [sources_by_pos[key] for key in sources_touching_pos]
                    
                    log.info("Breaking at value %s" % each_val)

        if touching:
            #G = x.es(diag=False).subgraph()
            
           
            if not no_barriers:
                break
            
        #complete_graph.es(source_val_ge=each_val)["weight"] = 0
        

    if no_barriers:
        return 0,0,[]
        
    #elif len(sources_touching) < 2:
        
        #log.info("Spaces not touching")
        
        #return 0,0,[]
    
    #elif len(sources_touching) > 2:
        #log.info("Multiple touching spaces")
    
        #if True:
            #min_d = (sx+sy+sz)*1000
            #for a,b in product(sources_touching, sources_touching):
        
                #if a==b:
                    #continue
                #source_id=x.vs.find(name=sources[a]['pos_id']).index
                #target_id=x.vs.find(name=sources[b]['pos_id']).index
                ##log.info("source/target %s %s" %(type(source_id), type(target_id)))

                #d = x.shortest_paths_dijkstra(source_id,
                                              #target_id,
                                              #weights="weight")[0][0]
                
                #if d < min_d:
                    
                    #min_d = d 
                    #source_a = a
                    #source_b = b
                    ##source = int(sources[a]['pos_id'])
                    ##target = int(sources[b]['pos_id'])
                    
        #elif False:
            
            #max_d = 0
            #for a,b in product(sources_touching, sources_touching):
            
                #if a==b:
                    #continue
                
                #source_id=x.vs.find(name=sources[a]['pos_id']).index
                #target_id=x.vs.find(name=sources[b]['pos_id']).index
                ##log.info("%s:%s source/target %s %s" %(a,b,source_id, target_id))

                #d = x.shortest_paths_dijkstra(source_id,
                                              #target_id,
                                              #weights="weight")[0][0]
                ##log.info(repr(d))
                #if d > max_d:
                    #log.info (d)
                    #max_d = d 
                    #source_a = a
                    #source_b = b
                    #log.info(max_d, source_a, source_b)
                    ##source = int(sources[a]['pos_id'])
                    ##target = int(sources[b]['pos_id'])
        
        
        #else:
            #max_d = 0
            #for a,b in product(sources_touching, sources_touching):
        
                #if a==b:
                    #continue
                
                
                
                #d = (sources[a]['pos'][0]-sources[b]['pos'][0])**2 +\
                    #(sources[a]['pos'][1]-sources[b]['pos'][1])**2 +\
                    #(sources[a]['pos'][2]-sources[b]['pos'][2])**2 
                
                #if d > max_d:
                    #max_d = d 
                    #source_a = a
                    #source_b = b
                    #source = pos_(sysz, sz, sources[a]['pos'])
                    #target = pos_(sysz, sz, sources[b]['pos'])
            
    #else:
        #log.debug("Spaces touching")
        #source_a = sources_touching[0]
        #source_b = sources_touching[1]
        #source = pos_(sysz,sz,sources[sources_touching[0]]['pos'])
        #target = pos_(sysz,sz,sources[sources_touching[1]]['pos'])
    
    #with warnings.catch_warnings():
        #warnings.simplefilter("ignore")

        #log.info(source_a)
        #log.info(source_b)

        #log.info(sources[source_a]['pos_id'])
        #log.info(sources[source_b]['pos_id'])
        #sys.stdout.flush()


        #source_id=x.vs.find(name=sources[source_a]['pos_id']).index
        #target_id=x.vs.find(name=sources[source_b]['pos_id']).index
        

        #log.info(source_id)
        #log.info(target_id)        
        #paths = G.get_shortest_paths(source_id, target_id, mode="ALL") 
        
    
           
    #if len(paths[0]) == 0:
        #log.debug("path length zero")
        #return 0,0, []
    #else:
        
        #return source_a, source_b, [inv_pos(sysz, sz, int(G.vs[v]['name'])) for v in paths[0]]    
                    
    #return 0,0, []
    
@cy.cfunc
def push(heap : list,
         item : tuple) -> cy.void:
    """Push item onto heap, maintaining the heap invariant."""
    heap.append(item)
    _siftdown(heap, 0, len(heap)-1)

@cy.cfunc    
def _siftdown(heap : list,
              startpos : cy.int,
              pos : cy.int) -> cy.void:
    
    newitem : tuple = heap[pos]
    # Follow the path to the root, moving parents down until finding a place
    # newitem fits.
    
    parentpos : cy.int
    parent : tuple 
    while pos > startpos:
        parentpos  = (pos - 1) >> 1
        parent  = heap[parentpos]
        if newitem < parent:
            heap[pos] = parent
            pos = parentpos
            continue
        break
    heap[pos] = newitem

@cy.cfunc 
def pop(heap: list):
    """Pop the smallest item off the heap, maintaining the heap invariant."""
    
    lastelt : tuple = heap.pop()    # raises appropriate IndexError if heap is empty
    
    if heap:
        returnitem :tuple = heap[0]
        heap[0] = lastelt
        _siftup(heap, 0)
        return returnitem
    
    return lastelt

@cy.cfunc 
def _siftup(heap : list,
            pos : cy.int):
    
    endpos : cy.int = len(heap)
    startpos : cy.int = pos
    newitem : tuple = heap[pos]
    
    # Bubble up the smaller child until hitting a leaf.
    rightpos : cy.int
    childpos :cy.int = 2*pos + 1    # leftmost child position
    pos : cy.int
    
    while childpos < endpos:
        # Set childpos to index of smaller child.
        rightpos = childpos + 1
        if rightpos < endpos and not heap[childpos] < heap[rightpos]:
            childpos = rightpos
        # Move the smaller child up.
        heap[pos] = heap[childpos]
        pos  = childpos
        childpos = 2*pos + 1
    
    # The leaf at pos is empty now.  Put newitem there, and bubble it up
    # to its final resting place (by sifting its parents down).
    heap[pos] = newitem
    
    _siftdown(heap, startpos, pos)
        
@cy.ccall
def pos_(sysz : cy.int,
         sz : cy.int,
         node : tuple,
         ) -> cy.int :
    

    '''The pos_ function converts matrix position to a position integer'''
    
    return node[0]*sysz + node[1]*sz + node[2]

@cy.ccall
@cy.returns(tuple)
def inv_pos(sysz : cy.int,
            sz : cy.int,
            d : cy.int,
            ):##
##cpdef tuple inv_pos(int sy, int sz, int d):
    '''The inv_pos function converts position integer to matrix position'''
    
    x: cy.int = d//(sysz)
    rem: cy.int = d%(sysz)
    y :cy.int = rem//sz
    z :cy.int = rem%sz
    
    return x,y,z  

    

                      


